# admm-lone-solver-tests.jl
#
# Tests for methods in the ADMM L1 solver methods

using Convex;
using Mosek;

function AllADMMTests()
  TestDHSMultiplies();
  TestSolveQP_admm();
  TestSolveQP_graph_form(verbose_level=0);
  TestSolveQP_A_graph_form(verbose_level=0);
end

function TestDHSMultiplies()
  dd = 16;
  kk = 3;
  nn = kk * dd;
  Hd = Hadamard.hadmat(Hadamard.intlog(dd));
  H = kron(eye(kk), Hd) / sqrt(dd);
  num_tests = 20;
  num_failed = 0;  # Number of true that fail
  num_T_failed = 0;  # Number of transposes that fail
  n_work_vec = zeros(nn);
  d_work_vec = zeros(dd);
  for test_ind = 1:num_tests
    s_vec = sign(randn(nn)) / sqrt(kk);
    S = diagm(s_vec[1:dd]);
    for ll = 2:kk
      indset = ((ll - 1) * dd + 1):(ll * dd);
      S = vcat(S, diagm(s_vec[indset]));
    end
    D = randn(nn);
    x = randn(dd);
    v = randn(nn);
    DHSx_desired = diagm(D) * H * S * x;
    DHSx = zeros(nn);
    MultiplyByDHS!(x, s_vec, D, d_work_vec, DHSx);
    STHDv_desired = S' * H * diagm(D) * v;
    STHDv = zeros(dd);
    MultiplyByDHST!(v, s_vec, D, n_work_vec, STHDv);
    if (norm(DHSx - DHSx_desired) > 1e-10)
      num_failed += 1;
    end
    if (norm(STHDv - STHDv_desired) > 1e-10)
      num_T_failed += 1;
    end
  end
  if (num_failed > 0)
    warn("Failure of ", num_failed, " of ", num_tests,
         " tests of MultiplyByDHS!");
  else
    println("Passed all MultiplyByDHS! tests");
  end
  if (num_T_failed > 0)
    warn("Failure of ", num_T_failed, " of ", num_tests,
         " tests of MultiplyByDHST!");
  else
    println("Passed all MultiplyByDHST! tests");
  end
end

function TestSolveQP_admm(; eps_accuracy = 1e-8, rho=.05)
  srand(0);
  dd = 16;
  kk = 3;
  nn = kk * dd;
  Hd = Hadamard.hadmat(Hadamard.intlog(dd));
  H = kron(eye(kk), Hd) / sqrt(dd);
  num_tests = 5;
  num_failed = 0;
  for test_ind = 1:num_tests
    s_vec = sign(randn(nn)) / sqrt(kk);
    S = diagm(s_vec[1:dd]);
    for ll = 2:kk
      indset = ((ll - 1) * dd + 1):(ll * dd);
      S = vcat(S, diagm(s_vec[indset]));
    end
    D = randn(nn).^3;
    A = diagm(D) * H * S;
    c_vec = randn(nn);
    # Now get CVX objective
    x_cvx = Variable(dd);
    obj = minimize(norm(A * x_cvx - c_vec, 1) + .5 * sumsquares(x_cvx));
    solve!(obj, MosekSolver(QUIET=true));
    (x_fo, total_cgs) = SolveQP_admm(dd, s_vec, D, c_vec,
                                     eps_abs=eps_accuracy,
                                     eps_rel=eps_accuracy, verbose_level=0,
                                     MAXITER=25000, rho = rho);
    cvx_obj_val = (norm(A * x_cvx.value - c_vec, 1)
                   + .5 * norm(x_cvx.value, 2)^2);
    first_order_obj_val = (norm(A * x_fo - c_vec, 1)
                           + .5 * norm(x_fo, 2)^2);
    relative_diff = (first_order_obj_val - cvx_obj_val) / first_order_obj_val;
    println("*** Relative difference = ", relative_diff, " ***");
    if (relative_diff > 1e-3)
      num_failed += 1;
    end
  end
  if (num_failed > 0)
    warn("Failure of ", num_failed, " of ", num_tests,
         " tests of SolveQP_admm");
  else
    println("Passed all SolveQP_admm tests");
  end
end

function TestSolveQP_graph_form(; eps_accuracy = 1e-8, verbose_level=2, rho=1.0)
  srand(0);
  dd = 16;
  kk = 3;
  nn = kk * dd;
  Hd = Hadamard.hadmat(Hadamard.intlog(dd));
  H = kron(eye(kk), Hd) / sqrt(dd);
  num_tests = 5;
  num_failed = 0;
  for test_ind = 1:num_tests
    s_vec = sign(randn(nn)) / sqrt(kk);
    S = diagm(s_vec[1:dd]);
    for ll = 2:kk
      indset = ((ll - 1) * dd + 1):(ll * dd);
      S = vcat(S, diagm(s_vec[indset]));
    end
    D = randn(nn).^3;
    A = diagm(D) * H * S;
    c_vec = randn(nn);
    # Now get CVX objective
    x_cvx = Variable(dd);
    obj = minimize(norm(A * x_cvx - c_vec, 1) + .5 * sumsquares(x_cvx));
    solve!(obj, MosekSolver(QUIET=true));
    (x_fo, total_cgs) = SolveQP_graph_form(dd, s_vec, D, c_vec,
                                           eps_abs=eps_accuracy,
                                           eps_rel=eps_accuracy,
                                           verbose_level=verbose_level,
                                           MAXITER=25000, rho = float(rho));
    cvx_obj_val = (norm(A * x_cvx.value - c_vec, 1)
                   + .5 * norm(x_cvx.value, 2)^2);
    first_order_obj_val = (norm(A * x_fo - c_vec, 1)
                           + .5 * norm(x_fo, 2)^2);
    relative_diff = (first_order_obj_val - cvx_obj_val) / first_order_obj_val;
    println("*** Relative difference = ", relative_diff, " (",
            total_cgs, " total CG steps) ***");
    if (relative_diff > 1e-3)
      num_failed += 1;
    end
  end
  if (num_failed > 0)
    warn("Failure of ", num_failed, " of ", num_tests,
         " tests of SolveQP_admm");
  else
    println("Passed all SolveQP_admm tests");
  end
end

function TestSolveQP_A_graph_form(; eps_accuracy = 1e-8, rho=1.0,
                                  verbose_level=2)
  srand(0);
  dd = 25;
  nn = 100;
  num_tests = 5;
  num_failed = 0;
  for test_ind = 1:num_tests
    A = randn(nn, dd);
    c_vec = randn(nn);
    # Now get CVX objective
    x_cvx = Variable(dd);
    obj = minimize(norm(A * x_cvx - c_vec, 1) + .5 * sumsquares(x_cvx));
    solve!(obj, MosekSolver(QUIET=true));
    (x_fo, iters) =
      SolveQP_graph_form(A, c_vec, eps_abs=eps_accuracy,
                         eps_rel=eps_accuracy, verbose_level=verbose_level,
                         MAXITER=25000, rho = float(rho));
    cvx_obj_val = (norm(A * x_cvx.value - c_vec, 1)
                   + .5 * norm(x_cvx.value, 2)^2);
    first_order_obj_val = (norm(A * x_fo - c_vec, 1)
                           + .5 * norm(x_fo, 2)^2);
    relative_diff = (first_order_obj_val - cvx_obj_val) / first_order_obj_val;
    println("*** Relative difference = ", relative_diff, " (",
            iters, " iterations of FO) ***");
    if (relative_diff > 1e-3)
      num_failed += 1;
    end
  end
  if (num_failed > 0)
    warn("Failure of ", num_failed, " of ", num_tests,
         " tests of SolveQP_admm");
  else
    println("Passed all SolveQP_admm tests");
  end
end

