# admm-lone-solver.jl
#
# Uses ADMM to solve certain linear-quadratic problems.

# (x, cg_iters) = SolveQP_admm(d, s, D, c;
#                              x_init, eps_abs, eps_rel, rho,
#                              update_rho, verbose, MAXITER,
#                              run_data)
#
# Applies ADMM to solve quadratic problems of the form
#
# min. norm(A * x - c, 1) + (1/2) * norm(x)^2
#
# where A = DHS for matrices D, H, S of the following forms.
#
# D is a diagonal matrix, otherwise arbitrary
# S is a stacked block of diagoinal matrices,
#
#  S' = [S1 S2 ... Sk]
#
# where Si is a d-by-d diagonal matrix. d should be a power of 2.
#
# H is a block-diagonal Hadamard matrix, where
#
#  H = kron(eye(k), Hd) / sqrt(d)
#
# and Hd is the d-by-d standard {+1, -1}-valued Hadamard matrix,
# defined by the recursion
#
# H1 = [1 1; 1 -1];  H{d+1} = [Hd Hd; Hd -Hd].
function SolveQP_admm(dd::Int64, s_vec::Vector{Float64},
                      D::Vector{Float64}, c_vec::Vector{Float64};
                      x_init::Vector{Float64} = zeros(dd),
                      eps_abs::Float64 = 1e-5,
                      rho::Float64 = 0.5,
                      eps_rel::Float64 = 1e-4,
                      update_rho::Bool = false,
                      verbose_level::Int64 = 0,
                      MAXITER::Int64 = 1000,
                      run_data::Matrix{Float64} = zeros(0, 0))
  primal_residual_norm = Inf;
  dual_residual_norm = Inf;
  norm_Ax = 0.0;
  norm_ATy = 0.0;
  norm_y = 0.0;
  nn = length(s_vec);
  kk = convert(Int64, nn / dd);
  s_vec = sign.(s_vec) / sqrt(kk);
  x_iter = zeros(dd);
  y_iter = zeros(nn);
  n_work_vec = zeros(nn);
  d_work_vec = zeros(dd);
  resid_dual = zeros(dd);
  resid_primal = zeros(nn);
  Ax = zeros(nn);  # A * x
  ATy = zeros(dd);  # A' * y
  ATlambda = zeros(dd);  # A' * lambda
  lambda = zeros(nn);
  obj = Inf;
  total_cgs = 0;
  iter = 0;
  while ((primal_residual_norm >
          sqrt(nn) * eps_abs + eps_rel * max(norm_Ax, norm_y)) &&
         (dual_residual_norm >
          sqrt(nn) * eps_abs + eps_rel * norm_ATy) &&
         iter < MAXITER)
    iter += 1;
    (x_iter, num_cgs) =
      CGS.InvertHadamardSystem(dd, s_vec, rho, D.^2,
                               ATlambda - rho * ATy, x_init=x_iter);
    total_cgs += num_cgs;
    MultiplyByDHS!(x_iter, s_vec, D, d_work_vec, Ax);
    y_iter = c_vec + (sign.(Ax - lambda / rho - c_vec) .*
                      max.(abs.(Ax - lambda / rho - c_vec) - 1/rho, 0));
    resid_primal = Ax - y_iter;
    lambda -= rho * resid_primal;
    MultiplyByDHST!(resid_primal, s_vec, D, n_work_vec, resid_dual);
    MultiplyByDHST!(y_iter, s_vec, D, n_work_vec, ATy);
    MultiplyByDHST!(lambda, s_vec, D, n_work_vec, ATlambda);
    primal_residual_norm = norm(resid_primal);
    dual_residual_norm = rho * norm(resid_dual);
    if (update_rho)
      if (dual_residual_norm > 10 * primal_residual_norm)
        println("\t ** halving rho **");
        rho = rho / 2;
      elseif (primal_residual_norm > 10 * dual_residual_norm)
        println("\t ** doubling rho **");
        rho = rho * 2;
      end
    end
    norm_Ax = norm(Ax);
    norm_y = norm(y_iter);
    norm_ATy = norm(ATy);
    obj = norm(x_iter)^2 / 2 + norm(Ax - c_vec, 1);
    if (!isempty(run_data))
      run_data[iter, 1] = primal_residual_norm;
      run_data[iter, 2] = dual_residual_norm;
      run_data[iter, 3] = obj;
    end
    if (verbose_level >= 2)
      println("Primal: ", primal_residual_norm,
              "\tDual: ", dual_residual_norm,
              "\tObj: ", obj, "\tCGs: ", num_cgs);
    end
  end
  return (x_iter, total_cgs);
end

# (x, cg_iters) =
#   SolveQP_graph_form(d, s, D, c;
#                      x_init, eps_abs, eps_rel, rho,
#                      update_rho, verbose, MAXITER,
#                      run_data)
#
# Applies Parikh and Boyd's graph form proximal splitting to solve quadratic
# problems of the form
#
# min. norm(A * x - c, 1) + (1/2) * norm(x)^2
#
# where A = DHS for matrices D, H, S of the following forms.
#
# D is a diagonal matrix, otherwise arbitrary
# S is a stacked block of diagoinal matrices,
#
#  S' = [S1 S2 ... Sk]
#
# where Si is a d-by-d diagonal matrix. d should be a power of 2.
#
# H is a block-diagonal Hadamard matrix, where
#
#  H = kron(eye(k), Hd) / sqrt(d)
#
# and Hd is the d-by-d standard {+1, -1}-valued Hadamard matrix,
# defined by the recursion
#
# H1 = [1 1; 1 -1];  H{d+1} = [Hd Hd; Hd -Hd].
function SolveQP_graph_form(dd::Int64, s_vec::Vector{Float64},
                            D::Vector{Float64}, c_vec::Vector{Float64};
                            x_init::Vector{Float64} = zeros(dd),
                            eps_abs::Float64 = 1e-5,
                            rho::Float64 = 0.5,
                            eps_rel::Float64 = 1e-4,
                            update_rho::Bool = false,
                            verbose_level::Int64 = 0,
                            MAXITER::Int64 = 1000,
                            run_data::Matrix{Float64} = zeros(0, 0))
  primal_residual_norm = Inf;
  dual_residual_norm = Inf;
  norm_duals = 0.0;
  norm_primals = 0.0;
  nn = length(s_vec);
  kk = convert(Int64, nn / dd);
  s_vec = sign.(s_vec) / sqrt(kk);
  x_iter = zeros(dd);
  y_iter = zeros(nn);
  x_half = zeros(dd);
  y_half = zeros(nn);
  n_work_vec = zeros(nn);
  d_work_vec = zeros(dd);
  resid_dual = zeros(dd + nn);
  resid_primal = zeros(dd + nn);
  Ax = zeros(nn);  # A * x
  ATlambda = zeros(dd);  # Scratch vector for A' * stuff
  lambda = zeros(dd);
  nu = zeros(nn);
  obj = Inf;
  total_cgs = 0;
  iter = 0;
  while ((primal_residual_norm >
          sqrt(nn) * eps_abs + eps_rel * max(norm_primals)) &&
         (dual_residual_norm >
          sqrt(nn) * eps_abs + eps_rel * norm_duals) &&
         iter < MAXITER)
    iter += 1;
    # Store residuals for later use
    resid_dual[1:dd] = x_iter;
    resid_dual[(dd+1):end] = y_iter;
    # Compute update steps
    # x_half = (rho / (1 + rho)) * (x_iter - lambda);
    # n_work_vec = y_iter - nu - c_vec;
    # y_half = (c_vec + sign(n_work_vec) .* max(abs(n_work_vec) - 1/rho, 0));
    @inbounds begin
      for jj = 1:dd
        x_half[jj] = (rho / (1 + rho)) * (x_iter[jj] - lambda[jj]);
      end
      for jj = 1:nn
        n_work_vec[jj] = y_iter[jj] - nu[jj] - c_vec[jj];
        y_half[jj] = (c_vec[jj] + sign.(n_work_vec[jj]) .*
                      max.(abs.(n_work_vec[jj]) - 1/rho, 0));
      end
    end
    # TODO(jduchi): Could replace these inline additions to make
    # things run faster if we did them with for-loops.
    MultiplyByDHST!(y_half + nu, s_vec, D, n_work_vec, ATlambda);
    (x_iter, num_cgs) =
      CGS.InvertHadamardSystem(dd, s_vec, 1.0, D.^2,
                               ATlambda + x_half + lambda, x_init=x_iter);
    total_cgs += num_cgs;
    MultiplyByDHS!(x_iter, s_vec, D, d_work_vec, y_iter);
    # NOTE: These @inbounds methods make everything run much faster.
    # They avoid calling the "elementwise" and "collect" operations on
    # vectors.
    @inbounds for jj = 1:dd
      lambda[jj] += (x_half[jj] - x_iter[jj]);
    end
    @inbounds for jj = 1:nn
      nu[jj] += (y_half[jj] - y_iter[jj]);
    end

    # Primal residual = [x_{k+1}; y_{k+1}] - [x_{k + 1/2}, y_{k + 1/2}]
    # Dual residual = rho * [x_k - x_{k + 1}; y_k - y_{k + 1}].
    @inbounds for jj = 1:dd
      resid_primal[jj] = x_half[jj] - x_iter[jj];
      resid_dual[jj] -= x_iter[jj];
    end
    @inbounds for jj = 1:nn
      resid_primal[dd + jj] = y_half[jj] - y_iter[jj];
      resid_dual[dd + jj] -= y_iter[jj];
    end
    primal_residual_norm = norm(resid_primal);
    dual_residual_norm = rho * norm(resid_dual);
    if (update_rho)
      if (dual_residual_norm > 10 * primal_residual_norm)
        println("\t ** halving rho **");
        rho = rho / 2;
      elseif (primal_residual_norm > 10 * dual_residual_norm)
        println("\t ** doubling rho **");
        rho = rho * 2;
      end
    end
    norm_primals = max(sqrt(dot(x_iter, x_iter) + dot(y_iter, y_iter)),
                       sqrt(dot(y_half, y_half) + dot(x_half, x_half)));
    norm_duals = rho * sqrt(dot(lambda, lambda) + dot(nu, nu));
    obj = dot(x_iter, x_iter) / 2 + norm(y_iter - c_vec, 1);
    if (!isempty(run_data))
      run_data[iter, 1] = primal_residual_norm;
      run_data[iter, 2] = dual_residual_norm;
      run_data[iter, 3] = obj;
    end
    if (verbose_level >= 2)
      println("Primal: ", primal_residual_norm,
              "\tDual: ", dual_residual_norm,
              "\tObj: ", obj, "\tCGs: ", num_cgs);
    end
  end
  return (x_iter, total_cgs);
end

# (x, iter) = SolveQP_graph_form(A::Matrix{Float64}, c_vec::Vector{Float64};
#                                x_init::Vector{Float64} = zeros(size(A, 2)),
#                                eps_abs::Float64 = 1e-5,
#                                rho::Float64 = 1.0,
#                                eps_rel::Float64 = 1e-4,
#                                update_rho::Bool = false,
#                                verbose_level::Int64 = 0,
#                                MAXITER::Int64 = 1000,
#                                use_cg::Bool = false)
#
# Applies Parikh and Boyd's graph form proximal splitting to solve quadratic
# problems of the form
#
# min. norm(A * x - c, 1) + (1/2) * norm(x)^2
#
# starting from the given initialization point.
function SolveQP_graph_form(A::Matrix{Float64}, c_vec::Vector{Float64};
                            x_init::Vector{Float64} = zeros(size(A, 2)),
                            eps_abs::Float64 = 1e-5,
                            rho::Float64 = 1.0,
                            eps_rel::Float64 = 1e-4,
                            update_rho::Bool = false,
                            verbose_level::Int64 = 0,
                            MAXITER::Int64 = 1000,
                            use_cg::Bool = false)
  primal_residual_norm = Inf;
  dual_residual_norm = Inf;
  norm_duals = 0.0;
  norm_primals = 0.0;
  nn = length(c_vec);
  dd = size(A, 2);
  iter = 0;
  x_iter = zeros(dd);
  y_iter = zeros(nn);
  x_half = zeros(dd);
  y_half = zeros(nn);
  n_work_vec = zeros(nn);
  d_work_vec = zeros(dd);
  resid_dual = zeros(dd + nn);
  resid_primal = zeros(dd + nn);
  lambda = zeros(dd);
  nu = zeros(nn);
  obj = Inf;
  IplusAtAinv = 0;
  AtA = 0;
  if (use_cg)
    AtA = Symmetric(A' * A);
  else
    IplusAtAinv = inv(eye(dd) + A' * A);
  end
  while ((primal_residual_norm >
          sqrt(nn) * eps_abs + eps_rel * max(norm_primals)) &&
         (dual_residual_norm >
          sqrt(nn) * eps_abs + eps_rel * norm_duals) &&
         iter < MAXITER)
    iter += 1;
    # Store residuals for later use
    resid_dual[1:dd] = x_iter;
    resid_dual[(dd+1):end] = y_iter;
    # Compute update steps
    x_half = (rho / (1 + rho)) * (x_iter - lambda);
    n_work_vec = y_iter - nu - c_vec;
    y_half = c_vec + sign.(n_work_vec) .* max.(abs.(n_work_vec) - 1/rho, 0);
    d_work_vec = x_half + lambda + A' * (y_half + nu);
    if (use_cg)
      (x_iter, cg_iters) = CGS.ConjugateGradientSystem(AtA, d_work_vec);
    else
      x_iter = IplusAtAinv * d_work_vec;
    end
    y_iter = A * x_iter;
    lambda += (x_half - x_iter);
    nu += (y_half - y_iter);

    # Primal residual = [x_{k+1}; y_{k+1}] - [x_{k + 1/2}, y_{k + 1/2}]
    # Dual residual = rho * [x_k - x_{k + 1}; y_k - y_{k + 1}].
    resid_primal[1:dd] = x_half - x_iter;
    resid_primal[(dd+1):end] = y_half - y_iter;
    resid_dual[1:dd] -= x_iter;
    resid_dual[(dd+1):end] -= y_iter;
    primal_residual_norm = norm(resid_primal);
    dual_residual_norm = rho * norm(resid_dual);
    if (update_rho)
      if (dual_residual_norm > 10 * primal_residual_norm)
        println("\t ** halving rho **");
        rho = rho / 2;
      elseif (primal_residual_norm > 10 * dual_residual_norm)
        println("\t ** doubling rho **");
        rho = rho * 2;
      end
    end
    norm_primals = max(sqrt(dot(x_iter, x_iter) + dot(y_iter, y_iter)),
                       sqrt(dot(y_half, y_half) + dot(x_half, x_half)));
    norm_duals = rho * sqrt(dot(lambda, lambda) + dot(nu, nu));
    obj = dot(x_iter, x_iter) / 2 + norm(y_iter - c_vec, 1);
    if (verbose_level >= 2)
      println("Primal: ", primal_residual_norm,
              "\tDual: ", dual_residual_norm,
              "\tObj: ", obj, "\titer: ", iter);
    end
  end
  return (x_iter, iter);
end


# MultiplyByDHS!(x, s, D, d_work_vec, result)
#
# Sets result = diagm(D) * H * S * x, where H and S are defined as in
# the header to this file. work_vec is assumed to be a d-dimensional
# vector, where x is d-dimensional. This also assumes (without checking) that d
# is a power of 2.
function MultiplyByDHS!(x::Vector{Float64},
                        s_vec::Vector{Float64},
                        D::Vector{Float64},
                        d_work_vec::Vector{Float64},
                        result::Vector{Float64})
  result[:] = 0;
  dd = length(x);
  nn = length(s_vec);
  kk = round(Int64, nn / dd);
  @inbounds for ll = 1:kk
    ind_start = (ll - 1) * dd;
    # d_work_vec[:] = s_vec[indset] .* x;
    for jj = 1:dd
      d_work_vec[jj] = s_vec[ind_start + jj] * x[jj];
    end
    Hadamard.hadmult!(d_work_vec);
    # result[indset] += D[indset] .* d_work_vec;
    for jj = 1:dd
      result[ind_start + jj] += D[ind_start + jj] * d_work_vec[jj];
    end
  end
  @inbounds for jj = 1:nn
    result[jj] /= sqrt(dd);
  end
end

# MultiplyByDHST!(v, s, D, n_work_vec, result)
#
# Sets result = S' * H * diagm(D) * v, where H and S are defined as in
# the header to this file. The vector n_work_vec is a size n vector,
# where v is of size n as well. result is of size d.
function MultiplyByDHST!(v::Vector{Float64},
                         s_vec::Vector{Float64},
                         D::Vector{Float64},
                         n_work_vec::Vector{Float64},
                         result::Vector{Float64})
  dd = length(result);
  result[:] = 0;
  n_work_vec[:] = 0;
  nn = length(v);
  kk = round(Int64, nn / dd);
  # n_work_vec[:] = D .* v;
  @inbounds for jj = 1:nn
    n_work_vec[jj] = D[jj] * v[jj];
  end
  @inbounds for ll = 1:kk
    ind_start = (ll - 1) * dd;
    ind_end = ind_start + dd;
    indset = ((ll - 1) * dd + 1):(ll*dd);
    Hadamard.hadmult!(view(n_work_vec, indset));
    # result[:] += s_vec[indset] .* n_work_vec[indset];
    for jj = 1:dd
      result[jj] += s_vec[ind_start + jj] * n_work_vec[ind_start + jj];
    end
  end
  @inbounds for jj = 1:dd
    result[jj] /= sqrt(dd);
  end
end
