###############################################################################
############################### TESTING #######################################
###############################################################################

function TestAllCGs()
  TestNewtonSolve();
  TestNewtonMultiplications();
  TestMultiplyByOrthogonal();
  TestInvertHadamardSystem();
  TestMultiplyByIplusHDH();
end

function TestNewtonSolve()
  dd = 4;
  kk = 3;
  nn = dd * kk;
  H2 = [1 1; 1 -1];
  H4 = [H2 H2; H2 -H2] / 2;
  H2 = H2 / sqrt(2);
  H = kron(eye(kk), H4);
  num_tests = 20;
  num_failed = 0;
  for test_ind = 1:num_tests
    b = rand(dd + nn);
    s_vec = sign(1 - 2 * rand(nn));
    s_vec = s_vec / sqrt(kk);
    S = diagm(s_vec[1:dd]);
    for ll = 2:kk
      S = vcat(S, diagm(s_vec[((ll - 1)*dd + 1):(ll * dd)]));
    end
    D1 = rand(nn);
    D2 = rand(nn);
    D3 = rand(nn);
    D1 = max(D1, D2) + 1e-5;
    D3 = max(D2, D3) + 1e-5;
    Z = [H * S zeros(nn, nn); zeros(nn, dd) eye(nn)];
    G = [diagm(D1) diagm(D2); diagm(D2) diagm(D3)];
    A = Z' * G * Z;
    result_good = A \ b;
    (result_to_test, iters) =
      InvertNewtonSystem(nn, dd, s_vec, D1, D2, D3, b, eps_cg = 1e-8);
    if (norm(result_to_test - result_good) > 1e-6)
      num_failed += 1;
    end
  end
  if (num_failed > 0)
    warn("Failure of ", num_failed, " of ", num_tests, " Newton solves.");
  else
    println("Passed all InvertNewtonSystem tests");
  end
end

# TestNewtonMultiplications()
#
# Tests the methods MultiplyByNewtonMatrix! and MultiplyByPreconditioner! with
# random matrices, comparing with explicit calculations.
function TestNewtonMultiplications()
  dd = 4;
  kk = 3;
  nn = dd * kk;
  H2 = [1 1; 1 -1];
  H4 = [H2 H2; H2 -H2] / 2;
  H2 = H2 / sqrt(2);
  H = kron(eye(kk), H4);
  result_good = zeros(dd + nn);
  result_to_test = zeros(dd + nn);
  two_n_work_vec = zeros(2 * nn);
  d_work_vec = zeros(dd);
  n_work_vec = zeros(nn);
  num_tests = 20;
  num_newton_mult_failed = 0;
  num_pc_failed = 0;
  for test_ind = 1:num_tests
    x = randn(dd + nn);
    s_vec = sign(1 - 2 * rand(nn));
    s_vec = s_vec / sqrt(kk);
    S = diagm(s_vec[1:dd]);
    for ll = 2:kk
      S = vcat(S, diagm(s_vec[((ll - 1)*dd + 1):(ll * dd)]));
    end
    D1 = rand(nn);
    D2 = rand(nn);
    D3 = rand(nn);
    Z = [H * S zeros(nn, nn); zeros(nn, dd) eye(nn)];
    G = [diagm(D1) diagm(D2); diagm(D2) diagm(D3)];
    result_good = Z' * G * Z * x;
    MultiplyByNewtonMatrix!(nn, dd, kk, s_vec, D1, D2, D3, x,
                            two_n_work_vec, d_work_vec,
                            result_to_test);
    if (norm(result_to_test - result_good) > 1e-10)
      num_newton_mult_failed += 1;
    end
    # Now test preconditioner
    result_good = Z' * inv(G) * Z * x;
    MultiplyByPreconditioner!(nn, dd, kk, s_vec, D1, D2, D3, x,
                              n_work_vec, two_n_work_vec, d_work_vec,
                              result_to_test);
    if (norm(result_to_test - result_good) > 1e-10)
      num_pc_failed += 1;
    end
  end
  if (num_newton_mult_failed > 0)
    warn("Failure of ", num_newton_mult_failed, " of ", num_tests,
         " tests of Newton matrix multiplication");
  else
    println("Passed all Newton multiplication tests");
  end
  if (num_pc_failed > 0)
    warn("Failure of ", num_pc_failed, " of ", num_tests,
         " tests of preconditioner matrix multiplication");
  else
    println("Passed all preconditioner multiplication tests");
  end
end

# TestMultiplyByOrthogonal()
#
# Tests the methods MultiplyByOrthogonal! and MultiplyByOrthogonalT! with
# random matrices, comparing with explicit calculations.
function TestMultiplyByOrthogonal()
  dd = 4;
  kk = 3;
  nn = dd * kk;
  H2 = [1 1; 1 -1];
  H4 = [H2 H2; H2 -H2] / 2;
  H2 = H2 / sqrt(2);
  H = kron(eye(kk), H4);
  result_good = zeros(2 * nn);
  result_to_test = zeros(2 * nn);
  num_tests = 20;
  num_failed = 0;
  for test_ind = 1:num_tests
    x = randn(dd + nn);
    s_vec = sign(1 - 2 * rand(nn));
    s_vec = s_vec / sqrt(kk);
    S = diagm(s_vec[1:dd]);
    for ll = 2:kk
      S = vcat(S, diagm(s_vec[((ll - 1)*dd + 1):(ll * dd)]));
    end
    result_good[1:nn] = H * S * x[1:dd];
    result_good[(nn+1):end] = x[(dd+1):end];
    MultiplyByOrthogonal!(nn, dd, kk, s_vec, x, result_to_test);
    if (norm(result_to_test - result_good) > 1e-10)
      num_failed += 1;
    end
  end
  if (num_failed > 0)
    warn("Failure of ", num_failed, " of ", num_tests,
         " tests of MultiplyByOrthogonal");
  else
    println("Passed all orthogonal multiplication tests");
  end
  # Now test the Transposed version
  result_good = zeros(dd + nn);
  result_to_test = zeros(dd + nn);
  num_tests = 20;
  num_failed = 0;
  d_work_vec = zeros(dd);
  for test_ind = 1:num_tests
    x = randn(2 * nn);
    s_vec = sign(1 - 2 * rand(nn));
    s_vec = s_vec / sqrt(kk);
    S = diagm(s_vec[1:dd]);
    for ll = 2:kk
      S = vcat(S, diagm(s_vec[((ll - 1)*dd + 1):(ll * dd)]));
    end
    result_good[1:dd] = S' * H * x[1:nn];
    result_good[(dd+1):end] = x[(nn+1):end];
    MultiplyByOrthogonalT!(nn, dd, kk, s_vec, x, d_work_vec, result_to_test);
    if (norm(result_to_test - result_good) > 1e-10)
      num_failed += 1;
    end
  end
  if (num_failed > 0)
    warn("Failure of ", num_failed, " of ", num_tests,
         " tests of MultiplyByOrthogonalT");
  else
    println("Passed all transposed orthogonal multiplication tests");
  end
end

# TestInvertHadamardSystem()
#
# Tests that the method InvertHadamardSystem behaves as expected.
function TestInvertHadamardSystem()
  dd = 4;
  kk = 3;
  nn = dd * kk;
  H4 = Hadamard.hadmat(2);
  H4 = H4 / 2;
  H = kron(eye(kk), H4);
  result_good = zeros(dd);
  result_to_test = zeros(dd);
  num_tests = 20;
  num_failed = 0;
  num_precond_failed = 0;
  Id = eye(dd);
  for test_ind = 1:num_tests
    b = randn(dd);
    rho = rand();
    s_vec = sign(randn(nn)) / sqrt(kk);
    D = rand(nn).^4;
    alpha = rand();
    (result_to_test, cg_iters) =
      InvertHadamardSystem(dd, s_vec, rho, D, b, eps_cg = 1e-11,
                           alpha=alpha);
    S = diagm(s_vec[1:dd]);
    for ll = 2:kk
      indset = ((ll - 1) * dd + 1):(ll * dd);
      S = vcat(S, diagm(s_vec[indset]));
    end
    result_good = (alpha * Id + rho * S' * H * diagm(D) * H * S) \ b;
    if (norm(result_good - result_to_test) > 1e-10)
      num_failed += 1;
    end
    result_good = (Id + rho * S' * H * diagm(D) * H * S) \ b;
    (result_to_test, cg_iters) =
      InvertHadamardSystemP(dd, s_vec, rho, D, b, eps_cg = 1e-11);
    if (norm(result_good - result_to_test) > 1e-10)
      num_precond_failed += 1;
    end
  end
  if (num_failed > 0)
    warn("Failure of ", num_failed, " of ", num_tests,
         " tests of InvertHadamardSystem");
  else
    println("Passed all InvertHadamardSystem tests");
  end
  if (num_precond_failed > 0)
    warn("Failure of ", num_precond_failed, " of ", num_tests,
         " tests of InvertHadamardSystemP");
  else
    println("Passed all InvertHadamardSystemP tests");
  end
end

# TestMultiplyByIplusHDH()
#
# Tests multiplication of (I + rho * S' * H * D * H * S) by random
# vectors, as well as multiplication of
#
# W' * (I + rho * D)^{-1} * W
#
# for W = H * S.
function TestMultiplyByIplusHDH()
  dd = 4;
  kk = 3;
  nn = dd * kk;
  H4 = Hadamard.hadmat(2);
  H4 = H4 / 2;
  H = kron(eye(kk), H4);
  result_good = zeros(dd);
  result_to_test = zeros(dd);
  work_vec = zeros(dd);
  num_tests = 20;
  num_failed = 0;
  num_inv_failed = 0;
  Id = eye(dd);
  In = eye(nn);
  for test_ind = 1:num_tests
    x = randn(dd);
    s_vec = sign(1 - 2 * rand(nn)) / sqrt(kk);
    S = diagm(s_vec[1:dd]);
    for ll = 2:kk
      indset = ((ll - 1) * dd + 1):(ll * dd);
      S = vcat(S, diagm(s_vec[indset]));
    end
    D = rand(nn);
    rho = rand();
    alpha = rand();
    result_good = (alpha * Id + rho * S' * H * diagm(D) * H * S) * x;
    MultiplyByIplusHDH!(x, dd, kk, rho, s_vec, D, work_vec, result_to_test,
                        alpha);
    if (norm(result_to_test - result_good) > 1e-10)
      num_failed += 1;
    end
    result_good = (S' * H) * ((In + rho * diagm(D)) \ (H * S * x));
    MultiplyByIplusHDinvH!(x, dd, kk, rho, s_vec, D, work_vec, result_to_test);
    if (norm(result_to_test - result_good) > 1e-10)
      num_inv_failed += 1;
    end
  end
  if (num_failed > 0)
    warn("Failure of ", num_failed, " of ", num_tests,
         " tests of TestMultiplyByIplusHDH");
  else
    println("Passed all MultiplyByIplusHDH tests");
  end
  if (num_inv_failed > 0)
    warn("Failure of ", num_inv_failed, " of ", num_tests,
         " tests of TestMultiplyByIplusHDinvH");
  else
    println("Passed all MultiplyByIplusHDinvH tests");
  end
end

###### Explicit Constructions for Testing ######

function ExplicitlyConstructHDH(dd::Int64,
                                s_vec::Vector{Float64},
                                D::Vector{Float64})
  nn = length(s_vec);
  kk = convert(Int64, nn / dd);
  s_vec = sign(s_vec) / sqrt(kk);
  Hd = Hadamard.hadmat(Hadamard.intlog(dd)) / sqrt(dd);
  W = zeros(nn, dd);
  for ll = 1:kk
    indset = ((ll - 1) * dd + 1):(ll * dd);
    W[indset, :] = Hd * spdiagm(s_vec[indset]);
  end
  return W' * diagm(D) * W;
end

function ExplicitlyConstructNewtonMatrix(dd::Int64,
                                         s_vec::Vector{Float64},
                                         D1::Vector{Float64},
                                         D2::Vector{Float64},
                                         D3::Vector{Float64})
  nn = length(s_vec);
  kk = convert(Int64, nn / dd);
  H_small = hadmat(round(Int64, log2(dd))) / sqrt(dd);
  W = zeros(nn, dd);
  s_vec = sign(s_vec) / sqrt(kk);
  for ll = 1:kk
    indset = ((ll - 1) * dd + 1):(ll * dd);
    W[indset, :] = H_small * spdiagm(s_vec[indset]);
  end
  Z = [W zeros(nn, nn); zeros(nn, dd) eye(nn)];
  G = [diagm(D1) diagm(D2); diagm(D2) diagm(D3)];
  return Z' * G * Z;
end

function ExplicitlyConstructPreconditioner(dd::Int64,
                                           s_vec::Vector{Float64},
                                           D1::Vector{Float64},
                                           D2::Vector{Float64},
                                           D3::Vector{Float64})
  nn = length(s_vec);
  kk = convert(Int64, nn / dd);
  H_small = hadmat(round(Int64, log2(dd))) / sqrt(dd);
  W = zeros(nn, dd);
  s_vec = sign(s_vec) / sqrt(kk);
  for ll = 1:kk
    indset = ((ll - 1) * dd + 1):(ll * dd);
    W[indset, :] = H_small * spdiagm(s_vec[indset]);
  end
  Z = [W zeros(nn, nn); zeros(nn, dd) eye(nn)];
  inv_part = 1./(D1 .* D3 - D2.^2);
  Ginv = [diagm(D3 .* inv_part) -diagm(D2 .* inv_part);
          -diagm(D2 .* inv_part) diagm(D1 .* inv_part)];
  # Ginv = [eye(nn) zeros(nn, nn); zeros(nn, nn) diagm(1./D3)];
  return Z' * Ginv * Z;
end
