# conjugate-solver.jl
#
# Some solvers for particular conjugate gradient methods

module CGS

include("Hadamard.jl");
include("conjugate-solver-tests.jl");

# (x, iters) = ConjugateGradientSystem(M::Symmetric{Float64, Array{Float64,2}},
#                                      b::Vector{Float64};
#                                      x_init::Vector{Float64} = zeros(0),
#                                      eps_cg::Float64 = 1e-6)
#
# Performs conjugate gradient for (at most) n steps, where n is the dimension
# of the symmetric matrix M, to solve the positive definite system
#
# b = (eye(n) + M) * x
#
# Begins iterations from the given x_init vector.
function ConjugateGradientSystem(M::Symmetric{Float64, Array{Float64,2}},
                                 b::Vector{Float64};
                                 x_init::Vector{Float64} = zeros(0),
                                 eps_cg::Float64 = 1e-6)
  nn = size(M, 1);
  maxiter = nn;
  x = zeros(nn);
  if (!isempty(x_init))
    x[:] = x_init[:];
  end
  if (nn != length(b))
    error("Bad vector sizes.");
  end
  resid = b - (M * x + x);
  p_vec = copy(resid);
  resid_norm_square = dot(resid, resid);
  norm_b_square = dot(b, b);
  iter = 0;
  Mp = zeros(nn);
  while (iter < nn && resid_norm_square > eps_cg^2 * norm_b_square)
    Mp = M * p_vec + p_vec;
    alpha = resid_norm_square / dot(p_vec, Mp);
    x += alpha * p_vec;
    resid -= alpha * Mp;
    new_resid_square = dot(resid, resid);
    p_vec = resid + (new_resid_square / resid_norm_square) * p_vec;
    resid_norm_square = new_resid_square;
    iter += 1;
  end
  return (x, iter);
end

# (x, iter) = InvertHadamardSystem(d, s, rho, D, b; x0, eps_cg, residuals)
#
# Sets x to be the solution to the (positive definite) linear equation
#
# (alpha * I + rho * S' * H * D * H * S) * x = b
#
# where I is the d-by-d identity, D is the corresponding diagonal
# matrix to the vector D, and S is the matrix
#
# S' = [S1 S2 ... Sk]
#
# of diagonal matrices Si. The Si are d-by-d diagonal matrices, where
# the i-th such matrix has diagonal entries given by the i-th block of
# entries of the vector s. We scale s to have entries +/-1 / sqrt(k).
#
# The CG method is solved to relative tolerance eps_cg (default 10^(-6)).
# Also returns the number of iterations of CG.
function InvertHadamardSystem(dd::Int64,
                              s_vec::Vector{Float64},
                              rho::Float64,
                              D::Vector{Float64},
                              b::Vector{Float64};
                              x_init::Vector{Float64} = zeros(0),
                              eps_cg::Float64 = 1e-6,
                              residuals::Vector{Float64} = Array{Float64}(0),
                              alpha::Float64 = 1.0)
  nn = length(s_vec);
  kk = convert(Int64, nn / dd);
  # s_vec = sign(s_vec) / sqrt(kk);
  maxiter = dd;
  A_p_vec = zeros(dd);
  d_work_vec = zeros(dd);
  x = zeros(dd);
  if (!isempty(x_init))
    x[:]  = x_init[:];
  end
  # Check sizes:
  if (nn != length(s_vec) || nn != length(D) || dd != length(b))
    error("Bad vector sizes.");
  end
  MultiplyByIplusHDH!(x, dd, kk, rho, s_vec, D,
                      d_work_vec, A_p_vec, alpha);
  # NOTE: Replaced with @inbounds for faster looping and construction
  # resid = b - A_p_vec;
  resid = zeros(dd);
  @inbounds for jj = 1:dd
    resid[jj] = b[jj] - A_p_vec[jj];
  end
  p_vec = copy(resid);
  resid_square = dot(resid, resid);
  norm_b_squared = dot(b, b);
  iter = 0;
  while (resid_square / norm_b_squared > eps_cg^2 &&
         iter <= maxiter)
    iter = iter + 1;
    MultiplyByIplusHDH!(p_vec, dd, kk, rho, s_vec, D,
                        d_work_vec, A_p_vec, alpha);
    stepsize = resid_square / dot(p_vec, A_p_vec);
    # Replaced additions with for loops for speed
    for jj = 1:dd
      x[jj] += stepsize * p_vec[jj];
      resid[jj] -= stepsize * A_p_vec[jj];
    end
    new_resid_square = dot(resid, resid);
    if (!isempty(residuals))
      residuals[iter] = new_resid_square;
    end
    # p_vec = resid + (new_resid_square / resid_square) * p_vec;
    @inbounds for jj = 1:dd
      p_vec[jj] = resid[jj] + (new_resid_square / resid_square) * p_vec[jj];
    end
    resid_square = new_resid_square;
  end
  return (x, iter);
end

# (x, iter) = InvertHadamardSystemP(d, s, rho, D, b; x0, eps_cg, residuals)
#
# Uses a preconditioned conjugate gradient method (instead of a
# vanilla CG method as above) to solve the linear system in
# InvertHadamardSystem.
#
# Practically speaking, it appears that the non-pre-conditioned method
# has better performance.
function InvertHadamardSystemP(dd::Int64,
                               s_vec::Vector{Float64},
                               rho::Float64,
                               D::Vector{Float64},
                               b::Vector{Float64};
                               x_init::Vector{Float64} = zeros(0),
                               eps_cg::Float64 = 1e-6,
                               residuals::Vector{Float64} = Array{Float64}(0))
  nn = length(s_vec);
  kk = convert(Int64, nn / dd);
  # s_vec = sign(s_vec) / sqrt(kk);
  maxiter = dd;
  A_p_vec = zeros(dd);
  d_work_vec = zeros(dd);
  x = zeros(dd);
  if (!isempty(x_init))
    x[:]  = x_init[:];
  end
  MultiplyByIplusHDH!(x, dd, kk, rho, s_vec, D,
                      d_work_vec, A_p_vec);
  resid = b - A_p_vec;
  z_vec = zeros(dd);
  MultiplyByIplusHDinvH!(resid, dd, kk, rho, s_vec, D, d_work_vec, z_vec);
  p_vec = copy(z_vec);
  old_resid_dot_z = dot(resid, z_vec);
  norm_b_squared = dot(b, b);
  for ii = 1:maxiter
    MultiplyByIplusHDH!(p_vec, dd, kk, rho, s_vec, D,
                        d_work_vec, A_p_vec);
    alpha = old_resid_dot_z / dot(p_vec, A_p_vec);
    x += alpha * p_vec;
    resid -= alpha * A_p_vec;
    new_resid_square = dot(resid, resid);
    if (!isempty(residuals))
      residuals[ii] = new_resid_square;
    end
    if (new_resid_square / norm_b_squared < eps_cg^2)
      return (x, ii);
    end
    MultiplyByIplusHDinvH!(resid, dd, kk, rho, s_vec, D, d_work_vec, z_vec);
    resid_dot_z = dot(z_vec, resid);
    beta = resid_dot_z / old_resid_dot_z;
    old_resid_dot_z = resid_dot_z;
    p_vec = z_vec + beta * p_vec;
  end
  return (x, maxiter);
end

# MultiplyByIplusHDH!(x, d, k, rho, s_vec, D, d_work_vec, result, alpha=1)
#
# Sets result to be equal to
#
# (alpha * I + rho * S' * H * D * H * S) * x
#
# where I is the d-by-d identity, D is the corresponding diagonal
# matrix to the vector D, and S is the matrix
#
# S' = [S1 S2 ... Sk]
#
# of diagonal matrices Si. The vector d_work_vec is assumed to be of
# size d, where x and result are both vectors of dimension d.
function MultiplyByIplusHDH!(x::Vector{Float64},
                             dd::Int64,
                             kk::Int64,
                             rho::Float64,
                             s_vec::Vector{Float64},
                             D::Vector{Float64},
                             d_work_vec::Vector{Float64},
                             result::Vector{Float64},
                             alpha::Float64 = 1.0)
  result[:] = 0;
  @inbounds for ll = 1:kk
    init_ind = (ll - 1) * dd;
    # indset = (init_ind + 1):(init_ind + dd)
    # d_work_vec[:] = s_vec[indset] .* x;
    for jj = 1:dd
      d_work_vec[jj] = s_vec[init_ind + jj] * x[jj];
    end
    Hadamard.hadmult!(d_work_vec);
    # d_work_vec .*= D[indset];
    for jj = 1:dd
      d_work_vec[jj] *= D[init_ind + jj];
    end
    Hadamard.hadmult!(d_work_vec);
    d_work_vec /= dd;
    # result[:] += rho * s_vec[indset] .* d_work_vec;
    for jj = 1:dd
      result[jj] += rho * s_vec[init_ind + jj] * d_work_vec[jj];
    end
  end
  result[:] += alpha * x;
end

# MultiplyByIplusHDinvH!(x, d, k, rho, s_vec, D, d_work_vec, result)
#
# Sets result to be equal to
#
# W' * (I + rho * D)^{-1} * W * x
#
# where W = H * S, as in the description of InvertHadamardSystem.
function MultiplyByIplusHDinvH!(x::Vector{Float64},
                                dd::Int64,
                                kk::Int64,
                                rho::Float64,
                                s_vec::Vector{Float64},
                                D::Vector{Float64},
                                d_work_vec::Vector{Float64},
                                result::Vector{Float64})
  result[:] = 0;
  @inbounds for ll = 1:kk
    indset = ((ll - 1) * dd + 1):(ll * dd);
    d_work_vec[:] = s_vec[indset] .* x;
    Hadamard.hadmult!(d_work_vec);
    d_work_vec ./= (1 + rho * D[indset]);
    Hadamard.hadmult!(d_work_vec);
    d_work_vec /= dd;
    result[:] += s_vec[indset] .* d_work_vec;
  end
end

# (x, iter) = InvertNewtonSystem(nn::Int64, dd::Int64,
#                        s_vec::Vector{Float64},
#                        D1::Vector{Float64},
#                        D2::Vector{Float64},
#                        D3::Vector{Float64},
#                        b::Vector{Float64};
#                        x_init::Vector{Float64} = zeros(dd + nn),
#                        eps_cg::Float64 = 1e-6,
#                        residual_norms::Vector{Float64} = Array{Float64}(0))
#
# Let the matrix S be defined by
#
#  S' = [S1 S2 ... Sk]
#
# where the matrices Si are d-by-d diagonal matrices, and k = n/d, and Si is
# the diagonal of the corresponding ith block of the vector s_vec. We force
# s_vec to have entries in {-1 / sqrt(k), 1/sqrt(k)}. Then S is an n-by-d
# matrix. Let H be the matrix
#
#  H = kron(eye(k), Hd)
#
# where Hd denotes the d-by-d (natural) Hadamard matrix, which satisfies Hd =
# Hd' and Hd^2 = I_d, the d-by-d identity.
#
# Let the matrix A be defined by
#
# A = [W 0]^T [ D1 D2 ] [W 0]
#     [0 I]   [ D2 D3 ] [0 I]
#
# where Di are the corresponding diagonals of the given vectors Di, and
#
#  W = H * S,
#
# so that W' * W = I_d, the d-by-d identity matrix. This method uses
# preconditioned conjugate gradients to return the solution A x = b, where we
# solve to relative residual error eps_cg.
#
# See the associated notes on preconditioning for explanations of our
# preconditioner.
function InvertNewtonSystem(nn::Int64, dd::Int64,
                            s_vec::Vector{Float64},
                            D1::Vector{Float64},
                            D2::Vector{Float64},
                            D3::Vector{Float64},
                            b::Vector{Float64};
                            x_init::Vector{Float64} = zeros(dd + nn),
                            eps_cg::Float64 = 1e-6,
                            eps_cg_abs::Float64 = 1e-6,
                            residual_norms::Vector{Float64} = Array{Float64}(0))
  kk = convert(Int64, nn / dd);
  s_vec = sign(s_vec) / sqrt(kk);
  maxiter = 2 * nn;
  two_n_work_vec = zeros(2 * nn);
  n_work_vec = zeros(nn);
  d_work_vec = zeros(dd);
  z_vec = zeros(dd + nn);
  Ap_vec = zeros(dd + nn);
  x = copy(x_init);
  if (length(D1) != nn || length(D2) != nn || length(D3) != nn ||
      length(b) != (dd + nn) || length(s_vec) != nn)
    error("Wrong vector sizes in Newton multiplication");
  elseif (2^Hadamard.intlog(dd) != dd)
    error("Need dimension to be a power of 2");
  end
  MultiplyByNewtonMatrix!(nn, dd, kk, s_vec, D1, D2, D3, x_init,
                          two_n_work_vec, d_work_vec,
                          Ap_vec);
  resid = b - Ap_vec;
  # MultiplyByPreconditioner!(nn, dd, kk, s_vec, D1, D2, D3, resid,
  #                           n_work_vec, two_n_work_vec, d_work_vec, z_vec);
  MultiplyBySimplePreconditioner!(dd, D1, D3, resid, z_vec);
  p_vec = copy(z_vec);
  norm_b = norm(b);
  old_resid_dot_z = dot(resid, z_vec);
  for ii = 1:maxiter
    MultiplyByNewtonMatrix!(nn, dd, kk, s_vec, D1, D2, D3, p_vec,
                            two_n_work_vec, d_work_vec, Ap_vec);
    alpha = old_resid_dot_z / dot(p_vec, Ap_vec);
    x = x + alpha * p_vec;
    resid = resid - alpha * Ap_vec;
    resid_norm = norm(resid);
    if (!isempty(residual_norms))
      residual_norms[ii] = resid_norm;
    end
    if (resid_norm / norm_b < eps_cg || resid_norm < eps_cg_abs)
      return (x, ii);
    end
    # MultiplyByPreconditioner!(nn, dd, kk, s_vec, D1, D2, D3, resid,
    #                            n_work_vec, two_n_work_vec, d_work_vec, z_vec);
    MultiplyBySimplePreconditioner!(dd, D1, D3, resid, z_vec);
    resid_dot_z = dot(z_vec, resid);
    beta = resid_dot_z / old_resid_dot_z;
    old_resid_dot_z = resid_dot_z;
    p_vec = z_vec + beta * p_vec;
  end
  return (x, maxiter);
end

# MultiplyByNewtonMatrix!(nn::Int64, dd::Int64, kk::Int64,
#                                 s_vec::Vector{Float64},
#                                 D1::Vector{Float64},
#                                 D2::Vector{Float64},
#                                 D3::Vector{Float64},
#                                 x::Vector{Float64},
#                                 two_n_work_vec::Vector{Float64},
#                                 d_work_vec::Vector{Float64},
#                                 result::Vector{Float64})
#
# Multiplies the vector x by the matrix
#
#  [W 0]^T [D1 D2] [W 0]
#  [0 I]   [D2 D3] [0 I]
#
# as described above, where the vectors two_n_work_vec and d_work_vec are of
# sizes nn and dd, respectively, and are used to avoid memory allocations.
function MultiplyByNewtonMatrix!(nn::Int64, dd::Int64, kk::Int64,
                                 s_vec::Vector{Float64},
                                 D1::Vector{Float64},
                                 D2::Vector{Float64},
                                 D3::Vector{Float64},
                                 x::Vector{Float64},
                                 two_n_work_vec::Vector{Float64},
                                 d_work_vec::Vector{Float64},
                                 result::Vector{Float64})
  MultiplyByOrthogonal!(nn, dd, kk, s_vec, x, two_n_work_vec);
  # Use result to store intermediate calculations.
  @inbounds begin
    result[1:nn] = (D1 .* two_n_work_vec[1:nn] +
                    D2 .* two_n_work_vec[(nn+1):(2*nn)]);
    two_n_work_vec[(nn+1):end] = (D2 .* two_n_work_vec[1:nn] +
                                  D3 .* two_n_work_vec[(nn+1):(2*nn)]);
    two_n_work_vec[1:nn] = result[1:nn];
  end
  MultiplyByOrthogonalT!(nn, dd, kk, s_vec, two_n_work_vec, d_work_vec, result);
end

function MultiplyBySimplePreconditioner!(dd::Int64,
                                         D1::Vector{Float64},
                                         D3::Vector{Float64},
                                         x::Vector{Float64},
                                         result::Vector{Float64})
  @inbounds begin
    # result[1:dd] = x[1:dd];
    result[1:dd] = x[1:dd] / mean(D1);
    result[(dd+1):end] = x[(dd+1):end] ./ D3;
  end
end

function MultiplyByPreconditioner!(nn::Int64, dd::Int64,
                                   kk::Int64,
                                   s_vec::Vector{Float64},
                                   D1::Vector{Float64},
                                   D2::Vector{Float64},
                                   D3::Vector{Float64},
                                   x::Vector{Float64},
                                   n_work_vec::Vector{Float64},
                                   two_n_work_vec::Vector{Float64},
                                   d_work_vec::Vector{Float64},
                                   result::Vector{Float64})
  MultiplyByOrthogonal!(nn, dd, kk, s_vec, x, two_n_work_vec);
  n_work_vec[:] = 1 ./ (D1 .* D3 - D2.^2);
  @inbounds begin
    # Use result to store intermediate calculations
    result[1:nn] = (D3 .* n_work_vec .* two_n_work_vec[1:nn]
                    - D2 .* n_work_vec .* two_n_work_vec[(nn+1):(2*nn)]);
    two_n_work_vec[(nn + 1):(2*nn)] =
    (-D2 .* n_work_vec .* two_n_work_vec[1:nn]
     + D1 .* n_work_vec .* two_n_work_vec[(nn+1):(2*nn)]);
    two_n_work_vec[1:nn] = result[1:nn];
  end
  MultiplyByOrthogonalT!(nn, dd, kk, s_vec, two_n_work_vec, d_work_vec, result);
end

# MultiplyByOrthogonal!(nn::Int64, dd::Int64, kk::Int64,
#                                s_vec::Vector{Float64},
#                                x::Vector{Float64},
#                                result::Vector{Float64})
#
# Sets the (2 * n)-dimensional vector result to be equal to
#
# [W 0] x
# [0 I]
#
# where W = H * S for the matrices H and S defined in InvertNewtonSystem. The
# vector x should be of size d + n.
function MultiplyByOrthogonal!(nn::Int64, dd::Int64, kk::Int64,
                               s_vec::Vector{Float64},
                               x::Vector{Float64},
                               result::Vector{Float64})
  sqrt_d = sqrt(dd);
  @inbounds for ll = 1:kk
    indset = ((ll - 1) * dd + 1):(ll * dd);
    result[indset] = s_vec[indset] .* x[1:dd];
    Hadamard.hadmult!(view(result, indset));
    result[indset] /= sqrt_d;
  end
  @inbounds result[(nn+1):(2 * nn)] = x[(dd+1):end];
end

# MultiplyByOrthogonalT!(nn::Int64, dd::Int64, kk::Int64,
#                                s_vec::Vector{Float64},
#                                x::Vector{Float64},
#                                result::Vector{Float64})
#
# Sets the (d+n)-dimensional vector result to be equal to
#
# [W 0]^T x
# [0 I]
#
# where W = H * S for the matrices H and S defined in InvertNewtonSystem.
# The vector x should be of size 2 * n.
function MultiplyByOrthogonalT!(nn::Int64, dd::Int64, kk::Int64,
                                s_vec::Vector{Float64},
                                x::Vector{Float64},
                                d_work_vec::Vector{Float64},
                                result::Vector{Float64})
  result[1:dd] = 0;
  sqrt_d = sqrt(dd);
  @inbounds for ll = 1:kk
    indset = ((ll - 1) * dd + 1):(ll * dd);
    d_work_vec[:] = x[indset];
    Hadamard.hadmult!(d_work_vec);
    result[1:dd] += s_vec[indset] .* d_work_vec / sqrt_d;
  end
  @inbounds result[(dd+1):end] = x[(nn+1):end];
end

end  # Module CGS
