# digit-experiment.jl
#
# Experiments with some digit classification datasets

module DigitExperiment
using PyPlot;

data_path = "./Data/";

if (!isdefined(:PRS))
  include("solvers.jl");
end

# (X, labels) = LoadDatasets(filename = "zip.train")
#
# Loads a collection of digits from the given file.
function LoadDatasets(filename = "zip.train")
  # First, read the training data
  X = readdlm(string(data_path, filename), ' ', Float64, '\n');
  labels = X[:, 1];
  X = X[:, 2:end];
  labels = convert(Vector{Int}, labels);
  return (X, labels);
end

# DisplayDigit(digit::Array{Float64})
#
# Given a vector (length 256) representation of a digit, plots the
# digit.
function DisplayDigit(digit::Array{Float64})
  figure();
  @assert length(digit[:]) == 256
  imshow(reshape(digit[:], 16, 16)', cmap="gray", interpolation="nearest");
  xticks([]); yticks([]);
end

# DisplayBehaviorOfOne(digit; k_factor = 3, p_fail = 0.0)
# 
function DisplayBehaviorOfOne(digit::Array{Float64};
                              k_factor::Int64 = 3, p_fail::Real = 0.0)
  (b, s_vec) = PRS.HadamardSignedData(digit, k_factor, p_fail = p_fail,
                                      dev = 0.0);
  dd = length(digit);
  digit_init = PRS.OrthogonalityInitializer(dd, s_vec, b, maxiter = 200);
  # digit_init = min(max(digit_init, -1.0), 1.0);
  (digit_est, objs, cg_steps) =
    PRS.ProxLinearQuadratic(digit_init, s_vec, b, maxiter = 30,
                            qp_solver = :graph, eps_qp=1e-4,
                            verbose_level=0);
  (digit_est, objs, cg_steps) =
    PRS.ProxLinearQuadratic(digit_est, s_vec, b, maxiter = 30,
                            qp_solver = :graph, eps_qp=1e-6,
                            verbose_level=0);
  flip_sign = sign(norm(digit_est - digit) - norm(digit_est + digit));
  
  println("Relative error: ",
          norm(digit_est + flip_sign * digit) / norm(digit));
  DisplayDigit(flip_sign * digit_init);
  DisplayDigit(flip_sign * digit_est);
  DisplayDigit(-digit);
end

# (num_success, total_cgs) = PerformExperiment(X; k_factor,
#                                              p_fail, dev, success_tol)
#
# Performs a digit experiment with the data given in the matrix X.
# Takes d * k measurements of each digit, corrupting p_fail of them,
# then uses the orthogonality initializer to seed the
# ProxLinearQuadratic method.
function PerformExperiment(X::Matrix{Float64};
                           k_factor::Int64 = 4,
                           p_fail::Real = 0.0,
                           dev::Real = 10.0,
                           success_tol::Real = 1e-2)
  dd = size(X, 2);
  num_success = 0;
  total_cgs = 0;
  for ii = 1:size(X, 1)
    digit = X[ii, :][:];
    (b, s_vec) = PRS.HadamardSignedData(digit, k_factor,
                                         p_fail = p_fail, dev = dev);
    digit_init = PRS.OrthogonalityInitializer(dd, s_vec, b, maxiter = 200);
    (digit_est, objs, cg_steps) =
      PRS.ProxLinearQuadratic(digit_init, s_vec, b, maxiter = 30,
                              qp_solver = :graph, eps_qp=1e-4,
                              verbose_level=0);
    rel_err = (min(norm(digit_est - digit), norm(digit_est + digit))
               / norm(digit));
    num_success += (rel_err < success_tol);
    total_cgs += cg_steps;
    println("[Experiment ", ii, " of ", size(X, 1), "] ",
            "rel error = ", rel_err,
            ", CGs = ", cg_steps);
  end
  return (num_success, total_cgs);
end

# (p_fail_vals, num_successes, num_cgs) = ExplorePFailures(n, k_factor)
#
# Performs 9 experiments, with failure (outlier) probabilities from 0
# through .2, of the PerformExperiment method. Uses the given k_factor
# and the first n datapoints in the digit dataset of LoadDatasets().
function ExplorePFailures(nn::Int64 = 500,
                          k_factor::Int64 = 4)
  (X, labels) = LoadDatasets();
  n_p_fails = 9;
  p_fail_vals = linspace(0, .2, n_p_fails);
  num_successes = zeros(Int64, n_p_fails);
  num_cgs = zeros(Int64, n_p_fails);
  for p_ind = 1:n_p_fails
    println("*** Experiment with p_fail = ", p_fail_vals[p_ind], " ***");
    (num_successes[p_ind],
     num_cgs[p_ind]) = PerformExperiment(X[1:nn, :], k_factor = k_factor,
                                         p_fail = p_fail_vals[p_ind]);
  end
  return (p_fail_vals, num_successes, num_cgs);
end

end  # module DigitExperiment
