# generate-data.jl
#
# This file performs the experiments and generates the figures associated with
# the paper
#
#  "Solving (most) of a set of quadratic equalities: Composite optimization for
#   robust phase retrieval"
#
# by John Duchi and Feng Ruan.
#
# The methods in this file can sometimes be *very* expensive, as there are a
# large number of experiments required to generate the data and replicate
# things. As such, partial experiments may be more useful, in which case the
# experimenter should simply comment out everything else.
#
# In the paper, we had access to a large cluster, which made everything
# substantially easier.

include("phase-experiments.jl");
include("solvers.jl");
PE = PhaseExperiments;
savepath = "./CompositeResults/";

# Each of the true/false flags below determines whether a particular
# set of experiments runs.

run_zero_noise = true;  # Whether to run (expensive) zero-noise experiments
run_outliers = true;  # Whether to run the experiments testing phase
                      # retrieval with outliers
run_digits = true;  # Whether to run the digit experiments
run_image = true;  # Whether to run the experiment with the RNA image.

############################################################
######### ---- Part 1: Zero-noise experiments ---- #########
############################################################

num_n_vals = 10;  # Number of different sample sizes to run.
num_tests_per = 100;  # Number of tests per individual sample size to run.
seed = 4634;  # Number chosen by rolling a die 4 times.

if (run_zero_noise)
  for dim = [400, 800, 1000, 3000]
    filename = string("pz_dim", dim, "_tests", num_tests_per,
                      "_nvals", num_n_vals,
                      "_seed", seed,
                      "_prox_lowtol_signs_biginit.txt");
    # Proximal-linear algorithm
    PE.ZeroNoisePhaseTransitionExperiment(filename,
                                          savepath=savepath,
                                          dim=dim,
                                          num_n_vals = num_n_vals,
                                          tolerance=1e-5,
                                          num_tests_per = num_tests_per,
                                          prop_init = .167,
                                          init_type = :LM,
                                          alg_type = :PL,
                                          signal_type = :signs,
                                          seed = seed);

    # Experiment with big initialization for Truncated Amplitude Flow (TAF)
    filename = string("pz_dim", dim, "_tests", num_tests_per,
                      "_nvals", num_n_vals,
                      "_seed", seed,
                      "_taf_lowtol_signs_biginit.txt");
    PE.ZeroNoisePhaseTransitionExperiment(filename,
                                          savepath=savepath,
                                          dim=dim,
                                          num_n_vals = num_n_vals,
                                          tolerance=1e-5,
                                          num_tests_per = num_tests_per,
                                          prop_init = .167,
                                          init_type = :LM,
                                          alg_type = :TAF,
                                          signal_type = :signs,
                                          seed = seed);
  end
end  # run_zero_noise

#############################################################
########### ---- Part 2: Outlier experiments ---- ###########
#############################################################

dim = 200;  # Dimension of experiment
# Different values for probability of failed measurement. Does a
# linearly spaced set of values between 0 and .3, with num_p_fail_vals
# thresholds.
num_p_fail_vals = 31;
seed = 5155;  # Number chosen by rolling a die

# Try three different algorithm types: small (:SM) for smallest magnitude
# measurements, largest magnitude measurements (:LM), and median-like
# initialization (:MTW).

if (run_outliers)
  # Proximal-linear algorithm
  filename = string("po_dim", dim, "_tests", num_tests_per,
                    "_pvals", num_p_fail_vals,
                    "_seed", seed,
                    "_proxgraph_lowtol_gauss_smallinit.txt");
  PE.PhaseTransitionExperiment(filename,
                               savepath=savepath,
                               fail_mode = :zero,
                               dim=dim,
                               num_p_fail_vals = num_p_fail_vals,
                               tolerance=1e-5,
                               num_tests_per = num_tests_per,
                               prop_init = .5,
                               init_type = :SM,
                               alg_type = :PL_graph,
                               signal_type = :gauss,
                               seed = seed);

  # Median-truncated Wirtinger flow
  filename = string("po_dim", dim, "_tests", num_tests_per,
                    "_pvals", num_p_fail_vals,
                    "_seed", seed,
                    "_mtw_lowtol_gauss_smallinit.txt");
  PE.PhaseTransitionExperiment(filename,
                               savepath=savepath,
                               fail_mode = :zero,
                               dim=dim,
                               num_p_fail_vals = num_p_fail_vals,
                               tolerance=1e-5,
                               num_tests_per = num_tests_per,
                               prop_init = .5,
                               init_type = :SM,
                               alg_type = :MTW,
                               signal_type = :gauss,
                               seed = seed);

  # Proximal-linear algorithm
  filename = string("po_dim", dim, "_tests", num_tests_per,
                    "_pvals", num_p_fail_vals,
                    "_seed", seed,
                    "_proxgraph_lowtol_gauss_biginit.txt");
  PE.PhaseTransitionExperiment(filename,
                               savepath=savepath,
                               fail_mode = :zero,
                               dim=dim,
                               num_p_fail_vals = num_p_fail_vals,
                               tolerance=1e-5,
                               num_tests_per = num_tests_per,
                               prop_init = .4,
                               init_type = :LM,
                               alg_type = :PL_graph,
                               signal_type = :gauss,
                               seed = seed);

  # Median-truncated Wirtinger flow
  filename = string("po_dim", dim, "_tests", num_tests_per,
                    "_pvals", num_p_fail_vals,
                    "_seed", seed,
                    "_mtw_lowtol_gauss_biginit.txt");
  PE.PhaseTransitionExperiment(filename,
                               savepath=savepath,
                               fail_mode = :zero,
                               dim=dim,
                               num_p_fail_vals = num_p_fail_vals,
                               tolerance=1e-5,
                               num_tests_per = num_tests_per,
                               prop_init = .4,
                               init_type = :LM,
                               alg_type = :MTW,
                               signal_type = :gauss,
                               seed = seed);

  # Prox-linear
  filename = string("po_dim", dim, "_tests", num_tests_per,
                    "_pvals", num_p_fail_vals,
                    "_seed", seed,
                    "_proxgraph_lowtol_gauss_mtwinit.txt");
  PE.PhaseTransitionExperiment(filename,
                               savepath=savepath,
                               dim=dim,
                               fail_mode = :zero,
                               num_p_fail_vals = num_p_fail_vals,
                               tolerance=1e-5,
                               num_tests_per = num_tests_per,
                               prop_init = .5,
                               init_type = :MTW,
                               alg_type = :PL_graph,
                               signal_type = :gauss,
                               seed = seed);

  # Median-truncated Wirtinger flow
  filename = string("po_dim", dim, "_tests", num_tests_per,
                    "_pvals", num_p_fail_vals,
                    "_seed", seed,
                    "_mtw_lowtol_gauss_mtwinit.txt");
  PE.PhaseTransitionExperiment(filename,
                               savepath=savepath,
                               dim=dim,
                               fail_mode = :zero,
                               num_p_fail_vals = num_p_fail_vals,
                               tolerance=1e-5,
                               num_tests_per = num_tests_per,
                               prop_init = .5,
                               init_type = :MTW,
                               alg_type = :MTW,
                               signal_type = :gauss,
                               seed = seed);
end  # run_outliers

###########################################################
########### ---- Part 3: Digit Experiments ---- ###########
###########################################################
include("digit-experiment.jl");
p_fail_vals = [];
num_successes = [];
num_cgs = [];
number_digit_experiments = 500;
digit_measurement_oversampling = 4;

if (run_digits)
  data_path = "./Data/";
  (p_fail_vals, num_successes, num_cgs) =
    DigitExperiment.ExplorePFailures(number_digit_experiments,
                                     digit_measurement_oversampling);
end

###########################################################
########### ---- Part 4: Image Experiments ---- ###########
###########################################################

# Only do the RNA experiment, with the jpg figure included, available for free
# use at https://visualsonline.cancer.gov/details.cfm?imageid=11167

include("image-processing.jl");
include("Hadamard.jl");
inds = [7];  # Make sure to only do the RNA part.
data_path = "./Data/";
savefile = string(data_path, "rna-particles.result");

# First, convert the image to float format we use
log_dim = 22;  # The log-dimension of the full problem. If log_dim =
               # 20, the images are 512-by-512. Set larger for more
               # accuracy, though the runtime increases.
IP.ConvertAFewImages(log_dim, inds, path = data_path);

data_struct = readdlm(string(data_path, "rna-particles.float"), ' ');
filename = "rna-particles.result";
vert_len = Int(data_struct[1]);
horiz_len = Int(data_struct[2]);
num_channels = Int(data_struct[3]);
len_image = vert_len * horiz_len * num_channels;
dim = Int(data_struct[4]);
x_star = map((x) -> Float64(x), data_struct[5:end]);

assert(length(x_star) == dim);

# Get randomized version of signal
oversampling_factor = 3;
(b, s_vec) = PRS.HadamardSignedData(x_star, oversampling_factor);
# Now, time to initialize

total_time = 0;
tic();
x_init = PRS.OrthogonalityInitializer(dim, s_vec, b, maxiter=300,
                                      use_big=true, p_val=.75, tol=1e-6);
(x_est, objs, cg_count) =
  PRS.ProxLinearQuadratic(x_init, s_vec, b, tol=1e-4, eps_qp=1e-4,
                          maxiter=40, verbose_level=1,
                          qp_solver = :graph);
(x_est, objs_high_prec, cg_count_high_prec) =
  PRS.ProxLinearQuadratic(x_est, s_vec, b, tol=1e-4, eps_qp=1e-5,
                          maxiter=40, verbose_level=1,
                          qp_solver = :graph);
total_time += toc();

rel_error = (min(norm(x_est[1:len_image] + x_star[1:len_image]),
                 norm(x_est[1:len_image] - x_star[1:len_image])) /
             norm(x_star));
println("** Relative solution error: ",
        rel_error,
        " (after ", cg_count, " CG iterations over ", length(objs),
        " iterations) **");

nn = oversampling_factor * dim;
signed_b = sqrt.(max.(b, 0));
d_work_vec = zeros(dim);
n_work_vec = zeros(nn);
Ax = zeros(nn);
D = ones(nn);
PRS.MultiplyByDHS!(x_est, s_vec, D, d_work_vec, Ax);
signed_b = sign(Ax) .* signed_b;
ATb = zeros(dim);
PRS.MultiplyByDHST!(signed_b, s_vec, D, n_work_vec, ATb);
rho = 1.0;
(x_smoothed, iter) = PRS.CGS.InvertHadamardSystem(dim, s_vec, rho, D,
                                                  ATb,
                                                  x_init = x_est,
                                                  eps_cg=1e-10,
                                                  alpha = 0.0);
rel_smoothed_err = (min(norm(x_smoothed[1:len_image] - x_star[1:len_image]),
                        norm(x_smoothed[1:len_image] + x_star[1:len_image])) /
                    norm(x_star[1:len_image]));
println("** Relative smoothed out error: ",
        rel_smoothed_err,
        " (after ", iter, " CG iterations) **");

# Now, time to calibrate timing and save our information and success.
v = randn(dim);
Hadamard.hadmult!(v);
tic();
for ii = 1:100
  Hadamard.hadmult!(v);
end
calibration_time = toc();
calibration_time /= 100;

total_conjugate_steps = cg_count + cg_count_high_prec;
relative_mat_mult_count = total_time / (oversampling_factor * calibration_time);
println("** Relative multiplication count = ", relative_mat_mult_count,
        " -- CG steps = ", total_conjugate_steps, " **");

# Save the results of the experiment in the given file
result = vcat(rel_error,
              rel_smoothed_err,
              cg_count,
              cg_count_high_prec,
              total_time,
              calibration_time,
              oversampling_factor,
              vert_len,
              horiz_len,
              num_channels,
              dim,
              x_init[:],
              x_est[:],
              x_smoothed[:]);

writedlm(savefile, result, ' ');

