# gradient-solvers.jl
#
# Performs gradient descent-like iterations for various phase
# retrieval problems.

# (x, objs) = TruncatedAmpFlow(A::Matrix{Float64}, b::Vector{Float64};
#                              x_init::Vector{Float64} = randn(size(A, 2)),
#                              maxiter::Int64 = length(b),
#                              verbose_level::Int64 = 1,
#                              gamma_thresh::Float64 = .7)
#
# Performs Wang, Giannakis, and Eldar's Trunacted Amplitude Flow for
# the given number of iterations, returning the estimated signal and a
# sequence of objective values.
function TruncatedAmpFlow(A::Matrix{Float64}, b::Vector{Float64};
                          x_init::Vector{Float64} = randn(size(A, 2)),
                          maxiter::Int64 = length(b),
                          verbose_level::Int64 = 1,
                          gamma_thresh::Float64 = .7)
  stepsize = .6;
  x = x_init;
  (nn, dd) = size(A);
  objs = zeros(maxiter+1);
  psi = sqrt(b);  # The magnitudes of signals rather than squares
  objs[1] = sum(psi - abs(A * x)).^2 / (2 * nn);
  D = ones(nn);
  Ax = zeros(nn);
  grad = zeros(dd);
  for iter = 1:maxiter
    # inds = (abs(A * x) .> (psi / (1 + gamma_thresh)));
    # Ax = A[inds, :] * x;
    # grad = A[inds, :]' * (Ax - psi[inds] .* (Ax ./ abs(Ax))) / nn;
    # The below code allows for faster iterations.
    D[(abs(A * x) .<= (psi / (1 + gamma_thresh)))] = 0;
    Ax = A * x;
    grad = A' * (D .* (Ax - psi .* (Ax ./ abs(Ax)))) / nn;
    D[:] = 1;
    x = x - stepsize * grad;
    if (verbose_level >= 1)
      println("[Iteration ", iter, "] Obj = ", objs[iter],
              " Grad norm = ", norm(grad));
    end
    objs[iter + 1] = sum(psi - abs(A * x)).^2 / (2 * nn);
  end
  return (x, objs);
end

# (x, objs) = TruncatedAmpFlow(x_init::Vector{Float64} = randn(dd),
#                              s_vec::Vector{Float64}, b::Vector{Float64};
#                              maxiter::Int64 = length(b),
#                              verbose_level::Int64 = 1,
#                              gamma_thresh::Float64 = .7)
#
# Performs Wang, Giannakis, and Eldar's Trunacted Amplitude Flow for the given
# number of iterations but using our Hadamard sensing matrices, returning the
# estimated signal and a sequence of objective values.
function TruncatedAmpFlow(x_init::Vector{Float64},
                          s_vec::Vector{Float64}, b::Vector{Float64};
                          measure_inds::Vector{Int64} = Array{Int64}(0),
                          maxiter::Int64 = length(b),
                          verbose_level::Int64 = 1,
                          stepsize::Float64 = .5 * length(x_init),
                          gamma_thresh::Float64 = .7,
                          tol::Float64 = 1e-10)
  x = copy(x_init);
  objs = zeros(maxiter);
  dd = length(x);
  x_best = zeros(dd);
  nn = length(s_vec);
  kk = convert(Int64, nn/dd);
  s_vec = sign(s_vec) / sqrt(kk);
  min_obj = Inf;
  if (isempty(measure_inds))
    measure_inds = 1:length(s_vec);
  end
  if (2^Hadamard.intlog(dd) != dd)
    error("Dimension must be power of 2.");
  end
  objs = zeros(maxiter);
  d_work_vec = zeros(dd);
  n_work_vec = zeros(nn);
  onevec = ones(nn);
  Ax = zeros(nn);
  grad = zeros(dd);
  b_vec = copy(b);
  b_vec[b_vec .< 0] = 0;
  R_vec = zeros(nn);
  R_vec[measure_inds] = 1;
  R_boolean = zeros(Bool, nn);
  R_boolean[measure_inds] = true;
  b_vec = copy(b);
  b_vec[!R_boolean] = 0;
  psi = sqrt(b_vec);  # The magnitudes of signals rather than squares
  for iter = 1:maxiter
    MultiplyByDHS!(x, s_vec, R_vec, d_work_vec, Ax);
    objs[iter] = sum((psi - abs(Ax)).^2) / (2 * nn);
    inds = (abs(Ax) .> (psi / (1 + gamma_thresh)));
    # grad = A[inds, :]' * (Ax[inds] - psi[inds] .* (Ax[inds] ./
    #            abs(Ax[inds]))) / nn.
    # Use onevec as the index vector
    onevec[!inds] = 0.0;
    MultiplyByDHST!((Ax - psi .* Ax ./ abs(Ax)), s_vec, onevec,
                    n_work_vec, grad);
    grad /= nn;
    x = x - stepsize * grad;
    if (verbose_level >= 1)
      println("[Iteration ", iter, "] Obj = ", objs[iter],
              " Grad norm = ", norm(grad));
    end
    onevec[:] = 1.0;
    if (objs[iter] < tol)
      objs = objs[1:iter];
      return (x, objs);
    end
  end
  return (x, objs);
end

# GradientDescentQ(A::Matrix{Float64}, b::Vector{Float64};
#                          x_init::Vector{Float64} = randn(size(A, 2)),
#                          maxiter::Int64 = length(b),
#                          verbose_level::Int64 = 1,
#                          tol::Float64 = 1e-9)
#
# Runs gradient descent on the quadratic phase retrieval objective
function GradientDescentQ(A::Matrix{Float64}, b::Vector{Float64};
                          x_init::Vector{Float64} = randn(size(A, 2)),
                          maxiter::Int64 = length(b),
                          verbose_level::Int64 = 1,
                          tol::Float64 = 1e-9)
  stepsize = .6;
  proj_dec_mult = .4;  # Projected decrease from gradient
  step_mult = .5;  # Amount to decrease stepsize if not enough progress
  x = x_init;
  (mm, nn) = size(A);
  objs = zeros(maxiter + 1);
  grad = zeros(nn);
  Ax = A * x;
  Ag = zeros(mm);
  objs[1] = mean(((Ax).^2 - b).^2) / 2;
  x_next = zeros(nn);
  for iter = 1:maxiter
    Ax = A * x;
    grad = 2 * A' * (Ax .* ((Ax).^2 - b)) / mm;
    grad_norm_sq = sum(grad.^2);
    Ag = A * grad;
    next_obj = mean(((Ax - stepsize * Ag).^2 - b).^2) / 2;
    while (next_obj > objs[iter] - stepsize * proj_dec_mult * grad_norm_sq)
      stepsize *= step_mult;
      next_obj = mean(((Ax - stepsize * Ag).^2 - b).^2) / 2;
    end
    x = x - stepsize * grad;
    stepsize /= step_mult;
    objs[iter + 1] = next_obj;
    if (objs[iter] < tol)
      objs = objs[1:iter];
      return (x, objs);
    end
  end
  return (x, objs);
end

# (x, objs) = MedianTruncatedWFlow(A::Matrix{Float64}, b::Vector{Float64};
#                                  x_init::Vector{Float64} = randn(size(A, 2)),
#                                  maxiter::Int64 = length(b),
#                                  verbose_level::Int64 = 1,
#                                  u_thresh::Float64 = 5.0,
#                                  h_thresh::Float64 = 12.0,
#                                  l_thresh::Float64 = .3)
#
# Performs Zhang, Chi, and Liang's "Median Truncated Wirtinger Flow" for the
# given number of iterations, returning the estimated signal and sequence of
# objective values.
function MedianTruncatedWFlow(A::Matrix{Float64}, b::Vector{Float64};
                              x_init::Vector{Float64} = randn(size(A, 2)),
                              maxiter::Int64 = length(b),
                              verbose_level::Int64 = 1,
                              u_thresh::Float64 = 5.0,
                              h_thresh::Float64 = 12.0,
                              l_thresh::Float64 = .3)
  stepsize = .2;
  x = x_init;
  (nn, dd) = size(A);
  objs = zeros(maxiter + 1);
  objs[1] = norm(b - (A * x).^2, 1) / nn;
  for iter = 1:maxiter
    # TODO(jduchi): pre-allocate these for faster iterations.
    Ax = A * x;
    absAx = abs(Ax);
    Ax_squared = (Ax).^2;
    K = median(abs(b - Ax_squared));
    inds = ((absAx .>= l_thresh * norm(x)) &
            (absAx .<= u_thresh * norm(x)) &
            (abs(b - Ax_squared) .<= h_thresh * K * absAx / norm(x)));
    grad = A[inds, :]' * ((Ax_squared[inds] - b[inds]) ./ Ax[inds]) / nn;
    x = x - stepsize * grad;
    if (verbose_level >= 1)
      println("[Iteration ", iter, "] Obj = ", objs[iter],
              " Grad norm = ", norm(grad));
    end
    objs[iter + 1] = norm(b - (A * x).^2, 1) / nn;
  end
  return (x, objs);
end

# MedianTruncatedWFlow(x_init::Vector{Float64}, s_vec::Vector{Float64},
#                               b::Vector{Float64};
#                               maxiter::Int64 = length(s_vec),
#                               verbose_level::Int64 = 1,
#                               u_thresh::Float64 = 5.0,
#                               good_err_thresh::Float64 = .8,
#                               l_thresh::Float64 = .3)
function MedianTruncatedWFlow(x_init::Vector{Float64}, s_vec::Vector{Float64},
                              b::Vector{Float64};
                              maxiter::Int64 = length(s_vec),
                              verbose_level::Int64 = 1,
                              u_thresh::Float64 = 5.0,
                              l_thresh::Float64 = .3,
                              good_err_thresh::Float64 = .8,
                              stepsize::Float64 = 1.0 * length(x_init))
  x = copy(x_init);
  dd = length(x);
  nn = length(s_vec);
  kk = convert(Int64, nn/dd);
  s_vec = sign(s_vec) / sqrt(kk);
  if (2^Hadamard.intlog(dd) != dd)
    error("Dimension must be power of 2.");
  end
  objs = zeros(maxiter);
  d_work_vec = zeros(dd);
  onevec = ones(nn);
  Ax = zeros(nn);
  Ax_squared = zeros(nn);
  grad = zeros(dd);
  absAx = zeros(nn);
  num_relative_error_inds =
    max(1, min(floor(Int64, good_err_thresh * nn), nn));
  for iter = 1:maxiter
    MultiplyByDHS!(x, s_vec, onevec, d_work_vec, Ax);
    absAx = abs(Ax);
    Ax_squared = absAx.^2;
    K = median(abs(b - Ax_squared));
    inds = ((absAx .>= l_thresh * norm(x) / sqrt(dd)) &
            (absAx .<= u_thresh * norm(x) / sqrt(dd)));
    # Keep the smaller relative error indices
    inds_by_relative_error = sortperm(abs(b - Ax_squared) * norm(x) ./ absAx,
                                      lt = <);
    inds[inds_by_relative_error[(num_relative_error_inds+1):end]] = 0;
    if (verbose_level >= 1)
      println("\t[Iter ", iter, "] num indices = ", sum(inds));
    end
    # grad = (D * A)' * ((Ax_squared[inds] - b[inds]) ./ Ax[inds]) / nn,
    # where D is a diagonal zero-one matrix indicating the non-zeros in inds.
    onevec[!inds] = 0;  # Make onevec the indicators
    # Use absAx as the work vector.
    MultiplyByDHST!((Ax_squared - b) ./ Ax, s_vec, onevec, absAx, grad);
    grad /= nn;
    x = x - stepsize * grad;
    onevec[:] = 1.0;
    objs[iter] = norm(b - Ax_squared, 1) / nn;
    if (verbose_level >= 1)
      println("[Iteration ", iter, "] Obj = ", objs[iter],
              " Grad norm = ", norm(grad));
    end
  end
  return (x, objs);
end
                              
function huber_loss(x::Vector{Float64},
                    y::Vector{Float64},
                    threshold::Float64)
  sum = 0.0;
  nn = length(x);
  mult = 0.0;
  diff = 0.0;
  @inbounds for ii = 1:nn
    absdiff = abs(x[ii] - y[ii]);
    mult = 1.0 * (absdiff < threshold);
    sum += mult * (absdiff)^2 / (2 * threshold);
    sum += (1 - mult) * (absdiff - threshold / 2);
  end
  return sum;
end

function huber_grad!(x::Vector{Float64},
                     y::Vector{Float64},
                     threshold::Float64,
                     g::Vector{Float64})
  nn = length(x);
  @inbounds for ii = 1:nn
    diff = x[ii] - y[ii];
    mult = 1.0 * (abs(diff) < threshold);
    g[ii] = mult * diff / threshold;
    g[ii] += (1 - mult) * sign(diff);
  end
end

# (x, iters) = SolveQP_Huber(A::Matrix{Float64}, c_vec::Vector{Float64},
#                            x_zero::Vector{Float64};
#                            eps_stationarity::Float64 = 1e-5,
#                            verbose_level::Int64 = 0,
#                            MAXITER::Int64 = 1000,
#                            smooth_amount::Float64 = eps_stationarity)
#
# Sets x to approximately minimize
#
#  huber(A * x - b, u) + .5 * (x - x_zero) * A' * A * (x - x_zero),
#
# where huber(v, u) is defined as the sum of
#
#  v[i]^2 / (2 * u)   if abs(v[i]) < u
#  abs(v[i]) - u/2    if abs(v[i]) >= u
#
# Performs gradient descent with backtracking line search to solve the problem.
function SolveQP_Huber(A::Matrix{Float64}, c_vec::Vector{Float64},
                       x_zero::Vector{Float64};
                       eps_stationarity::Float64 = 1e-5,
                       verbose_level::Int64 = 0,
                       MAXITER::Int64 = 1000,
                       smooth_amount::Float64 = eps_stationarity)
  (mm, nn) = size(A);
  Ax = zeros(mm);
  Ax_minus_zero = zeros(mm);
  AtAx = zeros(nn);
  AtA = A' * A;
  AtAx_zero = AtA * x_zero;
  grad = zeros(nn);
  m_work_vec = zeros(mm);
  n_work_vec = zeros(nn);
  x = copy(x_zero);
  x_next = zeros(nn);
  grad_norm = Inf;
  stepsize = 1.0;
  alpha_bt = .2;
  beta_bt = .5;
  obj = 0.0;
  gemv!('n', 1.0, A, x, 0.0, Ax);
  gemv!('n', 1.0, AtA, x, 0.0, AtAx);
  blascopy!(nn, x, 1, n_work_vec, 1);
  axpy!(-1.0, x_zero, n_work_vec);
  gemv!('n', 1.0, A, n_work_vec, 0.0, Ax_minus_zero);
  blascopy!(nn, x, 1, n_work_vec, 1);
  axpy!(-1.0, x_zero, n_work_vec);
  gemv!('n', 1.0, A, n_work_vec, 0.0, Ax_minus_zero);
  obj = (huber_loss(Ax, c_vec, smooth_amount) / mm +
         .5 * dot(Ax_minus_zero, Ax_minus_zero) / mm);
  iter = 0;
  while (grad_norm > sqrt(nn) * eps_stationarity
         && iter < MAXITER)
    if (verbose_level >= 2)
      println("[Iter ", iter, "] Obj: ", obj, "\tGrad: ", grad_norm);
    end
    # First, compute gradients, which are...
    # Ax = A * x; AtAx = A' * A * x
    iter += 1;
    huber_grad!(Ax, c_vec, smooth_amount, m_work_vec);
    gemv!('t', 1.0, A, m_work_vec, 0.0, grad);
    axpy!(1.0, AtAx, grad);
    axpy!(-1.0, AtAx_zero, grad);
    grad[:] /= mm;
    grad_norm = norm(grad);
    # Now, do a backtracking line search until we have made enough progress
    # x_next = x - stepsize * g;
    blascopy!(nn, x, 1, x_next, 1);
    axpy!(-stepsize, grad, x_next);

    # Now, backtracking line search, but have to compute objectives
    # Ax = A * x_next;
    # Ax_minus_zero = A * (x_next - x_zero);
    gemv!('n', 1.0, A, x_next, 0.0, Ax);
    blascopy!(nn, x_next, 1, n_work_vec, 1);
    axpy!(-1.0, x_zero, n_work_vec);
    gemv!('n', 1.0, A, n_work_vec, 0.0, Ax_minus_zero);
    obj_next = (huber_loss(Ax, c_vec, smooth_amount) / mm
                + .5 * dot(Ax_minus_zero, Ax_minus_zero) / mm);
    while (obj_next > obj - stepsize * alpha_bt * grad_norm^2)
      stepsize *= beta_bt;
      # x_next = x - stepsize * grad;
      # Ax = A * x_next;
      # Ax_minus_zero = A * (x_next - x_zero);
      # obj_next = (huber_loss(Ax, c_vec, smooth_amount) / mm
      #             + .5 * dot(Ax_minus_zero, Ax_minus_zero) / mm);
      
      blascopy!(nn, x, 1, x_next, 1);
      axpy!(-stepsize, grad, x_next);
      gemv!('n', 1.0, A, x_next, 0.0, Ax);
      blascopy!(nn, x_next, 1, n_work_vec, 1);
      axpy!(-1.0, x_zero, n_work_vec);
      gemv!('n', 1.0, A, n_work_vec, 0.0, Ax_minus_zero);
      obj_next = (huber_loss(Ax, c_vec, smooth_amount) / mm
                  + .5 * dot(Ax_minus_zero, Ax_minus_zero) / mm);
    end
    obj = obj_next;
    blascopy!(nn, x_next, 1, x, 1);
    stepsize /= beta_bt;
    gemv!('n', 1.0, AtA, x_next, 0.0, AtAx);
  end
  return (x, iter);
end
