# hadamard-solvers.jl
#
# Base solvers for the Hadamard sampling methods.

# (x, objs, cg_count) =
#   ProxLinearQuadratic(x, s, b;
#                       maxiter, measure_inds, verbose_level, tol, qp_solver, eps_qp)
#
# Uses SolveQP_admm or SolveQP_graph, the ADMM or splitting-based method for
# solving L1 + quadratic objectives, to perform the prox-linear method on the
# objective
#
#  min. || (R H S * x).^2 - b ||_1
#
# where x is a vector in R^d, the matrix S is n-by-d and of the form
#
#  S' = [S1 S2 ... Sk]
#
# where Si is the diagonal of the i-th block of the vector s (here k =
# n/d). The matrix H is the Kronecker product of I_k and the d-by-d normalized
# Hadamard matrix. The matrix R is a subsampling matrix of observed rows of H
# S, indicated by measure_inds, which is a vector of the same length as b
# indicating which indices of HS*x to include in the measurements.
#
# A copy of the vector b is internally modified so that all indices
# corresponding to *non-measurements* (as indicated by measure_inds or the R
# matrix) are zero, so they incur no loss.
#
# The returned parameters are the final estimated x, the objectives of the
# sequential iterations, and the count of the number of conjugate gradient
# iterations.
#
# The Symbol qp_solver should be one of :newton, :admm, and :graph, which use
# different QP solvers.
function ProxLinearQuadratic(x_init::Vector{Float64},
                             s_vec::Vector{Float64},
                             b::Vector{Float64};
                             measure_inds::Vector{Int64} = Array{Int64}(0),
                             maxiter::Int64 = length(s_vec),
                             verbose_level::Int64 = 1,
                             tol::Float64 = 1e-3,
                             qp_solver::Symbol = :graph,
                             eps_qp::Float64 = 1e-4)
  x_curr = copy(x_init);
  dd = length(x_curr);
  nn = length(s_vec);
  kk = ceil(Int64, nn/dd);
  s_vec = sign.(s_vec) / sqrt(kk);
  if (2^Hadamard.intlog(dd) != dd)
    error("Dimension must be power of 2.");
  end
  R_vec = zeros(nn);
  if (isempty(measure_inds))
    measure_inds = 1:length(s_vec);
  end
  R_vec[measure_inds] = 1;
  R_boolean = zeros(Bool, nn);
  R_boolean[measure_inds] = true;
  b_vec = copy(b);
  b_vec[.!R_boolean] = 0;
  iter = 1;
  grad_map_norm = Inf;
  objs = zeros(maxiter+1);
  Wx = zeros(nn);
  x_delta = zeros(dd);
  d_work_vec = zeros(dd);
  total_cg_calcs = 0;
  while (iter <= maxiter && grad_map_norm > tol)
    # First, compute objective
    # obj = norm((A * x).^2 - b)
    MultiplyByDHS!(x_curr, s_vec, R_vec, d_work_vec, Wx);
    obj = norm((Wx).^2 - b_vec, 1);
    objs[iter] = obj;
    if (verbose_level >= 1)
      println("[Iteration ", iter, "]  Objective: ",
              obj, "    step: ", grad_map_norm,
              "   Total CG steps: ", total_cg_calcs);
    end
    new_obj = obj + 1.0;
    # Now, time to compute derivative matrices, etc.
    D = 2 * Wx;
    c_vec = b_vec - Wx.^2;
    # Set x_delta to minimize
    #
    # norm(diagm(D) H * S * x - c_vec, 1) + norm(x)^2
    if (qp_solver == :admm || qp_solver == :graph)
      if (qp_solver == :admm)
        (x_delta, cg_iters) = SolveQP_admm(dd, s_vec, D/2, c_vec/2,
                                           x_init = zeros(dd), rho=.125,
                                           verbose_level=verbose_level,
                                           eps_abs = eps_qp,
                                           eps_rel = eps_qp);
        total_cg_calcs += 2 * cg_iters;
      else
        (x_delta, cg_iters) = SolveQP_graph_form(dd, s_vec, D/2, c_vec/2,
                                                 x_init = zeros(dd),
                                                 rho = 2.,
                                                 eps_abs = eps_qp,
                                                 eps_rel = eps_qp);
        total_cg_calcs += 2 * cg_iters;
      end
      stepsize = 1.0;
      x_delta_norm = norm(x_delta)
      while (new_obj > obj &&
             x_delta_norm * stepsize > tol / 2)  # Small backtracking search
        x_intermediate = x_curr + stepsize * x_delta;
        MultiplyByDHS!(x_intermediate, s_vec, R_vec, d_work_vec, Wx);
        new_obj = norm(Wx.^2 - b_vec, 1);
        stepsize *= .5;
        total_cg_calcs += 1;
      end
      x_delta *= stepsize;
    elseif (qp_solver == :newton)
      (x_delta, cg_iters) = SolveQP_newton(dd, s_vec, D/2, c_vec/2,
                                           newton_acc = eps_qp,
                                           verbose_level = verbose_level);
      total_cg_calcs += cg_iters;
    else
      error("Unknown QP solver ", qp_solver);
    end
    grad_map = 2 * x_delta;
    x_curr = x_curr + x_delta;
    grad_map_norm = norm(grad_map);
    iter = iter + 1;
  end
  MultiplyByDHS!(x_curr, s_vec, R_vec, d_work_vec, Wx);
  obj = norm((Wx).^2 - b_vec, 1);
  objs[iter] = obj;
  objs = objs[1:iter];
  if (verbose_level >= 1)
    println("[Iteration ", iter, "]  Objective: ",
            objs[end], "    step: ", grad_map_norm,
            "   Total CG steps: ", total_cg_calcs);
  end
  return (x_curr, objs, total_cg_calcs);
end

# x = OrthogonalityInitializer(d, s_vec, b;
#                              maxiter, p_val, verbose_level, tol)
#
# Performs orthogonality-promoting initialization of an image based on
# our Hadamard sampling scheme.
function OrthogonalityInitializer(dd::Int64,
                                  s_vec::Vector{Float64},
                                  b::Vector{Float64};
                                  maxiter::Int64 = 50,
                                  p_val::Float64 = .5,
                                  verbose_level::Int64 = 1,
                                  tol::Float64 = 1e-5,
                                  use_big::Bool = false)
  dir = randn(dd);
  dir = dir / norm(dir);
  nn = length(s_vec);
  kk = convert(Int64, nn / dd);
  b_med = quantile(b, p_val);
  D = float(b .<= b_med);
  if (use_big)
    D = float(b .> b_med);
  end
  if (sum(D) <= 1)
    error("Not enough indices in given quantile");
  end
  n_work_vec = zeros(nn);
  d_work_vec = zeros(dd);
  DAdir = zeros(nn);
  dir_change = zeros(dd);
  iter = 0;
  norm_change = Inf;
  while (iter < maxiter && norm_change > tol)
    iter += 1;
    dir_change[:] = dir;
    MultiplyByDHS!(dir, s_vec, D, d_work_vec, DAdir);
    MultiplyByDHST!(DAdir, s_vec, D, n_work_vec, d_work_vec);
    if (use_big)
      dir = d_work_vec;
    else
      dir = dir - d_work_vec;
    end
    dir = dir / norm(dir);
    dir_change -= dir;
    norm_change = norm(dir_change);
    if (verbose_level >= 2)
      println("[Iteration ", iter, "]: change = ", norm_change);
    end
  end
  # Now, solve simple median-like problem to find magnitude of the direction.
  MultiplyByDHS!(dir, s_vec, ones(nn), d_work_vec, n_work_vec);
  n_work_vec = n_work_vec.^2;
  # (see file robust-quadatics.jl)
  t_val = MinimizeOneDAbs(n_work_vec, b .* (b .>= 0));
  return dir * sqrt(t_val);
end
