# image-processing.jl
#
# A few methods for processing images and loading and saving them

module IP

using PyPlot;
include("Hadamard.jl");

# ConvertAFewImages(log_dim = 20)
#
# Takes 5 images and converts them to the Float64 format acceptable for use on
# Sherlock (which does not have image processing stuff)
function ConvertAFewImages(log_dim::Int64 = 20,
                           inds::Array{Int64} = Array{Int64}(0);
                           path::String = "")
  if (isempty(path))
    path = string(homedir(), "/Data/CompositeImages/");
  end
  files = ["blue-poison.png",  # [1]
           "owl.png", # [2]
           "killer-whale.png", # [3]
           "hedgehog.png", # [4]
           "annie.png", # [5]
           "glacier-crevasse.png", # [6]
           "rna-particles.png", #[7]
           "galaxy.png", #[8]
           ];
  if (isempty(inds))
    for ii = 1:length(files)
      println("Converting image ", path, files[ii]);
      ConvertImageToFloats(string(path, files[ii]), 1 << log_dim);
    end
  else
    for ii in inds
      println("Converting image ", path, files[ii]);
      ConvertImageToFloats(string(path, files[ii]), 1 << log_dim);
    end
  end
  # [1] https://aqua.org/explore/animals/blue-poison-dart-frog
  # [2] http://wdfw.wa.gov/living/species/graphics/owl_3.jpg
  # [3] http://www.bbc.com/news/magazine-37025092
  # [4] http://kids.nationalgeographic.com/animals/hedgehog/
  # [5] Home
  # [6] https://photolibrary.usap.gov/ Julian Race, NSF
  #     (goo.gl/wUqLqx)
  # [7] https://visualsonline.cancer.gov/details.cfm?imageid=11167
  # [8]
  # https://www.nasa.gov/image-feature/goddard/2017/hubble-hones-in-on-a-hypergiants-home
  # 5, 6, 7, 8 are definitely free domain.
end

# LoadImageFromFloats(filename::String; savefile::String = "")
#
# Loads a file stored as a .float file from the given filename, then writes
# it out in the different image format specified by savefile (also plots it).
function LoadImageFromFloats(filename::String; savefile::String = "")
  float_image = readdlm(filename, ' ');
  vert_len = convert(Int64, float_image[1]);
  horiz_len = convert(Int64, float_image[2]);
  num_channel = convert(Int64, float_image[3]);
  dim = convert(Int64, float_image[4]);
  x_star = map(x -> float(x), float_image[4:end]);
  len_image = vert_len * horiz_len * num_channel;
  image = reshape(x_star[1:len_image], vert_len, horiz_len, num_channel);
  figure();
  imshow(image);
  xticks([]); yticks([]);
  if (!isempty(savefile))
    savefig(savefile, bbox_inches="tight");
  end
end

# ConvertImageToFloats(filename::String, dim::Int64;
#                      savefile::String ="")
#
# Converts an image of type .png stored at the given filename into an image of
# floats, where the product of the height, width, and number of channels is the
# largest power of 2 less than or equal to dim. If dim is large enough, simply
# saves the whole image.
#
# The resulting numbers are stored in the given save file, which defaults to
# the original filename with its extension replaced by .float
function ConvertImageToFloats(filename::String, dim::Int64;
                              savefile::String ="")
  picture = imread(filename);
  if (size(picture, 3) >= 3)
    picture = picture[:, :, 1:3];
  end
  kk = 3;
  N = prod(size(picture));
  stride_size = sqrt(N / dim);
  if (stride_size < 1)
    stride_size = 1;
  end
  horiz_len = size(picture, 2);
  vert_len = size(picture, 1);
  h_inds = map((x) -> round(Int64, x),
               linspace(1, horiz_len, floor(horiz_len / stride_size)));
  v_inds = map((x) -> round(Int64, x),
               linspace(1, vert_len, floor(vert_len / stride_size)));
  sub_image = picture[v_inds, h_inds, :];
  len_sub_image = length(sub_image[:]);
  if (len_sub_image == N)
    # We did not shrink the size of the image at all
    log_size = Hadamard.intlog(len_sub_image);
    old_dim = dim;
    if (1 << log_size == len_sub_image)
      dim = len_sub_image;
    else
      dim = (1 << (log_size + 1));
    end
    println("[INFO] Image size not shrunk; dimension decreased from ",
            old_dim, " to ", dim);
  end
  x_star = vcat(sub_image[:], zeros(dim - len_sub_image));
  if (isempty(savefile))
    savefile_r = match(r"\.[A-Za-z]+", filename);
    savefile = filename;
    if (savefile_r != nothing)
      savefile = string(filename[1:savefile_r.offset], "float");
    else
      savefile = string(filename, ".float");
    end
  end
  result = vcat(length(v_inds), length(h_inds), size(picture, 3), dim, x_star[:]);
  writedlm(savefile, result, ' ');
end

# (rel_err, mult_normal_count, mult_refine_count, calibration_time, total_time) =
#   PlotResultsOfImageExperiment(filename::String,
#                                flip_image::Bool = false;
#                                init_im_save::String = "",
#                                est_im_save::String = "",
#                                refined_im_save::String = "")
#
# Loads the result of the script "image-from-float.jl", which is a dlm vector,
# and plots resulting images.
#
# If the files init_im_save, est_im_save, or refined_im_save are specified,
# saves the image corresponding to the initialization, estimated image, or
# refined estimated image to the given file.
#
# Returns four numbers: the relative error of the image, the count of matrix
# multiplies in the initial solution stage, the count of matrix multiplies in
# the refining stage, the time required to compute one matrix multiply, and the
# total time of execution, which allows computation of the "true" times.
function PlotResultsOfImageExperiment(filename::String,
                                      flip_image::Bool = false;
                                      init_im_save::String = "",
                                      est_im_save::String = "",
                                      refined_im_save::String = "")
  result = readdlm(filename, ' ');
  rel_error = float(result[1]);
  rel_smoothed_err = float(result[2]);
  cg_count = Int(result[3]);
  cg_count_high_prec = Int(result[4]);
  total_time = float(result[5]);
  calibration_time = float(result[6]);
  kk = Int(result[7]);
  vert_len = Int(result[8]);
  horiz_len = Int(result[9]);
  num_channels = Int(result[10]);
  dim = Int(result[11]);
  last_ordinal_ind = 11;
  x_init = map(x -> float(x),
              result[(last_ordinal_ind + 1):(last_ordinal_ind + dim)]);
  x_est = map(x -> float(x),
              result[(last_ordinal_ind + dim + 1):(last_ordinal_ind + 2 * dim)]);
  x_refined = map(x -> float(x),
                  result[(last_ordinal_ind + 2 * dim + 1):end]);
  # Reshape them into images and figure out errors
  len_image = vert_len * horiz_len * num_channels;
  x_init = reshape(x_init[1:len_image], vert_len, horiz_len, num_channels);
  x_est = reshape(x_est[1:len_image], vert_len, horiz_len, num_channels);
  x_refined = reshape(x_refined[1:len_image], vert_len, horiz_len, num_channels);
  
  s_val = (flip_image ? -1 : 1);
  figure();
  rgb_image = ceil.(UInt8, 255 * max.(min.(s_val * x_init, 1), 0));
  imshow(rgb_image, interpolation="nearest");
  xticks([]); yticks([]);
  if (!isempty(init_im_save))
    savefig(init_im_save, bbox_inches="tight");
  end
  figure();
  rgb_image = ceil.(UInt8, 255 * max.(min.(s_val * x_est, 1), 0));
  imshow(rgb_image, interpolation="nearest");
  xticks([]); yticks([]);
  if (!isempty(est_im_save))
    savefig(est_im_save, bbox_inches="tight");
  end
  figure();
  rgb_image = ceil.(UInt8, 255 * max.(min.(s_val * x_refined, 1), 0));
  imshow(rgb_image, interpolation="nearest");
  xticks([]); yticks([]);
  if (!isempty(refined_im_save))
    savefig(refined_im_save, bbox_inches="tight");
  end

  total_conjugate_steps = cg_count + cg_count_high_prec;
  relative_mat_mult_count = total_time / (kk * calibration_time);
  println("** CG steps = ", total_conjugate_steps, " **");
  println("** Count of multiplications possible in time: ",
          total_time / calibration_time, " **\n**\t(or ",
          total_time / (kk * calibration_time), " CG steps, efficiency ",
          total_conjugate_steps * (kk * calibration_time) / total_time,
          ") **");
  println("** Relative smoothed out error: ",
          rel_smoothed_err, " **");
  println("** Total time: ", total_time,
          " (", calibration_time, " per hadmult) **");
  return (rel_smoothed_err, kk * cg_count, kk * cg_count_high_prec,
          calibration_time, total_time);
end

end  # module IP
