# admm-lone-solver-tests.jl
#
# Tests for methods in the ADMM L1 solver methods

using Convex;
using Mosek;

function TestSolvePR_newton(;vbl = 0)
  srand(0);
  dd = 32;
  kk = 5;
  num_tests = 5;
  num_failed = 0;
  for test_ind = 1:num_tests
    x_star = randn(dd);
    (b, s_vec) = PRS.HadamardSignedData(x_star, kk);
    (x, cg_iters) =
      ProxLinearQuadratic(x_star + .5 * randn(dd), s_vec, b,
                          verbose_level = vbl,
                          qp_solver = :newton);
    if (norm(x - x_star) / norm(x_star) > 1e-3)
      num_failed += 1;
    end
  end
  if (num_failed > 0)
    warn("Failed ", num_failed, " of ", num_tests,
         " of ProxLinearQuadratic with Newton");
  else
    println("Passed all ProxLinearQuadratic tests with Newton.");
  end
end

# TestSolveQP_newton(; eps_accuracy = 1e-8, vbl = 0)
#
# Tests to see that the SolveQP_newton method is functional and works
# correctly.
function TestSolveQP_newton(; eps_accuracy = 1e-8, vbl = 0)
  srand(0);
  dd = 16;
  kk = 3;
  nn = kk * dd;
  Hd = Hadamard.hadmat(Hadamard.intlog(dd));
  H = kron(eye(kk), Hd) / sqrt(dd);
  num_tests = 5;
  num_failed = 0;
  for test_ind = 1:num_tests
    s_vec = sign(randn(nn)) / sqrt(kk);
    S = diagm(s_vec[1:dd]);
    for ll = 2:kk
      indset = ((ll - 1) * dd + 1):(ll * dd);
      S = vcat(S, diagm(s_vec[indset]));
    end
    D = randn(nn).^3;
    A = diagm(D) * H * S;
    c_vec = randn(nn);
    # Now get CVX objective
    x_cvx = Variable(dd);
    obj = minimize(norm(A * x_cvx - c_vec, 1) + .5 * sumsquares(x_cvx));
    solve!(obj, MosekSolver(QUIET=true));
    (x_newton, total_cgs) =
      SolveQP_newton(dd, s_vec, D, c_vec, verbose_level=vbl, newton_acc = 1e-5,
                     t_mult = 5.0);
    cvx_obj_val = (norm(A * x_cvx.value - c_vec, 1)
                   + .5 * norm(x_cvx.value, 2)^2);
    newton_obj_val = (norm(A * x_newton - c_vec, 1)
                           + .5 * norm(x_newton, 2)^2);
    relative_diff = (newton_obj_val - cvx_obj_val) / cvx_obj_val;
    if (vbl >= 1)
      println("*** Relative difference = ", relative_diff, " *** [CGs: ",
              total_cgs, "]");
    end
    if (relative_diff > 1e-3)
      num_failed += 1;
    end
  end
  if (num_failed > 0)
    warn("Failure of ", num_failed, " of ", num_tests,
         " tests of SolveQP_newton");
  else
    println("Passed all SolveQP_newton tests");
  end
end
