# lone-qp-solver.jl
#
# An interior point method that implements solutions of the problem
#
# min. ||D * H * S * x - c||_1 + norm(x)^2 / 2
#
# where D is diagonal, H is a block Hadamard transform matrix, and S is a
# collection of diagonal blocks.

# (x, cg_iters) = SolveQP_newton(dd::Int64, s_vec::Vector{Float64},
#                                D::Vector{Float64}, c::Vector{Float64};
#                                newton_acc::Float64 = 1e-5,
#                                t_param_max::Float64 = 1e6,
#                                t_mult::Float64 = 10.0,
#                                verbose_level::Int64 = 0)
#
# Solves the optimization problem described above. In particular, uses a
# standard interior point Newton method (with backtracking line search) to
# solve the problem
#
# min. norm(A * x - c, 1) + (1/2) * norm(x)^2.
#
# Here A = DHS for matrices D, H, S of the following forms.
#
# D is a diagonal matrix, otherwise arbitrary
# S is a stacked block of diagoinal matrices,
#
#  S' = [S1 S2 ... Sk]
#
# where Si is a d-by-d diagonal matrix. d should be a power of 2.
#
# H is a block-diagonal Hadamard matrix, where
#
#  H = kron(eye(k), Hd) / sqrt(d)
#
# and Hd is the d-by-d standard {+1, -1}-valued Hadamard matrix,
# defined by the recursion
#
# H1 = [1 1; 1 -1];  H{d+1} = [Hd Hd; Hd -Hd].
function SolveQP_newton(dd::Int64, s_vec::Vector{Float64},
                        D::Vector{Float64}, c::Vector{Float64};
                        newton_acc::Float64 = 1e-5,
                        t_param_max::Float64 = 1e6,
                        t_mult::Float64 = 10.0,
                        verbose_level::Int64 = 0)
  t_val = 1.0;
  nn = length(s_vec);
  x = zeros(dd);
  # First, get initial values of y.
  y = abs(copy(c)) + 1;
  d_work_vec = zeros(dd);
  n_work_vec = zeros(nn);
  Ax = zeros(nn);
  Adx = zeros(nn);
  grad_x = zeros(dd);
  grad_y = zeros(nn);
  grad = zeros(nn + dd);
  full_variables = zeros(nn + dd);
  alpha_mult = .2;
  beta_mult = .5;
  iter = 0;
  t_block_iter = 0;
  total_cgs = 0;
  while (t_val < t_param_max)
    newton_dec = Inf;
    t_block_iter += 1;
    iter = 0;
    PRS.MultiplyByDHS!(x, s_vec, D, d_work_vec, Ax);
    obj = t_val * sum(y) + (t_val/2) * norm(x)^2;
    obj = obj - sum(log(y - c + Ax)) - sum(log(y + c - Ax));
    while (newton_dec > newton_acc)
      iter += 1;
      # Compute gradient of
      #
      # f_t(x, y) = t * sum(y) + (t/2) * norm(x)^2 - sum(log(y - c + D * W * x))
      #              - sum(log(y + c - D * W * x)).
      #
      # Also compute its Hessian and solve the resulting Newton system.
      PRS.MultiplyByDHS!(x, s_vec, D, d_work_vec, Ax);
      diag_pos = (y - c + Ax);
      diag_neg = (y + c - Ax);
      PRS.MultiplyByDHST!(1./diag_pos, s_vec, D, n_work_vec, d_work_vec);
      PRS.MultiplyByDHST!(1./diag_neg, s_vec, D, n_work_vec, grad_x);
      grad_x = t_val * x + grad_x - d_work_vec;
      grad_y = t_val - 1 ./ diag_pos - 1 ./ diag_neg;
      grad[1:dd] = grad_x;
      grad[(dd+1):end] = grad_y;
      # Solve Newton system associated with interior point problem.
      Dy = 1 ./ diag_pos.^2 + 1 ./ diag_neg.^2;
      Dxy = 1 ./ diag_pos.^2 - 1 ./ diag_neg.^2;
      D1 = t_val + D .* Dy .* D;
      D2 = D .* Dxy;
      full_variables[1:dd] = x;
      full_variables[(dd+1):end] = y;
      (d_newton, cg_iters) =
        CGS.InvertNewtonSystem(nn, dd, s_vec, D1, D2, Dy, -grad);
      total_cgs += cg_iters;
      if (verbose_level >= 2)
        println("[Iteration (", t_block_iter,
                ", ", iter, ")]: Newton solve in ", cg_iters,
                " CG iters");
      end
      
      # Now, perform backtracking line search until solved
      dx = d_newton[1:dd];
      dy = d_newton[(dd+1):end];
      PRS.MultiplyByDHS!(dx, s_vec, D, d_work_vec, Adx);
      new_obj = Inf;
      stepsize = 1.0 / beta_mult;
      predicted_decrease = dot(grad, d_newton);
      if (verbose_level >= 3)
        println("\t[Iter ", iter, "] Predicted decrease: ", predicted_decrease);
      end
      while (new_obj > obj + alpha_mult * stepsize * predicted_decrease)
        stepsize *= beta_mult;
        if (any(y + stepsize * dy .<= abs(Ax - c)))
          new_obj = Inf;
        else
          new_obj = (t_val * sum(y + stepsize * dy)
                     + (t_val/2) * norm(x + stepsize * dx)^2
                     - sum(log(y + stepsize * dy - c + Ax + stepsize * Adx))
                     - sum(log(y + stepsize * dy + c - Ax - stepsize * Adx)));
        end
      end
      obj = new_obj;
      x = x + stepsize * dx;
      y = y + stepsize * dy;
      newton_dec = sqrt(abs(predicted_decrease));
      if (verbose_level >= 2)
        println("[Iteration (", t_block_iter, ", ", iter, ")]: ",
                "obj = ", obj, " newton decrement ", newton_dec);
      end
    end
    if (verbose_level >= 2)
      println("[Outer iter ", t_block_iter, " (", iter, ")] Newton dec = ",
              newton_dec);
    end
    t_val *= t_mult;
  end
  return (x, total_cgs);
end
