# phase-data-generation.jl
#
# Methods to generate data for the (robust) phase retrieval problem.

# (A, b, x) = GenerateData(nn, dd; dev = 1.0, cond_num = 1.0, rad = 1.0)
#
# Generates data from the model
#
#   b_i = (a_i^T * x)^2 + noise_i
#
# where noise_i = dev * Laplace are Laplace-distributed random
# variables and x is uniform on the sphere of radius rad.
function GenerateData(nn, dd; dev = 1.0, cond_num = 1.0, rad = 1.0)
  A = randn(max(nn, dd), min(dd, nn));
  (Q, R) = qr(A);
  lambdas = linspace(1.0, max(cond_num, 1.0), min(dd, nn));
  A = Q * diagm(lambdas);
  if (nn < dd)
    A = A';
  end
  # A = sign(A);
  x = randn(dd);
  x = rad * x / norm(x);
  b = (A * x).^2 + dev * sign.(2 * rand(nn) - 1) .* log.(rand(nn));
  return (A, b, x);
end

# (A, b, x) = GenerateData(nn, dd; dev = 1.0, cond_num = 1.0, rad = 1.0)
#
# Generates data from the model
#
#   b_i = (a_i^T * x)^2 + noise_i
#
# where noise_i = dev * Laplace are Laplace-distributed random
# variables and x is uniform on the sphere of radius rad.
# 
# A is generated as standard gaussian random matrix. 
function GenerateDataGaussian(nn, dd; dev = 1.0, cond_num = 1.0, rad = 1.0)
  A = randn(nn, dd) * diagm(linspace(1.0, max(cond_num, 1.0), dd));
  x = randn(dd);
  x = rad * x / norm(x);
  b = (A * x).^2 + dev * sign.(2 * rand(nn) - 1) .* log.(rand(nn));
  return (A, b, x);
end

# (obs_vec, s) = HadamardSignedData(x, k; p_fail, dev)
#
# Returns observations from Hadamard data.
#
# Given a vector x, generates a vector b by setting
#
# b = (A * x).^2,
#
# where the matrix A is defined as follows. Let S1, ..., Sk be diagonal
# d-dimensional (where x is d-dimensional) matrices, and let
# Hd be the d-by-d standard {+1, -1}-valued Hadamard matrix,
# defined by the recursion
#
# H1 = [1 1; 1 -1];  H{d+1} = [Hd Hd; Hd -Hd].
#
# Define H = kron(eye(k), Hd) / sqrt(d) and S' = [S1 S2 ... Sk]. Then A = HS.
#
# Randomly corrupts p_fail fraction of the data in b, setting those corrupted
# entries to be Laplace distributed with variance dev^2. Returns the
# observations and s, the length k*d stacked vector of all signs in S.
function HadamardSignedData(x::Vector{Float64}, kk::Int64;
                            p_fail::Real = 0.0, dev::Real = 1.0)
  dd = length(x);
  logd = Hadamard.intlog(dd);
  if (2^logd != dd)
    true_len = 2^(logd + 1);
    x = vcat(x, zeros(true_len - dd));
  end
  dd = length(x);
  nn = kk * dd;
  work_vec = zeros(dd);
  s_vec = sign.(randn(nn)) / sqrt(kk);
  obs_vec = zeros(nn);
  MultiplyByDHS!(x, s_vec, ones(nn), work_vec, obs_vec);
  obs_vec = obs_vec.^2;
  if (p_fail > 0)
    fail_inds = rand(nn) .< p_fail;
    n_fail = sum(fail_inds);
    if (n_fail > 0)
      obs_vec[fail_inds] = dev * sign.(randn(n_fail)) .* log.(rand(n_fail));
    end
  end
  return (obs_vec, s_vec);
end
