# phase-experiment-plotting.jl
#
# Provides a module that performs all the experiments in our paper

module PhaseExperimentPlotting

using PyPlot;

include("robust-quadratics.jl");
include("solvers.jl");

# Change as necessary
base_savepath = "/Users/frankruan/repository/compositeopt/Julia/Data/";
base_savepath = "/Users/jduchi/Data/CompositeResults/";

function isbool(x)
  try
    Bool(x)
    return true;
  catch
    return false;
  end
end

function meshgrid(x, y)
  return [repmat(x',length(y),1)[:] repmat(y,1,length(x))[:]];
end

# mrfs = PlotZeroNoiseExperiment(filename; num_n_vals, loadpath, dim, num_tests_per)
#
# Plots the results of the experiments with zero noise, which (likely) strongly
# depend on initialization. Returns a num_n_vals length vector of the mean
# failures to recover.
function PlotZeroNoiseExperiment(filename = "zero_noise_phase.txt";
                                 loadpath = base_savepath,
                                 num_n_vals::Int64 = 16,
                                 dim::Int64 = 1000,
                                 num_tests_per::Int64 = 50,
                                 num_to_plot::Int64 = num_tests_per,
                                 tolerance::Float64 = 1e-5)
  n_vals = linspace(1.8, 1.8 + .1 * (num_n_vals - 1), num_n_vals) * dim;
  n_vals = map((x) -> round(Int64, x), n_vals);
  results = readdlm(string(loadpath, "/", filename), ' ');
  results = reshape(results, 2 * num_tests_per, length(n_vals));
  recovery_failures = zeros(Bool, (num_tests_per, length(n_vals)));
  if (isbool(results[1]))
    recovery_failures = map((x) -> Bool(x), results[1:num_tests_per, :]);
  else
    recovery_failures = map(x -> (x > tolerance), results[1:num_tests_per, :]);
  end
  iteration_counts = map((x) -> Int64(x), results[(num_tests_per + 1):end, :]);
  # Get the actual number of experiments finished
  num_finished = min(findlast(iteration_counts[:, 1] .> 0), num_to_plot);
  println("Plotting the results of ", num_finished, " experiments.");
  recovery_failures = recovery_failures[1:num_finished, :];
  iteration_counts = iteration_counts[1:num_finished, :];
  mean_recovery_failures = zeros(length(n_vals));
  for ii = 1:length(n_vals)
    mean_recovery_failures[ii] = sum(recovery_failures[:, ii]) / num_finished;
  end

  # Plot recovery probabilities
  figure();
  plot(n_vals / dim, 1 - mean_recovery_failures);
  return mean_recovery_failures;
end

function CompareMultipleExperiments(
  file_list_a::Array{String},
  file_list_b::Array{String},
  n_vals::Array{Int64} = 33 * ones(Int64, length(file_list_a)),
  dims::Array{Int64} = 1000 * ones(Int64, length(file_list_a)),
  n_tests::Array{Int64} = 100 * ones(Int64, length(file_list_a));
  loadpath = string(homedir(), "/Data/CompositeResults/"),
  plot_name_a = "A",
  plot_name_b = "B",
  savefile_compare = "",
  savefile_success = "",
  tolerance::Float64 = 1e-3)
  # First, we check that both sides have the same number of experiment lists
  if (length(file_list_a) != length(file_list_b))
    error("Different numbers of files in the two lists");
  end
  # Save the linewidth to use
  num_comparisons = length(file_list_a);
  linewidth = 1;
  if (num_comparisons <= 2)
    linewidth = 2;
  end
  
  # Get the number of measurements (n_vals) that we actually care about
  num_n_vals = minimum(n_vals);
  # Now, construct the results, which include times a or b won and their
  # (empirical) recovery probabilities
  wins_a_over_b = zeros(num_n_vals, num_comparisons);
  wins_b_over_a = zeros(num_n_vals, num_comparisons);
  recovery_probs_a = zeros(num_n_vals, num_comparisons);
  recovery_probs_b = zeros(num_n_vals, num_comparisons);
  for ii = 1:num_comparisons
    num_tests = n_tests[ii];
    n_vals_this = n_vals[ii];
    results_a = readdlm(string(loadpath, "/", file_list_a[ii]), ' ');
    results_a = reshape(results_a, 2 * num_tests, n_vals_this);
    results_b = readdlm(string(loadpath, "/", file_list_b[ii]), ' ');
    results_b = reshape(results_b, 2 * num_tests, n_vals_this);
    recovery_failures_a = zeros(Bool, (num_tests, n_vals_this));
    recovery_failures_b = zeros(Bool, (num_tests, n_vals_this));
    if (isbool(results_a[1]))
      recovery_failures_a = map(x -> Bool(x), results_a[1:num_tests, :]);
    else
      recovery_failures_a = map(x -> x > tolerance, results_a[1:num_tests, :]);
    end
    if (isbool(results_b[1]))
      recovery_failures_b = map(x -> Bool(x), results_b[1:num_tests, :]);
    else
      recovery_failures_b = map(x -> x > tolerance, results_b[1:num_tests, :]);
    end
    iteration_counts_a = map(x -> Int64(x), results_a[(num_tests+1):end, :]);
    iteration_counts_b = map(x -> Int64(x), results_b[(num_tests+1):end, :]);
    num_finished = min(findlast(iteration_counts_a[:, 1] .> 0),
                       findlast(iteration_counts_b[:, 1] .> 0));
    println("Comparing ", num_finished, " experiments between ",
            file_list_a[ii], " and ", file_list_b[ii]);
    for jj = 1:min(n_vals_this, num_n_vals)
      wins_a_over_b[jj, ii] =
        sum(recovery_failures_b[1:num_finished, jj] .&
            .!recovery_failures_a[1:num_finished, jj]) / num_finished;
      wins_b_over_a[jj, ii] =
        sum(recovery_failures_a[1:num_finished, jj] .&
            .!recovery_failures_b[1:num_finished, jj]) / num_finished;
    end
    recovery_probs_a[:, ii] =
      1 - mean(recovery_failures_a[1:num_finished, :], 1)[1:num_n_vals];
    recovery_probs_b[:, ii] =
      1 - mean(recovery_failures_b[1:num_finished, :], 1)[1:num_n_vals];
  end
  

  max_advantage_ind = 0;
  for ii = 1:num_comparisons
    max_advantage_ind = max(findlast(wins_a_over_b[:, ii] .> 0),
                            findlast(wins_b_over_a[:, ii] .> 0),
                            max_advantage_ind);
  end
  max_advantage_ind = min(max_advantage_ind + 1, num_n_vals);
  println("Maximum advantage index:", max_advantage_ind);
  
  # Now it is time to actually plot the comparisons. First, we generate a plot
  # of the actual recovery probabilities
  # line_styles = ["-", "-.", ":"];
  comparable_n_rats = linspace(1.8, 1.8 + .1 * (num_n_vals - 1), num_n_vals);
  marker_styles = ["v", "+", "o", "d", "s"];
  marker_sizes = [6, 9, 6, 6, 5];
  marker_edges = [1, 2, 1, 1, 1];
  figure();
  for ii = 1:num_comparisons
    ind = mod(ii - 1, 5) + 1;
    plot(comparable_n_rats[1:max_advantage_ind],
         recovery_probs_a[1:max_advantage_ind, ii][:],
         color = "b",
         linestyle = "-",
         linewidth = linewidth,
         marker = marker_styles[ind],
         markersize = marker_sizes[ind],
         markeredgecolor = "b",
         markeredgewidth = marker_edges[ind]);
    plot(comparable_n_rats[1:max_advantage_ind],
         recovery_probs_b[1:max_advantage_ind, ii][:],
         color = "k",
         linestyle = ":",
         linewidth = linewidth,
         marker = marker_styles[ind],
         markersize = marker_sizes[ind],
         markeredgecolor = "k",
         markeredgewidth = marker_edges[ind]);
  end
  legend([plot_name_a, plot_name_b], fontsize=18, loc="lower right");
  xlabel(L"m/n", fontsize=20);
  ylabel("P(success)", fontsize=20);
  axis([comparable_n_rats[1], comparable_n_rats[max_advantage_ind],
        ylim()[1], ylim()[2]]);
  if (!isempty(savefile_success))
    savefig(savefile_success, bbox_inches="tight");
  end

  # Now plot the "wins" of one versus the other
  figure();
  for ii = 1:num_comparisons
    ind = mod(ii - 1, 5) + 1;
    plot(comparable_n_rats[1:max_advantage_ind],
         wins_a_over_b[1:max_advantage_ind, ii][:],
         color = "b",
         linestyle = "-",
         marker = marker_styles[ind],
         linewidth = linewidth,
         markeredgecolor = "b",
         markersize = marker_sizes[ind],
         markeredgewidth = marker_edges[ind]);
    plot(comparable_n_rats[1:max_advantage_ind],
         wins_b_over_a[1:max_advantage_ind, ii][:],
         color = "k",
         linestyle = ":",
         linewidth = linewidth,
         marker = marker_styles[ind],
         markeredgecolor = "k",
         markersize = marker_sizes[ind],
         markeredgewidth = marker_edges[ind]);
  end
  max_y = max(maximum(wins_b_over_a), maximum(wins_a_over_b));
  axis([comparable_n_rats[1], comparable_n_rats[max_advantage_ind],
        0.0, ceil(max_y, 2)]);
  legend([plot_name_a, plot_name_b], fontsize=18, loc="upper right");
  xlabel(L"m/n", fontsize=20);
  ylabel("Fraction of one-sided success", fontsize=18);
  if (!isempty(savefile_compare))
    savefig(savefile_compare, bbox_inches="tight");
  end
end
  

# (ad_a, ad_b) = CompareZeroNoiseExperiments(file_a, file_b;
#                                            loadpath = base_savepath,
#                                            num_n_vals::Int64 = 33,
#                                            dim::Int64 = 1000,
#                                            num_tests_per::Int64 = 100)
#
# Takes the results of PlotZeroNoiseExperiment above for two experiments, whose
# data are stored in file_a and file_b, and compares their results. In
# particular, counts the number of times the experiment A succeeded in recovery
# while experiment B failed, and vice versa, for each value of sample size n.
#
# The returned vectors ad_a and ad_b are both of size num_n_vals and contain
# counts (for each sample size) of the times that experiment A succeeded while
# B failed (ad_a) and vice versa (ad_b).
function CompareZeroNoiseExperiments(file_a, file_b;
                                     loadpath = base_savepath,
                                     num_n_vals::Int64 = 33,
                                     dim::Int64 = 1000,
                                     num_tests_per::Int64 = 100,
                                     plot_name_a = "A",
                                     plot_name_b = "B",
                                     savefile_compare = "",
                                     savefile_success = "",
                                     tolerance::Float64 = 1e-3)
  n_vals = linspace(1.8, 1.8 + .1 * (num_n_vals - 1), num_n_vals) * dim;
  n_vals = map((x) -> round(Int64, x), n_vals);
  results_a = readdlm(string(loadpath, "/", file_a), ' ');
  results_a = reshape(results_a, 2 * num_tests_per, length(n_vals));
  recovery_failures_a = zeros(Bool, (num_tests_per, length(n_vals)));
  if (isbool(results_a[1]))
    recovery_failures_a = map(x -> Bool(x), results_a[1:num_tests_per, :]);
  else
    recovery_failures_a = map(x -> (x > tolerance),
                              results_a[1:num_tests_per, :]);
  end
  iteration_counts_a = map((x) -> Int64(x), results_a[(num_tests_per + 1):end, :]);
  results_b = readdlm(string(loadpath, "/", file_b), ' ');
  results_b = reshape(results_b, 2 * num_tests_per, length(n_vals));
  recovery_failures_b = zeros(Bool, (num_tests_per, length(n_vals)));
  if (isbool(results_b[1]))
    recovery_failures_b = map((x) -> Bool(x), results_b[1:num_tests_per, :]);
  else
    recovery_failures_b = map(x -> (x > tolerance),
                              results_b[1:num_tests_per, :]);
  end
  iteration_counts_b = map((x) -> Int64(x), results_b[(num_tests_per + 1):end, :]);
  num_finished = min(findlast(iteration_counts_a[:, 1] .> 0),
                     findlast(iteration_counts_b[:, 1] .> 0));
  println("Comparing ", num_finished, " experiments.");
  advantages_a = zeros(length(n_vals));
  advantages_b = zeros(length(n_vals));
  for ii = 1:length(n_vals)
    advantages_a[ii] = sum(recovery_failures_b[1:num_finished, ii] .&
                           .!recovery_failures_a[1:num_finished, ii]) / num_finished;
    advantages_b[ii] = sum(recovery_failures_a[1:num_finished, ii] .&
                           .!recovery_failures_b[1:num_finished, ii]) / num_finished;
  end
  figure();
  max_advantage_n = max(findlast(advantages_a .> 0),
                        findlast(advantages_b .> 0));
  plot(n_vals[1:max_advantage_n] / dim, advantages_a[1:max_advantage_n],
       "bo-", linewidth=2, label=plot_name_a);
  plot(n_vals[1:max_advantage_n] / dim, advantages_b[1:max_advantage_n],
       "ks:", linewidth=2, label=plot_name_b);
  ylabel("Fraction of one-sided success", fontsize=20);
  xlabel(L"m/n", fontsize=20);
  legend(fontsize=20);
  if (!isempty(savefile_compare))
    savefig(savefile_compare, bbox_inches="tight");
  end
  
  # Plot recovery probabilities
  figure();
  mean_a_failures = mean(recovery_failures_a[1:num_finished, :], 1)[:];
  mean_b_failures = mean(recovery_failures_b[1:num_finished, :], 1)[:];
  plot(n_vals / dim, 1 - mean_a_failures,
       "bo-", linewidth=2, label=plot_name_a);
  plot(n_vals / dim, 1 - mean_b_failures,
       "ks:", linewidth=2, label=plot_name_b);
  ylabel("P(success)", fontsize=20);
  xlabel(L"m/n", fontsize=20);
  if (!isempty(savefile_success))
    savefig(savefile_success, bbox_inches="tight");
  end
  return (advantages_a, advantages_b);
end

# PlotPhaseTransitionExperiment
#
# Plots the results of the method PhaseTransitionExperiment.
function PlotPhaseTransitionExperiment(filename="result_phase_transition.txt";
                                       loadpath = base_savepath,
                                       dim::Int64 = 256,
                                       num_p_fail_vals::Int64 = 31,
                                       num_tests_per::Int64 = 50,
                                       p_low::Float64 = .2,
                                       p_high::Float64 = .8,
                                       save_surface::String = "",
                                       save_phase_box::String = "",
                                       save_iterations::String = "",
                                       save_p_recovery::String = "",
                                       tolerance::Float64 = 1e-3)
  results = readdlm(string(loadpath, "/", filename), ' ');
  n_vals = [round(Int64, 1.8 * dim); 2 * dim; round(Int64, 2.5 * dim);
            3 * dim; 4 * dim; 6 * dim; 8 * dim];
  p_fail_vals = linspace(0, .3, num_p_fail_vals);
  # results should be a 2 * num_tests_per -by- length(n_vals) -by-
  # length(p_fail_vals) matrix.
  results = reshape(results, 2 * num_tests_per, length(n_vals),
                    length(p_fail_vals));
  recovery_failures = zeros(Bool, (num_tests_per, length(n_vals),
                                   length(p_fail_vals)));
  if (isbool(results[1]))
    recovery_failures = map((x) -> Bool(x), results[1:num_tests_per, :, :]);
  else
    recovery_failures = map(x -> (x > tolerance), results[1:num_tests_per, :, :]);
  end
  iteration_counts =
    map((x) -> Int64(x), results[(num_tests_per+1):end, :, :]);
  mean_recovery_failures = zeros(length(n_vals), length(p_fail_vals));
  for ii = 1:length(n_vals)
    for jj = 1:length(p_fail_vals)
      mean_recovery_failures[ii, jj] =
        sum(recovery_failures[:, ii, jj]) / num_tests_per;
    end
  end

  # color_choices = ["kv-", "b+-", "ro-", "gd-", "ks--"];
  color_choices = ["k", "b", "r", "g", "k", "b", "r", "b"];
  marker_choices = ["v", "+", "o", "d", "s", "^", "*", "p"];
  line_choices = ["-", "-", "-", "-", "-.", "-.", "-.", "-."];
  marker_sizes = [6, 9, 6, 6, 5, 6, 6, 5];
  marker_edge_widths = [1, 2, 1, 1, 1, 1, 1, 1];
  fill_colors = ["k", "b", "r", "g", "k"];
  
  # First, plot recovery probabilities for each of the methods
  figure();
  plot_surface(repmat(n_vals, 1, length(p_fail_vals)),
               repmat(p_fail_vals', length(n_vals), 1),
               mean_recovery_failures,
               cmap = ColorMap("Blues"), rstride=1, cstride=1, linewidth=.25);
  if (!isempty(save_surface))
    savefig(save_surface, bbox_inches="tight");
  end
  # sample size vs. recovery probability for different failure probabilities.
  
  figure();
  legend_strings = Array{String}(length(n_vals));
  for ii = 1:length(n_vals)
    plot(p_fail_vals, 1 - mean_recovery_failures[ii, :][:],
         color = color_choices[ii],
         linestyle = line_choices[ii],
         marker = marker_choices[ii],
         linewidth=2,
         markersize=marker_sizes[ii],
         markeredgewidth=marker_edge_widths[ii]);
    legend_strings[ii] = string(L"m/n = ", round(n_vals[ii] / dim, 2));
  end
  if (!isempty(save_p_recovery))
    savefig(save_p_recovery, bbox_inches="tight");
  end

  # Now plot image of successful recoveries
  figure();
  imshow(1 - mean_recovery_failures, cmap="gray", interpolation="nearest");
  yticks(0:(length(n_vals)-1), n_vals / dim);
  xticks(0:2:length(p_fail_vals), p_fail_vals[1:2:end]);
  if (!isempty(save_phase_box))
    println("Saving with no interpolation");
    savefig(save_phase_box, bbox_inches="tight");
  end

  figure();
  key_colors = zeros(2, 100);
  key_colors[1, :] = (0:99) / 100;
  key_colors[2, :] = (0:99) / 100;
  imshow(key_colors, cmap="gray");
  yticks([]);
  xticks([1, 25, 50, 75, 100], [0, .25, .5, .75, 1.0]);
  savefig("recovery-colors.pdf", bbox_inches="tight");
  
  PlotPhaseIterationCounts(dim,
                           color_choices, marker_choices,
                           line_choices, marker_sizes, marker_edge_widths,
                           fill_colors, n_vals, p_fail_vals,
                           mean_recovery_failures, recovery_failures,
                           iteration_counts, legend_strings, p_low, p_high,
                           save_iterations);
  return (recovery_failures, iteration_counts);
end

# PlotPhaseIterationCounts
#
# Helper function to plot iteration counts for the
# PlotPhaseTransitionExperiment method.
function PlotPhaseIterationCounts(dim::Int64,
                                  color_choices, marker_choices,
                                  line_choices, marker_sizes, marker_edge_widths,
                                  fill_colors,
                                  n_vals,
                                  p_fail_vals,
                                  mean_recovery_failures::Array{Float64, 2},
                                  recovery_failures::Array{Bool, 3},
                                  iteration_counts::Array{Int64, 3},
                                  legend_strings::Array{String, 1},
                                  p_low::Float64, p_high::Float64,
                                  save_iterations::String)
  # Now, plot iteration counts on successes for each method
  max_iter = maximum(iteration_counts);
  med_successful_iterations = zeros(length(n_vals), length(p_fail_vals));
  p_low = max(p_low, 0.0);
  low_successful_iterations = zeros(length(n_vals), length(p_fail_vals));
  p_high = min(p_high, 1.0);
  high_successful_iterations = zeros(length(n_vals), length(p_fail_vals));
  std_successful = zeros(length(n_vals), length(p_fail_vals));
  for n_ind = 1:length(n_vals)
    non_all_fail_inds = find(mean_recovery_failures[n_ind, :][:] .< 1);
    for good_p_ind in non_all_fail_inds
      success_inds = find(.!recovery_failures[:, n_ind, good_p_ind]);
      med_successful_iterations[n_ind, good_p_ind] =
        median(iteration_counts[success_inds, n_ind, good_p_ind]);
      high_successful_iterations[n_ind, good_p_ind] =
        quantile(iteration_counts[success_inds, n_ind, good_p_ind], p_high);
      low_successful_iterations[n_ind, good_p_ind] =
        quantile(iteration_counts[success_inds, n_ind, good_p_ind], p_low);
      if (length(success_inds) > 1)
        std_successful[n_ind, good_p_ind] =
          (std(iteration_counts[success_inds, n_ind, good_p_ind]) /
           sqrt(length(success_inds)));
      else
        std_successful[n_ind, good_p_ind] = 0; # max_iter;
      end
    end
  end
  figure();
  low_y = Inf;
  high_y = 0;
  for n_ind = 1:length(n_vals)
    success_inds =
      find(mean_recovery_failures[n_ind, :][:] .< .8);
    # println(success_inds);
    # println(med_successful_iterations[n_ind, success_inds]);
    plot(p_fail_vals[success_inds],
         med_successful_iterations[n_ind, success_inds],
         color = color_choices[n_ind], linestyle = line_choices[n_ind],
         marker = marker_choices[n_ind],
         linewidth=2, markersize=marker_sizes[n_ind],
         markeredgewidth=marker_edge_widths[n_ind],
         label=legend_strings[n_ind]);
    errorbar(p_fail_vals[success_inds],
             med_successful_iterations[n_ind, success_inds],
             std_successful[n_ind, success_inds],
             color = color_choices[n_ind], linestyle = line_choices[n_ind],
             marker = marker_choices[n_ind],
             linewidth=1.5, markersize=marker_sizes[n_ind],
             markeredgewidth=marker_edge_widths[n_ind]);
    if (length(success_inds) > 0)
      low_y =
        min(low_y,
            minimum(med_successful_iterations[n_ind, success_inds]
                    - std_successful[n_ind, success_inds]));
      high_y =
        max(high_y,
            maximum(med_successful_iterations[n_ind, success_inds]
                    + std_successful[n_ind, success_inds]));
    end
  end
  min_plot_iter =
    minimum(low_successful_iterations[low_successful_iterations .> 0]);
  legend(loc = "upper right", fontsize = 16);
  low_y = floor(Int64, low_y);
  high_y = ceil(Int64, high_y);
  axis([p_fail_vals[1], p_fail_vals[end], low_y, high_y]);
  if (!isempty(save_iterations))
    savefig(save_iterations, bbox_inches="tight");
  end
  # big_n_vals = n_vals[n_vals .>= 2];
  # xticks(big_n_vals / dim, map((x) -> string(x),
  #                              round(big_n_vals/dim, 1)));
  # axis([2, 8.5, min_plot_iter - 1, maximum(high_successful_iterations)]);
end

# (wins, pv_wins, pv_vals) = GivePValues(filename = "phase_result_cor_1.txt";
#                                        num_experiments = 100,
#                                        loadpath = base_savepath)
#
# Compares performance of stochastic prox-linear method with stochastic
# gradient method. Loads data at filename (from loadpath) and computes the
# number of times the stochastic proximal method beat the stochastic gradient
# method over iterations, as well as the significance of the gaps in their
# objective values.
#
# Returns three vectors of length number-of-iterations, the first counting the
# number of experiments in which stochastic prox-linear won, the second
# a p-value for the hypothesis that each method is equally likely to have
# better objective value than the other (Bernoulli test), and the third
# a p-value for the gaps in function values relative to variance in function
# values.
function GivePValues(filename = "phase_result_cor_1.txt";
                     num_experiments = 100,
                     loadpath = base_savepath)
  nn = 500;
  dim = 50;
  Niter = 100000;
  maxiter = convert(Int, Niter/nn)+1;
  xiter = collect(1:1:(maxiter-1));
  num_prox_iter = convert(Int, Niter / nn) + 1;

  results = [];
  results = readdlm(string(loadpath, "/", filename), ' ');
  # results should be a (3 * num_experiments) -by- num_prox_iter matrix
  results = reshape(results, 3 * num_experiments,
                    num_prox_iter);
  result_prox = results[1:num_experiments, :];
  result_sprox = results[(num_experiments+1):(2*num_experiments), :];
  result_sgd = results[(2*num_experiments+1):(3*num_experiments), :];
  minvalue = min(minimum(result_prox, 2),
                 minimum(result_sprox, 2),
                 minimum(result_sgd, 2));
  result_prox = (result_prox - minvalue * ones(1, maxiter));
  result_sprox = (result_sprox - minvalue * ones(1, maxiter));
  result_sgd = (result_sgd - minvalue * ones(1, maxiter));
  result_sprox_vs_sgd = zeros(maxiter);
  wins_sprox_vs_sgd = zeros(maxiter);
  # Get the confidence intervals
  for iter = 1:maxiter
    result_sprox_vs_sgd[iter] =
      (mean(result_sgd[:, iter] - result_sprox[:, iter])
       / (std(result_sgd[:, iter] - result_sprox[:, iter]) /
          sqrt(num_experiments)));
    wins_sprox_vs_sgd[iter] = sum(result_sgd[:, iter] .> result_sprox[:, iter]);
  end
  p_vals_sgd_vs_sprox = 1 - (.5 * erf(result_sprox_vs_sgd / sqrt(2)) + 1/2);
  # Binomial tails for winning p-values
  p_vals_wins = exp(-num_experiments *
                    max(wins_sprox_vs_sgd / num_experiments - .5, 0).^2
                    / (2 * .5 * .5));
  return (wins_sprox_vs_sgd, p_vals_wins, p_vals_sgd_vs_sprox);
end

# PlotSimplePhaseRetrieval(filename = "result_noiseful.txt";
#  num_experiments = 250, loadpath = base_savepath,
#  ylimits = (), savefile = "")
#
# Plots the results of a phase retrieval experiment, as computed by
# the method BasicPhaseRetrievalExperiment().
function PlotSimplePhaseRetrieval(filename = "result_noiseful.txt";
                                  num_experiments = 100,
                                  loadpath = base_savepath,
                                  ylimits = (),
                                  savefile = "")
  nn = 500;
  dim = 50;
  Niter = 100000;
  maxiter = convert(Int, Niter/nn)+1;
  xiter = collect(1:1:(maxiter-1));
  num_prox_iter = convert(Int, Niter / nn) + 1;

  results = [];
  results = readdlm(string(loadpath, "/", filename), ' ');
  # results should be a (3 * num_experiments) -by- num_prox_iter matrix
  results = reshape(results, 3 * num_experiments,
                    num_prox_iter);
  result_prox = results[1:num_experiments, :];
  result_sprox = results[(num_experiments+1):(2*num_experiments), :];
  result_sgd = results[(2*num_experiments+1):(3*num_experiments), :];
  minvalue = min(minimum(result_prox, 2),
                 minimum(result_sprox, 2),
                 minimum(result_sgd, 2));
  result_prox = (result_prox - minvalue * ones(1, maxiter));
  result_sprox = (result_sprox - minvalue * ones(1, maxiter));
  result_sgd = (result_sgd - minvalue * ones(1, maxiter));
  result_prox_up = zeros(maxiter);
  result_prox_down = zeros(maxiter);
  result_prox_med = zeros(maxiter);
  result_sprox_up = zeros(maxiter);
  result_sprox_down = zeros(maxiter);
  result_sprox_med = zeros(maxiter);
  result_sgd_up = zeros(maxiter);
  result_sgd_down = zeros(maxiter);
  result_sgd_med = zeros(maxiter);
  # Get the confidence intervals
  for iter = 1:maxiter
    result_prox_up[iter] = quantile(result_prox[:, iter], 0.9);
    result_prox_med[iter] = quantile(result_prox[:, iter], 0.5);
    result_prox_down[iter] = quantile(result_prox[:, iter], 0.1);
    result_sprox_up[iter] = quantile(result_sprox[:, iter], 0.9);
    result_sprox_med[iter] = quantile(result_sprox[:, iter], 0.5);
    result_sprox_down[iter] = quantile(result_sprox[:, iter], 0.1);
    result_sgd_up[iter] = quantile(result_sgd[:, iter], 0.9);
    result_sgd_med[iter] = quantile(result_sgd[:, iter], 0.5);
    result_sgd_down[iter] = quantile(result_sgd[:, iter], 0.1);
  end

  # Colors for plotting
  alpha_level = .5;
  sgd_col = (0.5, 0.5, 0.5);
  sprox_col = (0.5, 1.0, 0.5);
  prox_col = (0.5, 0.5, 1.0);

  figure();
  semilogy(xiter, result_prox_med[xiter], "b-", linewidth = 2,
           label = "prox");
  semilogy(xiter, result_sprox_med[xiter], "g--", linewidth = 2,
           label = "sprox");
  semilogy(xiter, result_sgd_med[xiter], "k:", linewidth = 2,
           label = "sgd");
  fill_between(xiter, result_prox_down[xiter],
               result_prox_up[xiter],
               color = prox_col, alpha = alpha_level);
  semilogy(xiter, result_prox_down[xiter], "b-");
  semilogy(xiter, result_prox_up[xiter], "b-");
  fill_between(xiter, result_sprox_down[xiter],
               result_sprox_up[xiter],
               color = sprox_col, alpha = alpha_level);
  semilogy(xiter, result_sprox_down[xiter], "g--");
  semilogy(xiter, result_sprox_up[xiter], "g--");
  fill_between(xiter, result_sgd_down[xiter],
               result_sgd_up[xiter],
               color = sgd_col, alpha = alpha_level);
  semilogy(xiter, result_sgd_down[xiter], "k:");
  semilogy(xiter, result_sgd_up[xiter], "k:");
  legend();
  if (!isempty(ylimits))
    ylim(ylimits);
  end
  if (!isempty(savefile))
    savefig(savefile, bbox_inches="tight");
  end
end

# PlotConditionNumber(loadpath = base_savepath,
#                             savepath = "../Draft/Paper/New/figure/",
#                             num_experiments = 250)
#
# Loads data for the condition number experiment stored at loadpath and plots
# figures demonstrating performance in the folder specified by savepath.
function PlotConditionNumber(;loadpath = base_savepath,
                             savepath = "../Draft/Paper/New/figure/",
                             num_experiments = 250,
                             p_val = .1)
  Niter = 100000;
  n = 500;
  d = 50;
  condition_numbers = [1.0, 10.0, 100.0];
  num_prox_linear_iterations = convert(Int, Niter / n);
  Iteration = convert(Int, Niter/n)+1;
  xiter = collect(1:1:(Iteration-1));
  result_cond = readdlm(string(loadpath, "/result_cond_num_opt.txt"), ' ');
  # result_cond should be a 3 * num_experiments -by- 3 -by-
  # num_prox_linear_iterations + 1 tensor, because there are 3 condition
  # numbers and it consists of results for sprox, sgd, and prox_linear.
  result_cond = reshape(result_cond, 3 * num_experiments,
                        length(condition_numbers),
                        num_prox_linear_iterations + 1);
  # result_prox should be a num_experiments -by- 3 -by-
  # num_prox_linear_iterations+1 matrix (tensor).
  result_prox = result_cond[1:num_experiments, :, :];
  result_sprox = result_cond[(num_experiments + 1):(2 * num_experiments),
                             :, :];
  result_sgd = result_cond[(2 * num_experiments + 1):(3 * num_experiments),
                           :, :];

  for cond_ind = 1:length(condition_numbers)
    minvalues = min(minimum(result_prox[:, cond_ind, :], 2),
                    minimum(result_sprox[:, cond_ind, :], 2),
                    minimum(result_sgd[:, cond_ind, :], 2));
    result_prox[:, cond_ind, :] =
      result_prox[:, cond_ind, :] - minvalues * ones(1, Iteration);
    result_sprox[:, cond_ind, :] =
      result_sprox[:, cond_ind, :] - minvalues * ones(1, Iteration);
    result_sgd[:, cond_ind, :] =
      result_sgd[:, cond_ind, :] - minvalues * ones(1, Iteration);
  end

  result_prox_med = zeros(Iteration, length(condition_numbers));
  result_sgd_med = zeros(Iteration, length(condition_numbers));
  result_sprox_med = zeros(Iteration, length(condition_numbers));
  result_prox_up = zeros(Iteration, length(condition_numbers));
  result_sgd_up = zeros(Iteration, length(condition_numbers));
  result_sprox_up = zeros(Iteration, length(condition_numbers));
  result_prox_low = zeros(Iteration, length(condition_numbers));
  result_sgd_low = zeros(Iteration, length(condition_numbers));
  result_sprox_low = zeros(Iteration, length(condition_numbers));

  # Now get median and quantile performance for each
  for cond_ind = 1:length(condition_numbers)
    for iter = 1:Iteration
      result_prox_med[iter, cond_ind] =
        quantile(result_prox[:, cond_ind, iter], 0.5);
      result_sprox_med[iter, cond_ind] =
        quantile(result_sprox[:, cond_ind, iter], 0.5);
      result_sgd_med[iter, cond_ind] =
        quantile(result_sgd[:, cond_ind, iter], 0.5);
      result_prox_low[iter, cond_ind] =
        quantile(result_prox[:, cond_ind, iter], p_val);
      result_sprox_low[iter, cond_ind] =
        quantile(result_sprox[:, cond_ind, iter], p_val);
      result_sgd_low[iter, cond_ind] =
        quantile(result_sgd[:, cond_ind, iter], p_val);
      result_prox_up[iter, cond_ind] =
        quantile(result_prox[:, cond_ind, iter], 1 - p_val);
      result_sprox_up[iter, cond_ind] =
        quantile(result_sprox[:, cond_ind, iter], 1 - p_val);
      result_sgd_up[iter, cond_ind] =
        quantile(result_sgd[:, cond_ind, iter], 1 - p_val);
    end
  end

  result_sgd_low = max(result_sgd_low, 1e-10);
  result_sprox_low = max(result_sprox_low, 1e-10);
  result_prox_low = max(result_prox_low, 1e-10);

  # Colors for plotting
  alpha_level = .5;
  sgd_col = (0.5, 0.5, 0.5);
  sprox_col = (0.5, 1.0, 0.5);
  prox_col = (0.5, 0.5, 1.0);
  
  for cond_ind = 1:length(condition_numbers)
    figure();
    semilogy(xiter, result_prox_med[xiter, cond_ind],
             "b-", linewidth = 3, label = "prox");
    semilogy(xiter, result_sprox_med[xiter, cond_ind],
             "g--", linewidth = 3, label = "sprox");
    semilogy(xiter, result_sgd_med[xiter, cond_ind],
             "k:", linewidth = 3, label = "sgd");
    fill_between(xiter, result_prox_low[xiter, cond_ind],
                 result_prox_up[xiter, cond_ind],
                 color = prox_col, alpha=alpha_level);
    fill_between(xiter, result_sgd_low[xiter, cond_ind],
                 result_sgd_up[xiter, cond_ind],
                 color = sgd_col, alpha=alpha_level);
    fill_between(xiter,
                 result_sprox_low[xiter, cond_ind],
                 result_sprox_up[xiter, cond_ind],
                 color = sprox_col, alpha=alpha_level);
    title(string("cond = ", condition_numbers[cond_ind]));
    (ymin, ymax) = ylim();
    ylim(max(1e-4, ymin), ymax);
    legend();
    savefig(string(savepath, "cond_num", cond_ind, ".pdf"),
            bbox_inches="tight");
  end
end

end  # module PhaseExperimentPlotting
