# phase-experiments.jl
#
# Module for a number of Phase retrieval experiments.

module PhaseExperiments

include("solvers.jl");
include("Hadamard.jl");

# Change as necessary
base_savepath = "/Users/jduchi/Data/CompositeResults/";
base_savepath = "./CompositeResults/";

function meshgrid(x, y)
  return [repmat(x',length(y),1)[:] repmat(y,1,length(x))[:]];
end

# ZeroNoisePhaseTransitionExperiment(filename = "zero_noise_phase.txt";
#                                    num_n_vals::Int64 = 16,
#                                    savepath = base_savepath,
#                                    dim::Int64 = 1000,
#                                    num_tests_per::Int64 = 50,
#                                    tolerance::Float64 = 1e-3,
#                                    verbose_level::Int64 = 0,
#                                    prop_init::Float64 = .5,
#                                    init_type::Symbol = :LM,
#                                    alg_type::Symbol = :PL,
#                                    signal_type::Symbol = :signs,
#                                    seed::Int64 = 2017)
#
# Performs an experiment on recovery thresholds for phase transitions
# with zero noise. The parameters are as follows.
#
#  - filename - Name of file to store results
#  - dim - The dimension of problem to solve
#  - num_n_vals - The number of different sample sizes to use, where
#    we choose sample sizes linearly spaced from 1.8 * dim to
#    (1.8 + .1 * (num_n_vals-1)) * dim. Default is num_n_vals = 16, so
#    the largest sample size is 3.3.
#  - num_tests_per - Number of experiments to run for each parameter setting
#  - verbose_level - How verbose to be during runs
#  - prop_init - Proportion of data to use in spectral-like initializations.
#    Recommended setting when using init_type = :LM (largest) is 1/6 = .167,
#    for init_type = :SM (smallest) about 1/2 = .5 is our theoretical value.
#  - init_type - Type of initialization to use. :LM gives "largest" spectral
#    initializer, based on Wang, Giannakis, Eldar. :SM gives "smallest"
#    spectral initializer, and :RND gives random.
#  - alg_type - Whether to use prox-linear with accurate (Mosek) solves (:PL),
#    truncated amplitude flow (:TAF), or prox-linear with innacurate graph
#    form solves (:PL_graph)
#  - signal_type - The signal x to use, whether random signs (:signs) or
#    just standard Gaussian (whenever signal_type != :signs).
#  - seed - The seed for the random number generator
function ZeroNoisePhaseTransitionExperiment(filename = "zero_noise_phase.txt";
                                            num_n_vals::Int64 = 16,
                                            savepath = base_savepath,
                                            dim::Int64 = 1000,
                                            num_tests_per::Int64 = 50,
                                            tolerance::Float64 = 1e-3,
                                            verbose_level::Int64 = 0,
                                            prop_init::Float64 = .5,
                                            init_type::Symbol = :LM,
                                            alg_type::Symbol = :PL,
                                            signal_type::Symbol = :signs,
                                            seed::Int64 = 2017)
  srand(seed);
  n_vals = linspace(1.8, 1.8 + .1 * (num_n_vals-1), num_n_vals) * dim;
  n_vals = map((x) -> round(Int64, x), n_vals);
  recovery_vals = zeros(num_tests_per, length(n_vals));
  iteration_counts = zeros(Int64, num_tests_per, length(n_vals));
  for iter = 1:num_tests_per
    println("*** [Experiment: ", iter, " of ", num_tests_per, "] ***");
    for n_ind = 1:length(n_vals)
      x = randn(dim);
      if (signal_type == :signs)
        x = sign.(x);
      end
      nn = n_vals[n_ind];
      println("\t [Experiment: ", iter, "] n = ", nn);
      time_begin = time();
      A = randn(nn, dim);
      b = (A * x).^2;
      x_est_dir = zeros(dim);
      magnitude_est = 0.0;
      if (init_type == :LM)
        x_est_dir =
          PRS.OrthogonalityPromotingInitialization(A, b, prop_init=prop_init,
                                                    use_big = true);
        magnitude_est = sqrt(mean(b));
      elseif (init_type == :SM)
        (x_est_dir, magnitude_est) =
          PRS.NoisyOrthPromotingInit(A, b, prop_init);
      elseif (init_type == :RND)
        x_est_dir = randn(dim);
        x_est_dir = x_est_dir / norm(x_est_dir);
      else
        error("Unknown init_type ", init_type);
      end
      x_init = x_est_dir[:] * magnitude_est;
      x_est = zeros(dim);
      num_steps = 0;
      if (alg_type == :PL)
        (x_est, objs, num_steps) =
          PRS.ProxLinearQuadratic(A, b, x_init = x_init, tol=1e-4,
                                  maxiter = 35, qp_solver = :mosek,
                                  verbose_level = verbose_level);
      elseif (alg_type == :PL_graph)
        (x_est, objs, num_steps) =
          PRS.ProxLinearQuadratic(A / sqrt(nn), b / nn,
                                  x_init = x_init, tol=1e-3,
                                  eps_qp=1e-5,
                                  maxiter = 25, qp_solver = :graph,
                                  verbose_level = verbose_level);
        save_num_steps = num_steps;
        (x_est, objs, num_steps) =
          PRS.ProxLinearQuadratic(A / sqrt(nn), b / nn,
                                  x_init = x_est, tol=1e-3,
                                  eps_qp=1e-8,
                                  maxiter = 25, qp_solver = :graph,
                                  verbose_level = verbose_level);
        num_steps += save_num_steps;
      elseif (alg_type == :TAF)
        num_steps = 1000;
        (x_est, objs) = PRS.TruncatedAmpFlow(A, b, x_init = x_init,
                                             verbose_level = 0,
                                             maxiter = num_steps,
                                             gamma_thresh=.7);
      else
        error("Unknown algorithm ", alg_type);
      end
      recovery_vals[iter, n_ind] =
        min(norm(x_est - x) / norm(x), norm(x_est + x) / norm(x));
      iteration_counts[iter, n_ind] = num_steps;
      time_end = time();
      time_taken = time_end - time_begin;
      if (recovery_vals[iter, n_ind] > tolerance)
        println("\t * Failure (", round(recovery_vals[iter, n_ind], 4),
                ") after ", num_steps, " steps (",
                round(time_taken, 3), " seconds) *");
      else
        println("\t * Success (",
                round(recovery_vals[iter, n_ind], 4),
                ") after ", num_steps, " steps (",
                round(time_taken, 3), " seconds) *");
      end
    end
    result = vcat(recovery_vals, iteration_counts);
    writedlm(string(savepath, "/", filename), result, ' ');
  end
  return (recovery_vals, iteration_counts);
end

# (recovery_vals, iteration_counts) =
#   PhaseTransitionExperiment(filename = "result_phase_transition.txt";
#                             savepath = base_savepath,
#                             dim::Int64 = 256,
#                             num_tests_per::Int64 = 50,
#                             fail_mode::Symbol = :zero,
#                             alg_type::Symbol = :PL,
#                             init_type::Symbol = :SM,
#                             tolerance::Float64 = 1e-3,
#                             verbose_level::Int64 = 0)
#
# Performs an experiment on recovery probabilities for the prox-linear method
# with smallest eigenvector initializations. The input parameters are as
# follows:
#
#  - filename - The filename in which to store the results, which are stored as
#    a single stacked vector.
#  - savepath - The path to the folder containing the saved file
#  - dim - The dimension of problem/signal to attempt to recover
#  - num_p_fail_vals - The number of different proportions of data to corrupt.
#  - num_tests_per - Number of experiments to run for each of the various
#    problem settings.
#  - tolerance - Relative tolerance at which we declare success or failure of
#    recovery.
#  - init_type - Type of initialization of search (either :SM for smallest
#    eigenvector or :MTW for median Wirtinger init or :LM for largest eigenvector)
#  - alg_type - Type of algorithm of search (either :PL for prox linear,
#    :MTW for median-truncated Wirtinger flow, or :PL_graph for graph-solved
#    prox linear)
#  - fail_mode - A symbol, either :zero or :cauchy, determining how the
#    outliers are chosen. If :zero, then all outlying measurements are simply
#    set to zero, so that they must be included in any set of "smallest"
#    measurements. If :cauchy, then sets outliers to be the square root of the
#    magnitude of a random Cauchy variable (so they have 2 weak moments).
#  - signal_type - Symbol (:signs or otherwise) to choose random x.
#  - prop_init - float of quantile of data to use when initializing
#  - verbose_level - How much to print out stuff.
#
# Within each experiment, performs tests with failure probabilities ranging in
# linspace(0, .3, num_p_fail_vals) and sample sizes of 1.8 * dim, 2 * dim, 3 *
# dim, 4 * dim, 6 * dim, and 8 * dim. (num_p_fail_vals failure probabilites and
# 7 sample sizes).
# 
# Returns the matrices recovery_vals and iteration_counts, each of which is
# num_tests_per-by-7-by-num_p_fail_vals, where the second two indices index
# sample size and failure probability, respectively. recovery_vals[i, j, k] is
# the relative error of the final iterate in experiment i, with sample size
# indexed by j and failure probability indexed by k. iteration_counts is
# indexed the same way, but counts the number of iterations of the prox-linear
# method.
function PhaseTransitionExperiment(filename = "result_phase_transition.txt";
                                   savepath = base_savepath,
                                   dim::Int64 = 256,
                                   num_p_fail_vals::Int64 = 31,
                                   num_tests_per::Int64 = 50,
                                   fail_mode::Symbol = :zero,
                                   alg_type::Symbol = :PL,
                                   init_type::Symbol = :SM,
                                   tolerance::Float64 = 1e-3,
                                   verbose_level::Int64 = 0,
                                   signal_type::Symbol = :signs,
                                   prop_init::Float64 = .5,
                                   seed::Int64 = 2017)
  # Set seed number
  srand(seed);
  n_vals = [round(Int64, 1.8 * dim); 2 * dim; round(Int64, 2.5 * dim);
            3 * dim; 4 * dim; 6 * dim; 8 * dim];
  p_fail_vals = linspace(0, .3, num_p_fail_vals);
  recovery_vals = zeros(Float64, num_tests_per,
                        length(n_vals), length(p_fail_vals));
  iteration_counts = zeros(Int64, num_tests_per,
                           length(n_vals), length(p_fail_vals));
  for iter = 1:num_tests_per
    println("*** [Experiment: ", iter, " of ", num_tests_per, "] ***");
    for n_ind = 1:length(n_vals);
      x = zeros(dim);
      if (signal_type == :signs)
        x = sign.(randn(dim));
      else
        x = randn(dim);
      end
      for p_ind = 1:length(p_fail_vals)
        t_begin = time();
        p_fail = p_fail_vals[p_ind];
        nn = n_vals[n_ind];
        println("\t [Experiment: ", iter, "] n = ", nn,
                ", P(fail) = ", p_fail);
        A = randn(nn, dim);
        b = (A * x).^2;
        fail_inds = (rand(nn) .< p_fail);
        # Failures are simply 0, to screw everything up?
        if (fail_mode == :zero)
          b[fail_inds] = 0;
        elseif (fail_mode == :cauchy)
          b[fail_inds] = sqrt(abs(randn(sum(fail_inds))
                                  ./ randn(sum(fail_inds))));
        else
          error("Unknown failure mode ", fail_mode);
        end
        x_est_dir = zeros(dim);
        magnitude_est = 0.0;
        if (init_type == :SM)
          (x_est_dir, magnitude_est) =
            PRS.NoisyOrthPromotingInit(A, b, .5, false);
        elseif (init_type == :LM)
          (x_est_dir, magnitude_est) =
            PRS.NoisyOrthPromotingInit(A, b, .5, true);
        elseif (init_type == :MTW)
          (x_est_dir, magnitude_est) = PRS.MedianWirtingerInitialization(A, b);
        else
          error("Unknown initialization type ", init_type);
        end
        x_init = x_est_dir * magnitude_est;
        x_est = zeros(dim);
        num_steps = 0;
        if (alg_type == :PL)
          (x_est, objs, num_steps) =
            PRS.ProxLinearQuadratic(A, b, x_init = x_init, tol=1e-3,
                                     maxiter = 45, qp_solver = :mosek,
                                    verbose_level = verbose_level);
        elseif (alg_type == :PL_graph)
          (x_est, objs, num_steps) =
            PRS.ProxLinearQuadratic(A / sqrt(nn), b / nn,
                                    x_init = x_init, tol=1e-3,
                                    eps_qp=1e-5,
                                    maxiter = 20, qp_solver = :graph,
                                    verbose_level = verbose_level);
          save_num_steps = num_steps;
          (x_est, objs, num_steps) =
            PRS.ProxLinearQuadratic(A / sqrt(nn), b / nn,
                                    x_init = x_est, tol=1e-3,
                                    eps_qp=1e-8,
                                    maxiter = 20, qp_solver = :graph,
                                    verbose_level = verbose_level);
          num_steps += save_num_steps;
        elseif (alg_type == :MTW)
          num_steps = 1500;
          (x_est, objs) =
            PRS.MedianTruncatedWFlow(A, b,
                                     x_init = x_init, maxiter = num_steps,
                                     verbose_level = verbose_level);
        else
          error("Unknown algorithm type ", alg_type);
        end
        recovery_vals[iter, n_ind, p_ind] =
          min(norm(x_est - x) / norm(x), norm(x_est + x) / norm(x));
        iteration_counts[iter, n_ind, p_ind] = num_steps;
        t_end = time();
        time_taken = t_end - t_begin;
        if (recovery_vals[iter, n_ind, p_ind] > tolerance)
          println("\t * Failure (",
                  round(recovery_vals[iter, n_ind, p_ind], 4),
                  ") after ", num_steps, " steps (",
                  round(time_taken, 3), " seconds) *");
        else
          println("\t * Success (",
                  round(recovery_vals[iter, n_ind, p_ind], 4),
                  ") after ", num_steps, " steps (",
                  round(time_taken, 3), " seconds) *");
        end
      end
    end
    result = vcat(recovery_vals, iteration_counts);
    writedlm(string(savepath, "/", filename), result, ' ');
  end
  return (recovery_vals, iteration_counts);
end

# (rfs, its) = HadamardZeroNoise(filename = "result_hadamart.txt";
#                                savepath = base_savepath,
#                                dim::Int64 = 1024,
#                                num_tests_per::Int64 = 100,
#                                alg_type::Symbol = :PL,
#                                num_n_vals::Int64 = 20,
#                                num_inits::Int64 = 10,
#                                max_init_ratio::Float64 = 2.0,
#                                tolerance::Float64 = 5e-3,
#                                signal_type::Symbol = :signs,
#                                seed::Int64 = 1117)
#
# Tests phase retrieval with (random) Hadamard measurement matrices and no
# noise. Performs num_tests_per experiments with dim-dimensional problems,
# where each experiment is as follows.
#
# 1. The sample size is chosen from 1.7, 1.8, ..., 1.7 + .1 * (num_n_vals - 1)
#
# 2. The initial point from which the iterative method begins is chosen
#    randomly (in a uniform direction) but such that
#
#    norm(x_init - x) / norm(x) = ratio,
#
#    where ratio is chosen from the linearly spaced set of max_init_ratio /
#    num_inits, 2 * max_init_ratio / num_inits, ..., max_init_ratio.
#
# The method run depends on alg_type, which is one of :PL (prox-linear) or :TAF
# (truncated amplitude flow).
#
# The tolerance is the level at which a failure is declared.
#
# Returns the num_tests_per-by-num_n_vals-by-num_inits boolean recovery
# failures (rfs) and integer-valued iteration counts (its) matrices.
function HadamardZeroNoise(filename = "result_hadamard_zero.txt";
                           savepath = base_savepath,
                           dim::Int64 = 1024,
                           num_tests_per::Int64 = 100,
                           alg_type::Symbol = :PL,
                           num_n_vals::Int64 = 20,
                           num_inits::Int64 = 10,
                           max_init_ratio::Float64 = 2.0,
                           tolerance::Float64 = 5e-3,
                           signal_type::Symbol = :signs,
                           seed::Int64 = 1117)
  srand(seed);
  recovery_failures = ones(Bool, num_tests_per,
                           num_n_vals, num_inits);
  iteration_counts = zeros(Int64, num_tests_per,
                           num_n_vals, num_inits);
  init_ratios = linspace(max_init_ratio / num_inits,
                         max_init_ratio, num_inits);
  k_mults = linspace(1.7, 1.7 + .1 * (num_n_vals - 1), num_n_vals);
  dim = 1 << Hadamard.intlog(dim);
  for iter = 1:num_tests_per
    println("*** [Experiment: ", iter, " of ", num_tests_per, "] ***");
    x = randn(dim);
    if (signal_type == :signs)
      x = sign.(x);
    end
    for n_ind = 1:num_n_vals
      kk = ceil(Int64, k_mults[n_ind]);
      k_low = floor(Int64, k_mults[n_ind]);
      (b, s_vec) = PRS.HadamardSignedData(x, kk, p_fail=0.0);
      num_k_diff_inds = floor(Int64, (k_mults[n_ind] - k_low) * dim);
      measurement_inds =
        vcat(1:(k_low * dim),
             sort(k_low * dim + randperm(dim)[1:num_k_diff_inds]));
      for ratio_ind = 1:num_inits
        ratio = init_ratios[ratio_ind];
        println("\t [Experiment: ", iter, "] n = ", length(measurement_inds),
                ", init ratio = ", ratio);
        delta_x = randn(dim);
        delta_x *= (ratio * norm(x) / norm(delta_x));
        x_init = x + delta_x;
        x_est = zeros(dim);
        iter_count = 0;
        if (alg_type == :PL)
          (x_est, objs, iter_count) =
            PRS.ProxLinearQuadratic(x_init, s_vec, b,
                                    measure_inds = measurement_inds,
                                    maxiter = 50,
                                    eps_qp = 1e-4,
                                    verbose_level = 0,
                                    qp_solver = :graph);
        elseif (alg_type == :TAF)
          (x_est, objs) =
            PRS.TruncatedAmpFlow(x_init, s_vec, b,
                                 measure_inds = measurement_inds,
                                 maxiter = 2000,
                                 tol = 1e-7,
                                 verbose_level = 0);
          iter_count = length(objs);
        else
          error("Unknown algorithm type ", alg_type);
        end
        err_norm = min(norm(x_est - x), norm(x_est + x)) / norm(x);
        recovery_failures[iter, n_ind, ratio_ind] = err_norm > tolerance;
        iteration_counts[iter, n_ind, ratio_ind] = iter_count;
        if (recovery_failures[iter, n_ind, ratio_ind])
          println("\t * Failure after ", iter_count, " steps (error = ",
                  err_norm, ") *");
        else
          println("\t * Success after ", iter_count, " steps *");
        end
      end
    end
    result = vcat(recovery_failures, iteration_counts);
    writedlm(string(savepath, "/", filename), result, ' ');
  end
  return (recovery_failures, iteration_counts);
end

# HadamardOutlier(filename = "result_hadamard_outlier.txt";
#                          savepath = base_savepath,
#                          dim::Int64 = 1024,
#                          num_p_fails::Int64 = 1,
#                          num_tests_per::Int64 = 100,
#                          alg_type::Symbol = :PL,
#                          init_type::Symbol = :SM,
#                          prop_init::Float64 = .6,
#                          signal_type::Symbol = :signs,
#                          tolerance::Float64 = 5e-3,
#                          dev::Float64 = 10.0,
#                          seed::Int64 = 902)
#
# Performs experiments with Hadamard data and a number of outliers.
function HadamardOutlier(filename = "result_hadamard_outlier.txt";
                         savepath = base_savepath,
                         dim::Int64 = 1024,
                         num_p_fails::Int64 = 1,
                         num_tests_per::Int64 = 100,
                         alg_type::Symbol = :PL,
                         init_type::Symbol = :SM,
                         prop_init::Float64 = .6,
                         signal_type::Symbol = :signs,
                         tolerance::Float64 = 5e-3,
                         dev::Float64 = 10.0,
                         seed::Int64 = 902)
  # Set seed number
  srand(seed);
  num_p_fails = max(1, min(num_p_fails, 5));
  p_fail_vals = [0.0, .025, .05, .1, .2];
  p_fail_vals = p_fail_vals[1:num_p_fails];
  k_mults = [2, 3, 4, 5, 6];
  recovery_failures = zeros(Bool, num_tests_per, length(k_mults),
                            length(p_fail_vals));
  iteration_counts = zeros(Int64, num_tests_per,
                           length(k_mults), length(p_fail_vals));
  dim = 1 << Hadamard.intlog(dim);
  for iter = 1:num_tests_per
    println("*** [Experiment: ", iter, " of ", num_tests_per, "] ***");
    for k_ind = 1:length(k_mults)
      for p_ind = 1:length(p_fail_vals)
        p_fail = p_fail_vals[p_ind];
        kk = k_mults[k_ind];
        println("\t [Experiment: ", iter, "] n = ", dim * kk,
                ", P(fail) = ", p_fail);
        x = randn(dim);
        if (signal_type == :signs)
          x = sign.(x);
        end
        (b, s_vec) = PRS.HadamardSignedData(x, kk, p_fail = p_fail, dev=dev);
        b[b.<= 0] = 0;
        x_init = zeros(dim);
        if (init_type == :SM)
          x_init = PRS.OrthogonalityInitializer(dim, s_vec, b,
                                                maxiter=250, p_val=prop_init,
                                                verbose_level=0, tol=1e-6,
                                                use_big = false);
        elseif (init_type == :LM)
          x_init = PRS.OrthogonalityInitializer(dim, s_vec, b,
                                                maxiter=250, p_val=prop_init,
                                                verbose_level=0, tol=1e-6,
                                                use_big = true);
        elseif (init_type == :rand)
          delta_x = randn(dim);
          delta_x = prop_init * norm(x) / norm(delta_x);
          x_init = x + delta_x;
        else
          error("unknown initialization type ", init_type);
        end
        num_steps = 0;
        x_est = zeros(dim);
        if (alg_type == :PL)
          (x_est, objs, num_steps) =
            PRS.ProxLinearQuadratic(x_init, s_vec, b, maxiter=50,
                                    qp_solver = :graph, eps_qp=1e-4,
                                    verbose_level=0);
        elseif (alg_type == :TAF)
          num_steps = 3000;
          (x_est, objs) =
            PRS.TruncatedAmpFlow(x_init, s_vec, b,
                                 maxiter = num_steps, verbose_level=0,
                                 stepsize = (p_ind > 1 ? sqrt(dd) :
                                             .6*sqrt(dd)));
        else
          error("Unknown algorithm ", alg_type);
        end
        recovery_failures[iter, k_ind, p_ind] =
          min(norm(x_est - x) / norm(x), norm(x_est + x) / norm(x)) > tolerance;
        iteration_counts[iter, k_ind, p_ind] = num_steps;
        if (recovery_failures[iter, k_ind, p_ind])
          println("\t * Failure after ", num_steps, " steps *");
        else
          println("\t * Success after ", num_steps, " steps *");
        end
      end
    end
    result = vcat(recovery_failures, iteration_counts);
    writedlm(string(savepath, "/", filename), result, ' ');
  end
  return (recovery_failures, iteration_counts);
end

# BasicPhaseRetrievalExperiment(filename = "result_noiseless.txt";
#                                       savepath = base_savepath,
#                                       wiggle_left = false,
#                                       wiggle_right = true,
#                                       condition_number = 1.0,
#                                       deviation = 0.0,
#                                       prop_corrupted = 0.0,
#                                       corruption_deviation = 5.0,
#                                       num_experiments = 100)
#
# This experiment investigates the noiseless case phase retrieval
# problem, in which we observe
#
#  b_i = |a_i^T x|^2
#
# where x is in dimension 50 and we observe n = 500 samples.
#
# Compares stochastic methods with the prox-linear method.
#
# If deviation > 0, then adds Laplace noise of scale deviation to the b_i
# scalars. If wiggle_right is true, then sets the condition number of A to be
# condition_number. If wiggle_left is true, then scales the matrix A so its
# rows range in norm from 1 to condition_number.
function BasicPhaseRetrievalExperiment(filename = "result_noiseless.txt";
                                       savepath = base_savepath,
                                       wiggle_left = false,
                                       wiggle_right = true,
                                       condition_number = 1.0,
                                       deviation = 0.0,
                                       prop_corrupted = 0.0,
                                       corruption_deviation = 5.0,
                                       num_experiments = 100)
  # Set seed number 
  srand(2017);
  Niter = 100000;
  dim = 50;
  nn = 500;
  rad = 1.0;
  num_prox_linear_iterations = convert(Int, Niter / nn);
  result_prox = zeros(num_experiments, num_prox_linear_iterations + 1);
  result_sprox = zeros(num_experiments, num_prox_linear_iterations + 1);
  result_sgd = zeros(num_experiments, num_prox_linear_iterations + 1);
  step_inits = [1000, 100, 10, 1];
  step_powers = [0.9, 0.8, 0.7, 0.6];
  all_choice = meshgrid(1:4, 1:4);
  stepchoice = zeros(16, 2);
  for iter = 1:16
    stepchoice[iter, 1] = step_inits[all_choice[iter, 1]];
    stepchoice[iter, 2] = step_powers[all_choice[iter, 2]];
  end
  num_pre_iter = 3 * nn;
    
  for iter = 1:num_experiments
    println("[Experiment:] ", iter, " of ", num_experiments);
    A = [];
    b = [];
    x = [];
    if (wiggle_right)
      (A, b, x) = PRS.GenerateData(nn, dim, dev = deviation,
                                    cond_num = condition_number,
                                    rad = rad);
    else
      (A, b, x) = PRS.GenerateData(nn, dim, dev = deviation,
                                    cond_num = condition_number,
                                    rad = rad);
    end
    if (wiggle_left)
      A = diagm(linspace(1, condition_number, nn)) * A;
    end
    if (prop_corrupted > 0)
      prop_corrupted = min(1.0, prop_corrupted);
      corrupted_inds = zeros(Int64, nn);
      num_corrupted = max(1, floor(Int64, prop_corrupted * nn));
      corrupted_inds[randperm(nn)[1:num_corrupted]] = 1;
      b[find(corrupted_inds)] = corruption_deviation * randn(num_corrupted);
    end
    x_initial = randn(dim);
    x_initial = rad * x_initial / norm(x_initial);
    (index_sprox_opt, index_sgd_opt) =
      GetInitialBestStepsize(A, b, num_pre_iter, stepchoice, 1e-5, x_initial);

    (x_curr, objs_prox) =
      PRS.ProxLinearQuadratics(A, b, lambda = 1e-5,
                                x_init = x_initial,
		                solver_type = 1,
                                maxiter = num_prox_linear_iterations,
                                verbose = false);
    result_prox[iter, :] = objs_prox;
    (x_curr, objs_sprox) =
      PRS.StochasticProxLinearQuad(
        A, b, Niter, 
        init_stepsize = stepchoice[index_sprox_opt, 1],
        steppow = stepchoice[index_sprox_opt, 2],
        lambda = 1e-5,
        compute_every = nn, verbose = false, x_init = x_initial);
    result_sprox[iter, :] = objs_sprox;
    (x_curr, objs_sgd) =
      PRS.StochasticGradientLinearQuad(
        A, b, Niter, 
        init_stepsize = stepchoice[index_sgd_opt, 1],
        steppow = stepchoice[index_sgd_opt, 2],
        lambda = 1e-5,
        compute_every = nn, verbose = false, x_init = x_initial);
    result_sgd[iter, :] = objs_sgd;
  end
  result = vcat(result_prox, result_sprox, result_sgd);
  writedlm(string(savepath, "/", filename), result, ' ');
end

# (index_sprox, index_sgd) =
#  GetInitialBestStepsize(A::Matrix{Float64}, b::Vector{Float64},
#                         num_pre_iter::Int64,
#                         stepchoices::Matrix{Float64},
#                         lambda::Float64, x_init::Vector{Float64})
#
# Find the "best" stepsize for the proximal stochastic and SGD methods by
# running each for num_pre_iter iterations using the stepsizes specified by the
# K-by-2 matrix stepchoices (stepchoices[:, 1] indexing initial stepsize,
# stepchoices[:, 2] indexing the power). Returns the index with the best
# performance after num_pre_iter iterations.
function GetInitialBestStepsize(A::Matrix{Float64}, b::Vector{Float64},
                                num_pre_iter::Int64,
                                stepchoices::Matrix{Float64},
                                lambda::Float64, x_initial::Vector{Float64})
  nn = size(A, 1);
  num_stepchoices = size(stepchoices, 1);
  result_sprox_pre = zeros(num_stepchoices);
  result_sgd_pre = zeros(num_stepchoices);
  for iter_inner = 1:num_stepchoices
    (x_curr, objs_sprox) =
      PRS.StochasticProxLinearQuad(
        A, b, num_pre_iter,
	init_stepsize = stepchoices[iter_inner, 1],
        steppow = stepchoices[iter_inner, 2], 
	lambda = lambda, compute_every = nn,
        verbose = false,
        x_init = x_initial);
    result_sprox_pre[iter_inner] = objs_sprox[end];
    (x_curr, objs_sgd) =
      PRS.StochasticGradientLinearQuad(
        A, b, num_pre_iter,
	init_stepsize = stepchoices[iter_inner, 1],
        steppow = stepchoices[iter_inner, 2], 
	lambda = lambda, compute_every = nn, verbose = false,
        x_init = x_initial);
    result_sgd_pre[iter_inner] = objs_sgd[end];
  end
  index_sprox_opt = indmin(result_sprox_pre);
  index_sgd_opt = indmin(result_sgd_pre);
  return (index_sprox_opt, index_sgd_opt);
end

end  # module PhaseExperiments
