# quadratic-objs.jl
#
# Uses quadratic objectives for phase retrieval, along with composite
# stepping.

# QuadraticProxLinear(A, b; x_init, eps_acc, maxiter)
#
# Performs a prox-linear method on the objective
#
#   ||(Ax)^2 - b||^2
#
# with a little line search to make sure we make progress. Iteratively
# sets delta to minimize
#
#   ||(Ax)^2 + 2 * (A * x).* A * delta - b||^2 + ||delta||^2 / alpha
#      = ||(Ax)^2 + 2 * D * A * delta - b||^2 + ||delta||^2 / alpha
#
# for D = diag(A * x), which has solution
#
#   delta = (A' * D^2 * A + I / alpha) \ (A' * D * ((Ax)^2 - b)).
#
# Then updates x = x + delta, and checks if enough progress is made.
# Terminates when delta / alpha is smaller than eps_acc.
function QuadraticProxLinear(A::Matrix{Float64}, b::Vector{Float64};
                             x_init::Vector{Float64} = randn(size(A, 2)),
                             eps_acc::Float64 = 1e-4,
                             maxiter::Int64 = 2 * size(A, 1),
                             verbose_level::Int64 = 1)
  proj_dec_mult = .4;  # Multiplier on projected decrease from delta
  steplength = 2 * eps_acc;
  stepsize = 1.0 / mean(A.^2);
  (m, n) = size(A);
  D = spdiagm(ones(m));
  grad = zeros(n);
  Ax = A * x_init;
  x = copy(x_init);
  Ax_next = Ax;
  delta = zeros(n);
  obj = mean(((Ax).^2 - b).^2) / 2;
  objs = zeros(maxiter + 1);
  objs[1] = obj;
  iter = 1;
  while (steplength > eps_acc && iter < maxiter)
    Ax = A * x;
    next_obj = Inf;
    delta_dot_grad = -Inf;
    grad = 2 * A' * (Ax .* ((Ax).^2 - b)) / m;
    while (next_obj > obj + proj_dec_mult * delta_dot_grad)
      if (verbose_level >= 2)
        println("\t Backtracking; stepsize ", stepsize);
      end
      D.nzval[:] = Ax;
      delta = -(4 * A' * D.^2 * A + eye(n) / stepsize) \
        (2 * A' * D * ((Ax).^2 - b));
      Ax_next = A * (x + delta);
      next_obj = mean(((Ax_next).^2 - b).^2) / 2;
      delta_dot_grad = dot(grad, delta);
      stepsize /= 2.0;
    end
    stepsize *= 2.0;  # Get stepsize back where it was
    steplength = norm(delta) / stepsize;
    # Increment stepsize so we make more progress hopefully.
    stepsize *= 2.0;
    x = x + delta;
    obj = next_obj;
    iter += 1;
    objs[iter] = obj;
    if (verbose_level >= 1)
      println("[Iteration ", iter-1, "]  Objective: ",
              obj, "  step: ", steplength);
    end
  end
  objs = objs[1:iter];
  return (x, objs);
end

# SubgradientL2Obj(A, b; x_init, eps_acc, maxiter)
#
# Performs subgradient method on the objective
#
#   ||(Ax)^2 - b||_2
#
# with Polyak stepping, which guarantees progress. Terminates when
#
#   ||x_k - x_{k+1}|| / alpha_k,
#
# where alpha_k is the final stepsize, is smaller than eps_acc.
function SubgradientL2Obj(A::Matrix{Float64}, b::Vector{Float64};
                          x_init::Vector{Float64} = randn(size(A, 2)),
                          eps_acc::Float64 = 1e-4,
                          maxiter::Int64 = 2 * size(A, 1))
  (m, n) = size(A);
  Ax = A * x_init;
  x = copy(x_init);
  obj = norm((Ax).^2 - b);
  objs = zeros(maxiter + 1);
  objs[1] = obj;
  iter = 1;
  steplength = Inf;
  while (steplength > eps_acc && iter < maxiter)
    next_obj = Inf;
    g = 2 * A' * (Ax .* ((Ax).^2 - b) / obj);
    stepsize = obj / norm(g)^2;
    # Approximate step length by "progress" guarantee in subgradient method
    steplength = obj * stepsize;
    x = x - stepsize * g;
    Ax = A * x;
    obj = norm((Ax).^2 - b);
    iter += 1;
    objs[iter] = obj;
    println("[Iteration ", iter-1, "]  Objective: ",
            obj, "  step: ", steplength);
  end
  objs = objs[1:iter];
  return (x, objs);
end

# TODO(jduchi): Write a version of this for Hadamard maybe? That could
# be interesting.
