# robust-quadratics.jl
#
# Methods that "solve" the non-linear least-absolute-deviations problem
#
# min. (1/n) sum_{i = 1}^n |(a_i^T x)^2 - b_i| + (lambda/2) * norm(x)^2
#
# where a_i are d-dimensional vectors and x is a d-vector as well.
# Note that this problem has the composite structure
#
# f(x) = (1/n) sum_{i = 1}^n h(c_i(x))
#
# where h is convex (but non-smooth) and c_i are smooth (Lipschitz
# gradient) functions.
#
# Author: John Duchi <jduchi@stanford.edu>
# Author: Feng Ruan <fengruan@stanford.edu>

include("phase-data-generation.jl");

# t = MinimizeOneDAbs(a, b)
#
# Sets t to be a minimizer of the function
#
# f(t) = sum(abs(a * t - b)),
#
# which we accomplish by a simple binary search. Only functions when a
# and b are positive. That is, returns a t_0 such that for some t* with
# |t* - t_0| < tol, we have
#
# sum_i sign(b_i - t * a_i) a_i > 0
#
# for t < t* and for t > t* we have
#
# sum_i sign(b_i - t * a_i) a_i < 0.
function MinimizeOneDAbs(a::Vector{Float64}, b::Vector{Float64};
                         tol::Float64 = 1e-8)
  nn = length(a);
  t_val = 0.0;
  max_t_val = norm(b, 1) / maximum(abs.(a));
  min_t_val = -max_t_val;
  nn = length(a);
  while (max_t_val - min_t_val > tol)
    t_val = (max_t_val + min_t_val) / 2;
    s = sum(sign.(a * t_val - b) .* a);
    if (s < 0)
      min_t_val = t_val;
    else
      max_t_val = t_val;
    end
  end
  return t_val;
end

# (x_guess, est_rad) = NoisyOrthPromotingInit(A::Matrix{Float64},
#                                b::Vector{Float64},
#                                prop_init::Float64 = .5,
#                                use_big::Bool = false)
#
# Similar to the OrthogonalityPromotingInitialization, except that
# this method first takes the fraction of data sorted by b, finds the
# smallest eigenvector of A, then tries to estimate the approximate
# norm of the true x by solving a particular convex problem.
function NoisyOrthPromotingInit(A::Matrix{Float64},
                                b::Vector{Float64},
                                prop_init::Float64 = .5,
                                use_big::Bool = false)
  (nn, dd) = size(A);
  A_norms = sqrt.(sum(A.^2, 2))[:];
  A_uniform = A .* repmat(1./A_norms, 1, dd);
  x_guess = OrthogonalityPromotingInitialization(A, b, prop_init=prop_init,
                                                 use_big=use_big);
  # Now, we would like to find the best guess of the norm of x. To
  # do so, we set t to minimize
  #
  # sum_{i = 1}^n |b_i / ||a_i||^2 - t * (u_i' * x_guess)^2|
  #
  # where u_i = a_i / ||a_i|| are the sphere-ized versions of the
  # a_i. The intuition is that this objective (in the noiseless case)
  # should be minimized approximately when t = ||x^*||^2, if x_guess
  # is in the direction of x^*. Otherwise, small perturbations should
  # not change it much.
  #
  # Simplifying by letting c_i = (u_i' * x_guess)^2, we must find a t_0
  # such that for all t < t_0, we have
  #
  # sum_i sign(b_i - t * c_i) c_i > 0,
  #
  # while for t > t_0 we have
  #
  # sum_i sign(b_i - t * c_i) c_i < 0.
  t = MinimizeOneDAbs((A_uniform * x_guess).^2, b ./ (A_norms.^2));
  return (x_guess[:], sqrt(t));
end

# x = OrthogonalityPromotingInitialization(A::Matrix{Float64},
#                                          b::Vector{Float64};
#                                          prop_init::Float64 = .5,
#                                          use_big::Bool = false)
#
# Uses the strategy of the Wang, Giannakis, Eldar paper (Solving Systems of
# Random Quadratic Equations via Truncated Amplitude Flow) to initialize
# a guess at the x that solves b = (A * x).^2. Assumes that the data
# (A, b) come from a process such that b = (A * x).^2 for some x.
#
# The initialization uses the idea that most of the random directions (rows of
# A) should satisfy a_i' * x = 0, so by sorting the response b, we have that
#
#  (1) x should be approximately orthogonal to the vectors a_i for which b[i]
#      is small.
#  (2) x should (hopefully?) be more aligned to the vectors a_i for which b[i]
#      is large.
#
# Has two guessing strategies, use_big, which decides which of strategies (1)
# and (2) to use (default is to use strategy (1)). The prop_init is the
# proportion of entries of b to use to initialize, i.e. find the x guess.
function OrthogonalityPromotingInitialization(A::Matrix{Float64},
                                              b::Vector{Float64};
                                              prop_init::Float64 = .5,
                                              use_big::Bool = false)
  (nn, dd) = size(A);
  n_use = min(ceil(Int64, prop_init * nn), nn);
  A_norms = sqrt.(sum(A.^2, 2))[:];
  inds = sortperm(b ./ (A_norms.^2), lt = <);
  A_uniform = A .* repmat(1./A_norms, 1, dd);
  S_small = A_uniform[inds[1:n_use], :]' * A_uniform[inds[1:n_use], :];
  S_big = (A_uniform[inds[(nn - n_use + 1):end], :]'
           * A_uniform[inds[(nn - n_use + 1):end], :]);
  S_small = Symmetric(S_small / n_use, :U);
  S_big = Symmetric(S_big / (nn - n_use + 1), :U);
  x_guess = zeros(dd);
  if (use_big)
    data = eigs(S_big, nev = 1, which=:LM);
    x_guess = data[2];
  else
    # Get smallest eigenvalue
    (D, V) = eig(S_small);
    if (D[1] <= D[end])
      x_guess = V[:, 1];
    else
      x_guess = V[:, end];
    end
  end
  return x_guess[:];
end

# (x_guess, mag) = MedianWirtingerInitialization(A, b; y_thresh = 3.0)
#
# Computes a guess at the direction and magnitude of the signal x generating
# (most of) b = (A * x).^2, with outliers. Uses generalization of Candes et
# al.'s initialization based on Zhang, Chi, Liang 2016 paper.
function MedianWirtingerInitialization(A::Matrix{Float64}, b::Vector{Float64};
                                       y_thresh::Float64 = 3.0)
  (nn, dd) = size(A);
  magnitude_est = sqrt(median(b) / .455);
  b_diag = spdiagm(b .* (abs.(b) .<= y_thresh^2 * magnitude_est^2));
  Y = A' * b_diag * A;
  Y = Symmetric(Y / nn, :U);
  data = eigs(Y, nev = 1, which = :LM);
  x_guess = data[2];
  return (x_guess[:], magnitude_est);
end

# (x, objs, iters) =
#   ProxLinearQuadratic(A::Matrix{Float64}, b::Vector{Float64};
#                       x_init::Vector{Float64} = randn(size(A, 2)),
#                       tol::Float64 = 1e-3,
#                       maxiter::Int64 = length(b),
#                       verbose_level::Int64 = 1,
#                       eps_qp::Float64 = 1e-4,
#                       qp_solver::Symbol = :mosek)
#
# Using the linearized proximal iteration to (locally) minimize the
# problem described above. Initializes the iterations from the given
# point x_init (randomly chosen if unspecified) from a Gaussian with
# variance 1.
#
# Method stops iterating when the gradient mapping has norm less than
# tol. Returns the objective over the iterations. If qp_solver ==
# :mosek, then uses Mosek as the solver. If qp_solver == :ecos, uses
# ECOS. if qp_solver == :graph, uses the graph form solver (see
# SolveQP_graph_form in the file admm-lone-solver.jl).
function ProxLinearQuadratic(A::Matrix{Float64}, b::Vector{Float64};
                             x_init::Vector{Float64} = randn(size(A, 2)),
                             tol::Float64 = 1e-3,
                             maxiter::Int64 = length(b),
                             verbose_level::Int64 = 1,
                             eps_qp::Float64 = 1e-4,
                             qp_solver::Symbol = :mosek)
  x_curr = x_init;
  steplength = Inf;
  (nn, dd) = size(A);
  lipschitz_const = 2 * norm(A);
  iter = 0;
  total_iters = 0;
  objs = [];
  while (iter < maxiter && steplength > tol)
    obj = sum(abs.((A * x_curr).^2 - b) / nn);
    objs = cat(1, objs, obj);
    if (verbose_level >= 1)
      println("[Iteration ", iter, "]  Objective: ",
            obj, "    step: ", steplength);
    end
    dx = zeros(dd);
    # Minimize the loss
    # norm(inner_grad_mat * delta_x - (b - (A * x_curr).^2), 1)
    #  + (lipschitz_const / 2) * ||delta_x||_2^2
    inner_grad_mat = 2 * diagm(A * x_curr) * A;
    if (qp_solver == :ecos || qp_solver == :mosek)
      quietude = (verbose_level < 2);
      delta_x = Variable(dd);
      obj = minimize(sum(abs.((A * x_curr).^2
                             + inner_grad_mat * delta_x - b))
                     + (lipschitz_const / 2) * sumsquares(delta_x));
      if (qp_solver == :ecos)
        solve!(obj, ECOSSolver(verbose=!quietude));
      else  # qp_solver == :mosek
        solve!(obj, MosekSolver(QUIET=quietude));
      end
      dx = delta_x.value;
    elseif (qp_solver == :graph)
      (dx, iter_count) =
        SolveQP_graph_form(inner_grad_mat, b - (A * x_curr).^2,
                           eps_abs = eps_qp, eps_rel = eps_qp, rho = 4.0,
                           verbose_level = verbose_level);
      total_iters += iter_count;
    else
      error("Unknown QP solver ", qp_solver);
    end
    grad_map = lipschitz_const * dx;
    x_curr = x_curr + dx[:];
    steplength = norm(grad_map);
    iter = iter + 1;
  end
  obj = sum(abs.((A * x_curr).^2 - b) / nn);
  if (verbose_level >= 1)
    println("[Iteration ", iter, "]  Objective: ",
          obj, "    step: ", steplength);
  end
  objs = cat(1, objs, obj);
  if (qp_solver != :graph)
    Convex.clearmemory();
  end
  if (qp_solver == :graph)
    return (x_curr, objs, total_iters);
  else
    return (x_curr, objs, iter);
  end
end
                              
# (x, obj) = ProxLinearQuadratics(A::Matrix{Float64}, b::Vector{Float64};
#                                 x_init::Vector{Float64} = zeros(size(A, 2)),
#                                 lambda::Float64 = 1.0,
#                                 tol::Float64 = 1e-3,
#                                 solver_type::Int64 = 0)
#
# Using the linearized proximal iteration to (locally) minimize the
# problem described above. Initializes the iterations from the given
# point x_init (randomly chosen if x_init = 0) from a Gaussian with variance
#
# 1 / lambda.
#
# Method stops iterating when the gradient mapping has norm less than
# tol. Returns the objective over the iterations. If solver_type == 0, then
# uses ECOS as the solver. Otherwise, uses MosekSolver as the internal solver.
function ProxLinearQuadratics(A::Matrix{Float64}, b::Vector{Float64};
                              x_init::Vector{Float64} = zeros(size(A, 2)),
                              lambda::Float64 = 1.0,
                              tol::Float64 = 1e-3,
                              solver_type::Int64 = 0, 
                              maxiter::Int64 = 2 * size(A, 1), 
                              verbose::Bool = true)
  x_curr = x_init;
  if (norm(x_init) < 1e-5)
    x_curr = randn(size(A, 2)) / max(1, sqrt(lambda));
  end
  steplength = Inf;
  (nn, dd) = size(A);
  inner_lipschitz = sum(A.^2) / nn;
  AAt = A * A';
  iter = 0;
  objs = [];
  while (iter < maxiter)
    obj = sum(abs.((A * x_curr).^2 - b) / nn) + (lambda/2) * norm(x_curr)^2;
    objs = cat(1, objs, obj);
    if(verbose)
      println("[Iteration ", iter, "]  Objective: ",
            obj, "    step: ", steplength);
    end
    x_next = Variable(dd);
    inner_grad_mat = 2 * diagm(A * x_curr) * A;
    obj = minimize(sum(abs.((A * x_curr).^2
                           + inner_grad_mat * (x_next - x_curr) - b)) / nn
                   + (inner_lipschitz / 2) * sumsquares(x_next - x_curr)
                   + (lambda/2) * sumsquares(x_next));
    # obj.constraints += (norm(x_next) <= rad);
    if (solver_type == 0)
      solve!(obj, ECOSSolver(verbose=false));
    else
      solve!(obj, MosekSolver(QUIET=true));
    end
    grad_map = inner_lipschitz * (x_next.value - x_curr);
    x_curr = x_next.value[:];
    steplength = norm(grad_map);
    iter = iter + 1;
  end
  obj = sum(abs.((A * x_curr).^2 - b) / nn) + (lambda/2) * norm(x_curr)^2;
  if(verbose)
    println("[Iteration ", iter, "]  Objective: ",
          obj, "    step: ", steplength);
  end
  objs = cat(1, objs, obj);
  Convex.clearmemory();
  return (x_curr, objs);
end

# (x, obj) = ProxLinearLineSearch(A::Matrix{Float64}, b::Vector{Float64};
#                                 x_init::Vector{Float64} = zeros(size(A, 2)),
#                                 lambda::Float64 = 1.0,
#                                 tol::Float64 = 1e-3,
#                                 solver_type::Int64 = 0)
#
# Using the linearized proximal iteration to (locally) minimize the
# problem described above. Initializes the iterations from the given
# point x_init (randomly chosen if x_init = 0) from a Gaussian with variance
#
# 1 / lambda.
#
# Method stops iterating when the gradient mapping has norm less than
# tol. Returns the objective over the iterations. If solver_type == 0, then
# uses ECOS as the solver. Otherwise, uses MosekSolver as the internal solver.
function ProxLinearLineSearch(A::Matrix{Float64}, b::Vector{Float64};
                              x_init::Vector{Float64} = zeros(size(A, 2)),
                              lambda::Float64 = 1.0,
                              tol::Float64 = 1e-3,
                              solver_type::Int64 = 0, 
                              maxiter::Int64 = 2 * size(A, 1), 
                              verbose::Bool = true)
  x_curr = x_init;
  if (norm(x_init) < 1e-5)
    x_curr = randn(size(A, 2)) / max(1, sqrt(lambda));
  end
  steplength = Inf;
  (nn, dd) = size(A);
  inner_lipschitz = sum(A.^2) / nn;
  const_inner_lipschitz = sum(A.^2) / nn;
  AAt = A * A';
  iter = 0;
  objs = [];
  beta_suff_decrease = .8;
  while (steplength > tol && iter <= maxiter)
    iter += 1;
    obj = sum(abs.((A * x_curr).^2 - b) / nn) + (lambda/2) * norm(x_curr)^2;
    objs = cat(1, objs, obj);
    if (verbose)
    println("[Iteration ", iter, "]  Objective: ",
            round(obj, 7), "  grad-norm: ", round(steplength, 5),
            "  Lip: ", round(inner_lipschitz, 5));
    end
    x_next = Variable(dd);
    inner_grad_mat = 2 * diagm(A * x_curr) * A;
    # Now perform line search
    #
    # We should have f(x + Delta) - f(x) <= beta * (f_x(x + Delta) - f(x))
    # for some beta, where Delta is the update from the linearized problem.
    #
    # We perform a backtracking line-search over "stepsizes," which must
    # terminate eventually.
    predicted_new_obj = -Inf;
    new_obj = Inf;
    while (new_obj - objs[end] >
           beta_suff_decrease * (predicted_new_obj - objs[end]))
      obj = minimize(sum(abs.((A * x_curr).^2
                             + inner_grad_mat * (x_next - x_curr) - b)) / nn
                     + (inner_lipschitz / 2) * sumsquares(x_next - x_curr)
                     + (lambda/2) * sumsquares(x_next));
      # obj.constraints += (norm(x_next) <= rad);
      if (solver_type == 0)
        solve!(obj, ECOSSolver(verbose=false));
      else
        solve!(obj, MosekSolver(QUIET=true));
      end
      predicted_new_obj = (sum(abs.((A * x_curr).^2 +
                                   inner_grad_mat * (x_next.value - x_curr) -
                                   b)) / nn
                           + (lambda / 2) * sum(x_next.value.^2));
      new_obj = (sum(abs.((A * x_next.value).^2 - b)) / nn +
                 (lambda / 2) * sum(x_next.value.^2));
      grad_map = const_inner_lipschitz * (x_next.value - x_curr);
      steplength = norm(grad_map);
      inner_lipschitz = 2 * inner_lipschitz;
    end
    x_curr = x_next.value[:];
    iter = iter + 1;
    inner_lipschitz = inner_lipschitz / 4;
  end
  obj = sum(abs.((A * x_curr).^2 - b) / nn) + (lambda/2) * norm(x_curr)^2;
  if(verbose)
    println("[Iteration ", iter, "]  Objective: ",
            round(obj, 7), "  grad-norm: ", round(steplength, 5),
            "  Lip: ", round(inner_lipschitz, 5));
  end
  objs = cat(1, objs, obj);
  Convex.clearmemory();
  return (x_curr, objs);
end

# (x, obj) = StochasticProxLinearQuad(A::Matrix{Float64},
#                                  b::Vector{Float64},
#                                  maxiter::Int64 = 2 * size(A, 1);
#                                  x_init::Vector{Float64} = zeros(size(A, 2)),
#                                  lambda::Float64 = 1.0,
#                                  init_stepsize = 1.0,
#                                  steppow = .75,
#                                  compute_every::Int64 = ceil(Int64, size(A, 1)/2))
#
# Using the linearized proximal iteration to (locally) minimize the
# problem described above. Initializes the iterations from the given
# point x_init (randomly chosen if x_init = 0) from a Gaussian with variance
#
# 1 / lambda.
#
# Method stops iterating when the gradient mapping has norm less than
# tol. Returns the objective over the iterations.
function StochasticProxLinearQuad(A::Matrix{Float64},
                                  b::Vector{Float64},
                                  maxiter::Int64 = 2 * size(A, 1);
                                  x_init::Vector{Float64} = zeros(size(A, 2)),
                                  lambda::Float64 = 1.0,
                                  init_stepsize = 1.0,
                                  steppow = .75,
                                  compute_every::Int64 = ceil(Int64, size(A, 1)/2), 
                                  verbose::Bool = true)
  x_curr = x_init;
  if (norm(x_init) < 1e-5)
    if (lambda < 1)
      x_curr = randn(size(A, 2));
    else
      x_curr = randn(size(A, 2)) / sqrt(lambda);
    end
  end
  nn = size(A, 1);
  objs = [];
  for iter = 1:maxiter
    if (mod(iter - 1, compute_every) == 0)
      obj = sum(abs.((A * x_curr).^2 - b) / nn) + (lambda/2) * norm(x_curr)^2;
      objs = cat(1, objs, obj);
      if(verbose)
        println("[Iteration ", iter, "] Objective: ", obj);
      end
    end
    stepsize = init_stepsize * iter^(-steppow);
    ind = ceil(Int64, rand() * nn);
    A_ind = A[ind, :][:];
    # Construct linear approximation
    c = -dot(A_ind, x_curr)^2 - b[ind];
    lin_part = 2 * dot(A_ind, x_curr) * A_ind;
    # Now update x to be the minimizer of
    #
    # |c + lin_part' * x| + (1 / (2 * stepsize)) * norm(x - x_curr)^2
    #   + (lambda / 2) * norm(x)^2,
    #
    # which is equivalent to minimizing
    #
    # stepsize * |c + lin_part' * x| +
    #   (1 + stepsize * lambda)/2 * norm(x - x_curr / (1 + stepsize * lambda))^2.
    #
    center_vec = x_curr / (1 + stepsize * lambda);
    c = stepsize * c / (1 + stepsize * lambda);
    a = stepsize * lin_part / (1 + stepsize * lambda);
    x_curr = MinQuadPlusAbs(c, a, center_vec);
  end
  obj = sum(abs.((A * x_curr).^2 - b) / nn) + (lambda/2) * norm(x_curr)^2;
  objs = cat(1, objs, obj);
  if(verbose)
    println("[Final iteration] Objective: ", obj);
  end
  return (x_curr, objs);
end

# MinQuadPlusAbs(c::Float64, a::Vector{Float64}, u::Vector{Float64})
#
# Sets x to be the minimizer of
#
#  |c + a'*x| + (1/2) * norm(x - u)^2
#
# over all x.
function MinQuadPlusAbs(c::Float64, a::Vector{Float64}, u::Vector{Float64})
  lambda = -(dot(u, a) + c) / sum(a.^2);
  lambda = min(1.0, max(-1.0, lambda));
  x = u + lambda * a;
  return x;
end

# (x, objs) = StochasticGradientLinearQuad(A::Matrix{Float64},
#                                      b::Vector{Float64},
#                                      maxiter::Int64 = 2 * size(A, 1);
#                                      x_init::Vector{Float64} = zeros(size(A, 2)),
#                                      lambda::Float64 = 1.0,
#                                      init_stepsize = 1.0,
#                                      steppow = .75,
#                                      compute_every::Int64 = ceil(Int64, size(A, 1)/2),
#                                      do_project::Bool = true)
#
# Performs stochastic gradient descent on the absolute error objective
# |(a^T x)^2 - b|, while incorporating the L2 regularization
# parameter.
#
# If do_project is true, then performs projections so that x is
# constrained to lie in the set
#
# { x : norm(x) <= sqrt(2 * norm(b, 1) / (n * lambda)) },
#
# which is implied by the positivity of the objective.
function StochasticGradientLinearQuad(A::Matrix{Float64},
                                      b::Vector{Float64},
                                      maxiter::Int64 = 2 * size(A, 1);
                                      x_init::Vector{Float64} = zeros(size(A, 2)),
                                      lambda::Float64 = 1.0,
                                      init_stepsize = 1.0,
                                      steppow = .75,
                                      compute_every::Int64 = ceil(Int64, size(A, 1)/2),
                                      do_project::Bool = true, 
                                      verbose::Bool = true)
  x_curr = x_init;
  nn = size(A, 1);
  max_x_norm = sqrt(2 * sum(abs.(b)) / (nn * lambda));
  if (norm(x_init) < 1e-5)
    x_curr = randn(size(A, 2));
    x_curr = max_x_norm * x_curr / (2 * norm(x_curr));
  end
  objs = [];
  for iter = 1:maxiter
    if (mod(iter - 1, compute_every) == 0)
      obj = sum(abs.((A * x_curr).^2 - b) / nn) + (lambda/2) * norm(x_curr)^2;
      objs = cat(1, objs, obj);
      if(verbose)
        println("[Iteration ", iter, "] Objective: ", obj);
      end
    end
    stepsize = init_stepsize * iter^(-steppow);
    ind = ceil(Int64, rand() * nn);
    A_ind = A[ind, :][:];
    margin = dot(A_ind, x_curr)^2 - b[ind];
    g = 2 * sign(margin) * dot(A_ind, x_curr) * A_ind;
    x_curr = (x_curr - stepsize * g) / (1 + lambda * stepsize);
    if (do_project && norm(x_curr) > max_x_norm)
      x_curr = max_x_norm * x_curr / norm(x_curr);
    end
  end
  obj = sum(abs.((A * x_curr).^2 - b) / nn) + (lambda/2) * norm(x_curr)^2;
  objs = cat(1, objs, obj);
  if(verbose)
    println("[Final iteration] Objective: ", obj);
  end
  return (x_curr, objs);
end

