# simple-solvers.jl
#
# Includes only the solvers based on gradient descent and related ideas.

module PRS

using Base.LinAlg;
using Base.BLAS;

include("phase-data-generation.jl");
include("gradient-solvers.jl");
include("quadratic-objs.jl");

# x = OrthogonalityPromotingInitialization(A::Matrix{Float64},
#                                          b::Vector{Float64};
#                                          prop_init::Float64 = .5,
#                                          use_big::Bool = false)
#
# Uses the strategy of the Wang, Giannakis, Eldar paper (Solving Systems of
# Random Quadratic Equations via Truncated Amplitude Flow) to initialize
# a guess at the x that solves b = (A * x).^2. Assumes that the data
# (A, b) come from a process such that b = (A * x).^2 for some x.
#
# The initialization uses the idea that most of the random directions (rows of
# A) should satisfy a_i' * x = 0, so by sorting the response b, we have that
#
#  (1) x should be approximately orthogonal to the vectors a_i for which b[i]
#      is small.
#  (2) x should (hopefully?) be more aligned to the vectors a_i for which b[i]
#      is large.
#
# Has two guessing strategies, use_big, which decides which of strategies (1)
# and (2) to use (default is to use strategy (1)). The prop_init is the
# proportion of entries of b to use to initialize, i.e. find the x guess.
function OrthogonalityPromotingInitialization(A::Matrix{Float64},
                                              b::Vector{Float64};
                                              prop_init::Float64 = .5,
                                              use_big::Bool = false)
  (nn, dd) = size(A);
  n_use = min(ceil(Int64, prop_init * nn), nn);
  A_norms = sqrt.(sum(A.^2, 2))[:];
  inds = sortperm(b ./ (A_norms.^2), lt = <);
  A_uniform = A .* repmat(1./A_norms, 1, dd);
  S_small = A_uniform[inds[1:n_use], :]' * A_uniform[inds[1:n_use], :];
  S_big = (A_uniform[inds[(nn - n_use + 1):end], :]'
           * A_uniform[inds[(nn - n_use + 1):end], :]);
  S_small = Symmetric(S_small / n_use, :U);
  S_big = Symmetric(S_big / (nn - n_use + 1), :U);
  x_guess = zeros(dd);
  if (use_big)
    data = eigs(S_big, nev = 1, which=:LM);
    x_guess = data[2];
  else
    # Get smallest eigenvalue
    (D, V) = eig(S_small);
    if (D[1] <= D[end])
      x_guess = V[:, 1];
    else
      x_guess = V[:, end];
    end
  end
  return x_guess[:];
end

end  # module PRS
